@@ -1,4 +1,4 @@
-/* $OpenLDAP: pkg/ldap/clients/mail500/main.c,v 1.25.2.3 2000/06/13 17:57:02 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/clients/mail500/main.c,v 1.44 2002/01/04 20:17:30 kurt Exp $ */
 /*
  * Copyright (c) 1990 Regents of the University of Michigan.
  * All rights reserved.
@@ -10,7 +10,7 @@
  * software without specific prior written permission. This software
  * is provided ``as is'' without express or implied warranty.
  *
- * Copyright 1998-2000 The OpenLDAP Foundation
+ * Copyright 1998-2002 The OpenLDAP Foundation
  * COPYING RESTRICTIONS APPLY.  See COPYRIGHT File in top level directory
  * of this package for details.
  */
@@ -108,13 +108,13 @@ typedef struct baseinfo {
 } Base;
 
 Base	base[] = {
-	{"ou=People, dc=OpenLDAP, dc=org",
+	{"ou=People, dc=example, dc=com",
 		0, USER,
 		{"uid=%s", "cn=%s", NULL}},
-	{"ou=System Groups, ou=Groups, dc=OpenLDAP, dc=org",
+	{"ou=System Groups, ou=Groups, dc=example, dc=com",
 		1, 0xff,
 		{"(&(cn=%s)(associatedDomain=%h))", NULL, NULL}},
-	{"ou=User Groups, ou=Groups, dc=OpenLDAP, dc=org",
+	{"ou=User Groups, ou=Groups, dc=example, dc=com",
 		1, 0xff,
 		{"(&(cn=%s)(associatedDomain=%h))", NULL, NULL}},
 	{NULL}
@@ -163,7 +163,7 @@ main ( int argc, char **argv )
 	int		numto, ngroups, numerr, nargs;
 	int		i, j;
 
-	if ( (myname = strrchr( argv[0], '/' )) == NULL )
+	if ( (myname = strrchr( argv[0], *LDAP_DIRSEP )) == NULL )
 		myname = strdup( argv[0] );
 	else
 		myname = strdup( myname + 1 );
