@@ -1,6 +1,6 @@
-/* $OpenLDAP: pkg/ldap/clients/tools/ldapmodify.c,v 1.45.2.7 2000/07/04 17:58:44 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/clients/tools/ldapmodify.c,v 1.109.2.1 2002/04/18 15:19:55 kurt Exp $ */
 /*
- * Copyright 1998-2000 The OpenLDAP Foundation, All Rights Reserved.
+ * Copyright 1998-2002 The OpenLDAP Foundation, All Rights Reserved.
  * COPYING RESTRICTIONS APPLY, see COPYRIGHT file
  */
 /* ldapmodify.c - generic program to modify or add entries using LDAP */
@@ -29,24 +29,28 @@
 
 #include <ldap.h>
 
+#include "lutil_ldap.h"
 #include "ldif.h"
 #include "ldap_defaults.h"
+#include "ldap_log.h"
 
 static char	*prog;
 static char	*binddn = NULL;
-static struct berval passwd = { 0, NULL};
+static struct berval passwd = { 0, NULL };
+static char *ldapuri = NULL;
 static char	*ldaphost = NULL;
 static int	ldapport = 0;
 #ifdef HAVE_CYRUS_SASL
+static unsigned sasl_flags = LDAP_SASL_AUTOMATIC;
+static char *sasl_realm = NULL;
 static char	*sasl_authc_id = NULL;
 static char	*sasl_authz_id = NULL;
 static char	*sasl_mech = NULL;
-static int	sasl_integrity = 0;
-static int	sasl_privacy = 0;
+static char	*sasl_secprops = NULL;
 #endif
 static int	use_tls = 0;
-static int	ldapadd, replace, not, verbose, contoper, force;
-static LDAP	*ld;
+static int	ldapadd, not, verbose, contoper, force;
+static LDAP	*ld = NULL;
 
 #define LDAPMOD_MAXLINE		4096
 
@@ -97,32 +101,36 @@ usage( const char *prog )
 "usage: %s [options]\n"
 "	The list of desired operations are read from stdin or from the file\n"
 "	specified by \"-f file\".\n"
-"options:\n"
-"	-a\t\tadd values (default%s)\n"
-"	-b\t\tread values from files (for binary attributes)\n"
-"	-c\t\tcontinuous operation\n"
-"	-C\t\tchase referrals\n"
-"	-d level\tset LDAP debugging level to `level'\n"
-"	-D dn\t\tbind DN\n"
-"	-E\t\trequest SASL privacy (-EE to make it critical)\n"
-"	-f file\t\tperform sequence of operations listed in file\n"
-"	-F\t\tforce all changes records to be used\n"
-"	-h host\t\tLDAP server\n"
-"	-I\t\trequest SASL integrity checking (-II to make it\n"
-"		\tcritical)\n"
-"	-k\t\tuse Kerberos authentication\n"
-"	-K\t\tlike -k, but do only step 1 of the Kerberos bind\n"
-"	-M\t\tenable Manage DSA IT control (-MM to make it critical)\n"
-"	-n\t\tprint changes, don't actually do them\n"
-"	-p port\t\tport on LDAP server\n"
-"	-r\t\treplace values\n"
-"	-U user\t\tSASL authentication identity (username)\n"
-"	-v\t\tverbose mode\n"
-"	-w passwd\tbind password (for Simple authentication)\n"
-"	-X id\t\tSASL authorization identity (\"dn:<dn>\" or \"u:<user>\")\n"
-"	-Y mech\t\tSASL mechanism\n"
-"	-Z\t\tissue Start TLS request (-ZZ to require successful response)\n"
+"Add or modify options:\n"
+"  -a         add values (default%s)\n"
+"  -F         force all changes records to be used\n"
+
+"Common options:\n"
+"  -d level   set LDAP debugging level to `level'\n"
+"  -D binddn  bind DN\n"
+"  -f file    read operations from `file'\n"
+"  -h host    LDAP server\n"
+"  -H URI     LDAP Uniform Resource Indentifier(s)\n"
+"  -I         use SASL Interactive mode\n"
+"  -k         use Kerberos authentication\n"
+"  -K         like -k, but do only step 1 of the Kerberos bind\n"
+"  -M         enable Manage DSA IT control (-MM to make critical)\n"
+"  -n         show what would be done but don't actually update\n"
+"  -O props   SASL security properties\n"
+"  -p port    port on LDAP server\n"
+"  -P version procotol version (default: 3)\n"
+"  -Q         use SASL Quiet mode\n"
+"  -R realm   SASL realm\n"
+"  -U authcid SASL authentication identity\n"
+"  -v         run in verbose mode (diagnostics to standard output)\n"
+"  -w passwd  bind passwd (for simple authentication)\n"
+"  -W         prompt for bind passwd\n"
+"  -x         Simple authentication\n"
+"  -X authzid SASL authorization identity (\"dn:<dn>\" or \"u:<user>\")\n"
+"  -Y mech    SASL mechanism\n"
+"  -Z         Start TLS request (-ZZ to require successful response)\n"
 	     , prog, (strcmp( prog, "ldapadd" ) ? " is to replace" : "") );
+
     exit( EXIT_FAILURE );
 }
 
@@ -144,200 +152,393 @@ main( int argc, char **argv )
     /* Print usage when no parameters */
     if( argc < 2 ) usage( prog );
 
-    ldapadd = ( strcmp( prog, "ldapadd" ) == 0 );
+	/* strncmp instead of strcmp since NT binaries carry .exe extension */
+    ldapadd = ( strncmp( prog, "ldapadd", sizeof("ldapadd")-1 ) == 0 );
 
     infile = NULL;
     not = verbose = want_bindpw = debug = manageDSAit = referrals = 0;
-    authmethod = LDAP_AUTH_SIMPLE;
+    authmethod = -1;
 	version = -1;
 
-    while (( i = getopt( argc, argv, "acCD:d:EFf:h:IKkMnP:p:rtU:vWw:X:Y:Z" )) != EOF ) {
+    while (( i = getopt( argc, argv, "acrf:F"
+		"Cd:D:h:H:IkKMnO:p:P:QR:U:vw:WxX:Y:Z" )) != EOF )
+	{
 	switch( i ) {
+	/* Modify Options */
 	case 'a':	/* add */
 	    ldapadd = 1;
 	    break;
 	case 'c':	/* continuous operation */
 	    contoper = 1;
 	    break;
-	case 'C':
-		referrals++;
-		break;
-	case 'r':	/* default is to replace rather than add values */
-	    replace = 1;
-	    break;
-	case 'k':	/* kerberos bind */
-#ifdef LDAP_API_FEATURE_X_OPENLDAP_V2_KBIND
-		authmethod = LDAP_AUTH_KRBV4;
-#else
-		fprintf( stderr, "%s was not compiled with Kerberos support\n", argv[0] );
-		return( EXIT_FAILURE );
-#endif
-	    break;
-	case 'K':	/* kerberos bind, part 1 only */
-#ifdef LDAP_API_FEATURE_X_OPENLDAP_V2_KBIND
-		authmethod = LDAP_AUTH_KRBV41;
-#else
-		fprintf( stderr, "%s was not compiled with Kerberos support\n", argv[0] );
-		return( EXIT_FAILURE );
-#endif
+	case 'f':	/* read from file */
+		if( infile != NULL ) {
+			fprintf( stderr, "%s: -f previously specified\n", prog );
+			return EXIT_FAILURE;
+		}
+	    infile = strdup( optarg );
 	    break;
 	case 'F':	/* force all changes records to be used */
 	    force = 1;
 	    break;
-	case 'h':	/* ldap host */
-	    ldaphost = strdup( optarg );
+
+	/* Common Options */
+	case 'C':
+		referrals++;
+		break;
+	case 'd':
+	    debug |= atoi( optarg );
 	    break;
 	case 'D':	/* bind DN */
+		if( binddn != NULL ) {
+			fprintf( stderr, "%s: -D previously specified\n", prog );
+			return EXIT_FAILURE;
+		}
 	    binddn = strdup( optarg );
 	    break;
+	case 'h':	/* ldap host */
+		if( ldapuri != NULL ) {
+			fprintf( stderr, "%s: -h incompatible with -H\n", prog );
+			return EXIT_FAILURE;
+		}
+		if( ldaphost != NULL ) {
+			fprintf( stderr, "%s: -h previously specified\n", prog );
+			return EXIT_FAILURE;
+		}
+	    ldaphost = strdup( optarg );
+	    break;
+	case 'H':	/* ldap URI */
+		if( ldaphost != NULL ) {
+			fprintf( stderr, "%s: -H incompatible with -h\n", prog );
+			return EXIT_FAILURE;
+		}
+		if( ldapport ) {
+			fprintf( stderr, "%s: -H incompatible with -p\n", prog );
+			return EXIT_FAILURE;
+		}
+		if( ldapuri != NULL ) {
+			fprintf( stderr, "%s: -H previously specified\n", prog );
+			return EXIT_FAILURE;
+		}
+	    ldapuri = strdup( optarg );
+	    break;
+	case 'I':
+#ifdef HAVE_CYRUS_SASL
+		if( version == LDAP_VERSION2 ) {
+			fprintf( stderr, "%s: -I incompatible with version %d\n",
+				prog, version );
+			return EXIT_FAILURE;
+		}
+		if( authmethod != -1 && authmethod != LDAP_AUTH_SASL ) {
+			fprintf( stderr, "%s: incompatible previous "
+				"authentication choice\n",
+				prog );
+			return EXIT_FAILURE;
+		}
+		authmethod = LDAP_AUTH_SASL;
+		version = LDAP_VERSION3;
+		sasl_flags = LDAP_SASL_INTERACTIVE;
+		break;
+#else
+		fprintf( stderr, "%s: was not compiled with SASL support\n",
+			prog );
+		return( EXIT_FAILURE );
+#endif
+	case 'k':	/* kerberos bind */
+#ifdef LDAP_API_FEATURE_X_OPENLDAP_V2_KBIND
+		if( version > LDAP_VERSION2 ) {
+			fprintf( stderr, "%s: -k incompatible with LDAPv%d\n",
+				prog, version );
+			return EXIT_FAILURE;
+		}
+
+		if( authmethod != -1 ) {
+			fprintf( stderr, "%s: -k incompatible with previous "
+				"authentication choice\n", prog );
+			return EXIT_FAILURE;
+		}
+			
+		authmethod = LDAP_AUTH_KRBV4;
+#else
+		fprintf( stderr, "%s: not compiled with Kerberos support\n", prog );
+		return EXIT_FAILURE;
+#endif
+	    break;
+	case 'K':	/* kerberos bind, part one only */
+#ifdef LDAP_API_FEATURE_X_OPENLDAP_V2_KBIND
+		if( version > LDAP_VERSION2 ) {
+			fprintf( stderr, "%s: -k incompatible with LDAPv%d\n",
+				prog, version );
+			return EXIT_FAILURE;
+		}
+		if( authmethod != -1 ) {
+			fprintf( stderr, "%s: incompatible with previous "
+				"authentication choice\n", prog );
+			return EXIT_FAILURE;
+		}
+
+		authmethod = LDAP_AUTH_KRBV41;
+#else
+		fprintf( stderr, "%s: not compiled with Kerberos support\n", prog );
+		return( EXIT_FAILURE );
+#endif
+	    break;
+	case 'M':
+		/* enable Manage DSA IT */
+		if( version == LDAP_VERSION2 ) {
+			fprintf( stderr, "%s: -M incompatible with LDAPv%d\n",
+				prog, version );
+			return EXIT_FAILURE;
+		}
+		manageDSAit++;
+		version = LDAP_VERSION3;
+		break;
+	case 'n':	/* print deletes, don't actually do them */
+	    ++not;
+	    break;
+	case 'O':
+#ifdef HAVE_CYRUS_SASL
+		if( sasl_secprops != NULL ) {
+			fprintf( stderr, "%s: -O previously specified\n", prog );
+			return EXIT_FAILURE;
+		}
+		if( version == LDAP_VERSION2 ) {
+			fprintf( stderr, "%s: -O incompatible with LDAPv%d\n",
+				prog, version );
+			return EXIT_FAILURE;
+		}
+		if( authmethod != -1 && authmethod != LDAP_AUTH_SASL ) {
+			fprintf( stderr, "%s: incompatible previous "
+				"authentication choice\n", prog );
+			return EXIT_FAILURE;
+		}
+		authmethod = LDAP_AUTH_SASL;
+		version = LDAP_VERSION3;
+		sasl_secprops = strdup( optarg );
+#else
+		fprintf( stderr, "%s: not compiled with SASL support\n",
+			prog );
+		return( EXIT_FAILURE );
+#endif
+		break;
+	case 'p':
+		if( ldapport ) {
+			fprintf( stderr, "%s: -p previously specified\n", prog );
+			return EXIT_FAILURE;
+		}
+	    ldapport = atoi( optarg );
+	    break;
+	case 'P':
+		switch( atoi(optarg) ) {
+		case 2:
+			if( version == LDAP_VERSION3 ) {
+				fprintf( stderr, "%s: -P 2 incompatible with version %d\n",
+					prog, version );
+				return EXIT_FAILURE;
+			}
+			version = LDAP_VERSION2;
+			break;
+		case 3:
+			if( version == LDAP_VERSION2 ) {
+				fprintf( stderr, "%s: -P 2 incompatible with version %d\n",
+					prog, version );
+				return EXIT_FAILURE;
+			}
+			version = LDAP_VERSION3;
+			break;
+		default:
+			fprintf( stderr, "%s: protocol version should be 2 or 3\n",
+				prog );
+			usage( prog );
+			return( EXIT_FAILURE );
+		} break;
+	case 'Q':
+#ifdef HAVE_CYRUS_SASL
+		if( version == LDAP_VERSION2 ) {
+			fprintf( stderr, "%s: -Q incompatible with version %d\n",
+				prog, version );
+			return EXIT_FAILURE;
+		}
+		if( authmethod != -1 && authmethod != LDAP_AUTH_SASL ) {
+			fprintf( stderr, "%s: incompatible previous "
+				"authentication choice\n",
+				prog );
+			return EXIT_FAILURE;
+		}
+		authmethod = LDAP_AUTH_SASL;
+		version = LDAP_VERSION3;
+		sasl_flags = LDAP_SASL_QUIET;
+		break;
+#else
+		fprintf( stderr, "%s: not compiled with SASL support\n",
+			prog );
+		return( EXIT_FAILURE );
+#endif
+	case 'r':	/* replace (obsolete) */
+		break;
+
+	case 'R':
+#ifdef HAVE_CYRUS_SASL
+		if( sasl_realm != NULL ) {
+			fprintf( stderr, "%s: -R previously specified\n", prog );
+			return EXIT_FAILURE;
+		}
+		if( version == LDAP_VERSION2 ) {
+			fprintf( stderr, "%s: -R incompatible with version %d\n",
+				prog, version );
+			return EXIT_FAILURE;
+		}
+		if( authmethod != -1 && authmethod != LDAP_AUTH_SASL ) {
+			fprintf( stderr, "%s: incompatible previous "
+				"authentication choice\n",
+				prog );
+			return EXIT_FAILURE;
+		}
+		authmethod = LDAP_AUTH_SASL;
+		version = LDAP_VERSION3;
+		sasl_realm = strdup( optarg );
+#else
+		fprintf( stderr, "%s: not compiled with SASL support\n",
+			prog );
+		return( EXIT_FAILURE );
+#endif
+		break;
+	case 'U':
+#ifdef HAVE_CYRUS_SASL
+		if( sasl_authc_id != NULL ) {
+			fprintf( stderr, "%s: -U previously specified\n", prog );
+			return EXIT_FAILURE;
+		}
+		if( version == LDAP_VERSION2 ) {
+			fprintf( stderr, "%s: -U incompatible with version %d\n",
+				prog, version );
+			return EXIT_FAILURE;
+		}
+		if( authmethod != -1 && authmethod != LDAP_AUTH_SASL ) {
+			fprintf( stderr, "%s: incompatible previous "
+				"authentication choice\n",
+				prog );
+			return EXIT_FAILURE;
+		}
+		authmethod = LDAP_AUTH_SASL;
+		version = LDAP_VERSION3;
+		sasl_authc_id = strdup( optarg );
+#else
+		fprintf( stderr, "%s: not compiled with SASL support\n",
+			prog );
+		return( EXIT_FAILURE );
+#endif
+		break;
+	case 'v':	/* verbose mode */
+	    verbose++;
+	    break;
 	case 'w':	/* password */
 	    passwd.bv_val = strdup( optarg );
 		{
 			char* p;
 
-			for( p = optarg; *p == '\0'; p++ ) {
-				*p = '*';
+			for( p = optarg; *p != '\0'; p++ ) {
+				*p = '\0';
 			}
 		}
 		passwd.bv_len = strlen( passwd.bv_val );
 	    break;
-	case 'd':
-	    debug |= atoi( optarg );
-	    break;
-	case 'f':	/* read from file */
-	    infile = strdup( optarg );
-	    break;
-	case 'p':
-	    ldapport = atoi( optarg );
-	    break;
-	case 'n':	/* print adds, don't actually do them */
-	    ++not;
-	    break;
-	case 'v':	/* verbose mode */
-	    verbose++;
-	    break;
-	case 'M':
-		/* enable Manage DSA IT */
-		manageDSAit++;
-		break;
 	case 'W':
 		want_bindpw++;
 		break;
-	case 'P':
-		switch( atoi(optarg) )
-		{
-		case 2:
-			version = LDAP_VERSION2;
-			break;
-		case 3:
-			version = LDAP_VERSION3;
-			break;
-		default:
-			fprintf( stderr, "protocol version should be 2 or 3\n" );
-			usage( argv[0] );
-		}
-		break;
-	case 'I':
-#ifdef HAVE_CYRUS_SASL
-		sasl_integrity++;
-		authmethod = LDAP_AUTH_SASL;
-#else
-		fprintf( stderr, "%s was not compiled with SASL support\n",
-			argv[0] );
-		return( EXIT_FAILURE );
-#endif
-		break;
-	case 'E':
-#ifdef HAVE_CYRUS_SASL
-		sasl_privacy++;
-		authmethod = LDAP_AUTH_SASL;
-#else
-		fprintf( stderr, "%s was not compiled with SASL support\n",
-			argv[0] );
-		return( EXIT_FAILURE );
-#endif
-		break;
 	case 'Y':
 #ifdef HAVE_CYRUS_SASL
-		if ( strcasecmp( optarg, "any" ) && strcmp( optarg, "*" ) ) {
-			sasl_mech = strdup( optarg );
+		if( sasl_mech != NULL ) {
+			fprintf( stderr, "%s: -Y previously specified\n", prog );
+			return EXIT_FAILURE;
+		}
+		if( version == LDAP_VERSION2 ) {
+			fprintf( stderr, "%s: -Y incompatible with version %d\n",
+				prog, version );
+			return EXIT_FAILURE;
+		}
+		if( authmethod != -1 && authmethod != LDAP_AUTH_SASL ) {
+			fprintf( stderr, "%s: incompatible with authentication choice\n", prog );
+			return EXIT_FAILURE;
 		}
 		authmethod = LDAP_AUTH_SASL;
+		version = LDAP_VERSION3;
+		sasl_mech = strdup( optarg );
 #else
-		fprintf( stderr, "%s was not compiled with SASL support\n",
-			argv[0] );
+		fprintf( stderr, "%s: not compiled with SASL support\n",
+			prog );
 		return( EXIT_FAILURE );
 #endif
 		break;
-	case 'U':
-#ifdef HAVE_CYRUS_SASL
-		sasl_authc_id = strdup( optarg );
-		authmethod = LDAP_AUTH_SASL;
-#else
-		fprintf( stderr, "%s was not compiled with SASL support\n",
-			argv[0] );
-		return( EXIT_FAILURE );
-#endif
+	case 'x':
+		if( authmethod != -1 && authmethod != LDAP_AUTH_SIMPLE ) {
+			fprintf( stderr, "%s: incompatible with previous "
+				"authentication choice\n", prog );
+			return EXIT_FAILURE;
+		}
+		authmethod = LDAP_AUTH_SIMPLE;
 		break;
 	case 'X':
 #ifdef HAVE_CYRUS_SASL
-		sasl_authz_id = strdup( optarg );
+		if( sasl_authz_id != NULL ) {
+			fprintf( stderr, "%s: -X previously specified\n", prog );
+			return EXIT_FAILURE;
+		}
+		if( version == LDAP_VERSION2 ) {
+			fprintf( stderr, "%s: -X incompatible with LDAPv%d\n",
+				prog, version );
+			return EXIT_FAILURE;
+		}
+		if( authmethod != -1 && authmethod != LDAP_AUTH_SASL ) {
+			fprintf( stderr, "%s: -X incompatible with "
+				"authentication choice\n", prog );
+			return EXIT_FAILURE;
+		}
 		authmethod = LDAP_AUTH_SASL;
+		version = LDAP_VERSION3;
+		sasl_authz_id = strdup( optarg );
 #else
-		fprintf( stderr, "%s was not compiled with SASL support\n",
-			argv[0] );
+		fprintf( stderr, "%s: not compiled with SASL support\n",
+			prog );
 		return( EXIT_FAILURE );
 #endif
 		break;
 	case 'Z':
 #ifdef HAVE_TLS
+		if( version == LDAP_VERSION2 ) {
+			fprintf( stderr, "%s: -Z incompatible with version %d\n",
+				prog, version );
+			return EXIT_FAILURE;
+		}
+		version = LDAP_VERSION3;
 		use_tls++;
 #else
-		fprintf( stderr, "%s was not compiled with TLS support\n",
-			argv[0] );
+		fprintf( stderr, "%s: not compiled with TLS support\n",
+			prog );
 		return( EXIT_FAILURE );
 #endif
 		break;
 	default:
+		fprintf( stderr, "%s: unrecognized option -%c\n",
+			prog, optopt );
 	    usage( prog );
 	}
     }
 
-    if ( argc != optind )
+	if (version == -1) {
+		version = LDAP_VERSION3;
+	}
+	if (authmethod == -1 && version > LDAP_VERSION2) {
+#ifdef HAVE_CYRUS_SASL
+		authmethod = LDAP_AUTH_SASL;
+#else
+		authmethod = LDAP_AUTH_SIMPLE;
+#endif
+	}
+
+	if ( argc != optind )
 	usage( prog );
 
-	if ( ( authmethod == LDAP_AUTH_KRBV4 ) || ( authmethod ==
-			LDAP_AUTH_KRBV41 ) ) {
-		if( version > LDAP_VERSION2 ) {
-			fprintf( stderr, "Kerberos requires LDAPv2\n" );
-			return( EXIT_FAILURE );
-		}
-		version = LDAP_VERSION2;
-	}
-	else if ( authmethod == LDAP_AUTH_SASL ) {
-		if( version != -1 && version != LDAP_VERSION3 ) {
-			fprintf( stderr, "SASL requires LDAPv3\n" );
-			return( EXIT_FAILURE );
-		}
-		version = LDAP_VERSION3;
-	}
-
-	if( manageDSAit ) {
-		if( version != -1 && version != LDAP_VERSION3 ) {
-			fprintf(stderr, "manage DSA control requires LDAPv3\n");
-			return EXIT_FAILURE;
-		}
-		version = LDAP_VERSION3;
-	}
-
-	if( use_tls ) {
-		if( version != -1 && version != LDAP_VERSION3 ) {
-			fprintf(stderr, "Start TLS requires LDAPv3\n");
-			return EXIT_FAILURE;
-		}
-		version = LDAP_VERSION3;
-	}
-
     if ( infile != NULL ) {
 	if (( fp = fopen( infile, "r" )) == NULL ) {
 	    perror( infile );
@@ -361,10 +562,36 @@ main( int argc, char **argv )
 	(void) SIGNAL( SIGPIPE, SIG_IGN );
 #endif
 
+#ifdef NEW_LOGGING
+    lutil_log_initialize( argc, argv );
+#endif
+
     if ( !not ) {
-	if (( ld = ldap_init( ldaphost, ldapport )) == NULL ) {
-	    perror( "ldap_init" );
-	    return( EXIT_FAILURE );
+	if( ( ldaphost != NULL || ldapport ) && ( ldapuri == NULL ) ) {
+		if ( verbose ) {
+			fprintf( stderr, "ldap_init( %s, %d )\n",
+				ldaphost != NULL ? ldaphost : "<DEFAULT>",
+				ldapport );
+		}
+
+		ld = ldap_init( ldaphost, ldapport );
+		if( ld == NULL ) {
+			perror("ldapmodify: ldap_init");
+			return EXIT_FAILURE;
+		}
+
+	} else {
+		if ( verbose ) {
+			fprintf( stderr, "ldap_initialize( %s )\n",
+				ldapuri != NULL ? ldapuri : "<DEFAULT>" );
+		}
+
+		rc = ldap_initialize( &ld, ldapuri );
+		if( rc != LDAP_SUCCESS ) {
+			fprintf( stderr, "Could not create LDAP session handle (%d): %s\n",
+				rc, ldap_err2string(rc) );
+			return EXIT_FAILURE;
+		}
 	}
 
 	/* referrals */
@@ -378,7 +605,7 @@ main( int argc, char **argv )
 
 
 	if (version == -1 ) {
-		version = 3;
+		version = LDAP_VERSION3;
 	}
 
 	if( ldap_set_option( ld, LDAP_OPT_PROTOCOL_VERSION, &version )
@@ -389,12 +616,11 @@ main( int argc, char **argv )
 		return EXIT_FAILURE;
 	}
 
-	if ( use_tls && ldap_start_tls_s( ld, NULL, NULL ) != LDAP_SUCCESS ) {
+	if ( use_tls && ( ldap_start_tls_s( ld, NULL, NULL ) != LDAP_SUCCESS )) {
+		ldap_perror( ld, "ldap_start_tls" );
 		if ( use_tls > 1 ) {
-			ldap_perror( ld, "ldap_start_tls" );
 			return( EXIT_FAILURE );
 		}
-		fprintf( stderr, "WARNING: could not start TLS\n" );
 	}
 
 	if (want_bindpw) {
@@ -404,42 +630,38 @@ main( int argc, char **argv )
 
 	if ( authmethod == LDAP_AUTH_SASL ) {
 #ifdef HAVE_CYRUS_SASL
-		int	minssf = 0, maxssf = 0;
+		void *defaults;
 
-		if ( sasl_integrity > 0 )
-			maxssf = 1;
-		if ( sasl_integrity > 1 )
-			minssf = 1;
-		if ( sasl_privacy > 0 )
-			maxssf = 100000; /* Something big value */
-		if ( sasl_privacy > 1 )
-			minssf = 56;
-		
-		if ( ldap_set_option( ld, LDAP_OPT_X_SASL_MINSSF,
-			(void *)&minssf ) != LDAP_OPT_SUCCESS ) {
-			fprintf( stderr, "Could not set LDAP_OPT_X_SASL_MINSSF"
-				"%d\n", minssf);
-			return( EXIT_FAILURE );
-		}
-		if ( ldap_set_option( ld, LDAP_OPT_X_SASL_MAXSSF,
-			(void *)&maxssf ) != LDAP_OPT_SUCCESS ) {
-			fprintf( stderr, "Could not set LDAP_OPT_X_SASL_MINSSF"
-				"%d\n", minssf);
-			return( EXIT_FAILURE );
+		if( sasl_secprops != NULL ) {
+			rc = ldap_set_option( ld, LDAP_OPT_X_SASL_SECPROPS,
+				(void *) sasl_secprops );
+			
+			if( rc != LDAP_OPT_SUCCESS ) {
+				fprintf( stderr,
+					"Could not set LDAP_OPT_X_SASL_SECPROPS: %s\n",
+					sasl_secprops );
+				return( EXIT_FAILURE );
+			}
 		}
 		
-		rc = ldap_negotiated_sasl_bind_s( ld, binddn, sasl_authc_id,
-				sasl_authz_id, sasl_mech,
-				passwd.bv_len ? &passwd : NULL,
-				NULL, NULL );
+		defaults = lutil_sasl_defaults( ld,
+			sasl_mech,
+			sasl_realm,
+			sasl_authc_id,
+			passwd.bv_val,
+			sasl_authz_id );
+
+		rc = ldap_sasl_interactive_bind_s( ld, binddn,
+			sasl_mech, NULL, NULL,
+			sasl_flags, lutil_sasl_interact, defaults );
 
 		if( rc != LDAP_SUCCESS ) {
-			ldap_perror( ld, "ldap_negotiated_sasl_bind_s" );
+			ldap_perror( ld, "ldap_sasl_interactive_bind_s" );
 			return( EXIT_FAILURE );
 		}
 #else
-		fprintf( stderr, "%s was not compiled with SASL support\n",
-			argv[0] );
+		fprintf( stderr, "%s: not compiled with SASL support\n",
+			prog );
 		return( EXIT_FAILURE );
 #endif
 	}
@@ -467,7 +689,7 @@ main( int argc, char **argv )
 		c.ldctl_value.bv_len = 0;
 		c.ldctl_iscritical = manageDSAit > 1;
 
-		err = ldap_set_option( ld, LDAP_OPT_SERVER_CONTROLS, &ctrls );
+		err = ldap_set_option( ld, LDAP_OPT_SERVER_CONTROLS, ctrls );
 
 		if( err != LDAP_OPT_SUCCESS ) {
 			fprintf( stderr, "Could not set ManageDSAit %scontrol\n",
@@ -585,6 +807,23 @@ process_ldif_rec( char *rbuf, int count )
 	    }
 
 	    if ( strcasecmp( type, T_CHANGETYPESTR ) == 0 ) {
+#ifdef LIBERAL_CHANGETYPE_MODOP
+		/* trim trailing spaces (and log warning ...) */
+
+		int icnt;
+		for ( icnt = val.bv_len; --icnt > 0; ) {
+		    if ( !isspace( (unsigned char) val.bv_val[icnt] ) ) {
+			break;
+		    }
+		}
+
+		if ( ++icnt != val.bv_len ) {
+		    fprintf( stderr, "%s: illegal trailing space after \"%s: %s\" trimmed (line %d of entry \"%s\")\n",
+			    prog, T_CHANGETYPESTR, val.bv_val, linenum, dn );
+		    val.bv_val[icnt] = '\0';
+		}
+#endif /* LIBERAL_CHANGETYPE_MODOP */
+
 		if ( strcasecmp( val.bv_val, T_MODIFYCTSTR ) == 0 ) {
 			new_entry = 0;
 			expect_modop = 1;
@@ -613,6 +852,23 @@ process_ldif_rec( char *rbuf, int count )
 	}
 
 	if ( expect_modop ) {
+#ifdef LIBERAL_CHANGETYPE_MODOP
+	    /* trim trailing spaces (and log warning ...) */
+	    
+	    int icnt;
+	    for ( icnt = val.bv_len; --icnt > 0; ) {
+		if ( !isspace( (unsigned char) val.bv_val[icnt] ) ) {
+		    break;
+		}
+	    }
+	    
+	    if ( ++icnt != val.bv_len ) {
+		fprintf( stderr, "%s: illegal trailing space after \"%s: %s\" trimmed (line %d of entry \"%s\")\n",
+    			prog, type, val.bv_val, linenum, dn );
+		val.bv_val[icnt] = '\0';
+	    }
+#endif /* LIBERAL_CHANGETYPE_MODOP */
+
 	    expect_modop = 0;
 	    expect_sep = 1;
 	    if ( strcasecmp( type, T_MODOPADDSTR ) == 0 ) {
@@ -627,7 +883,7 @@ process_ldif_rec( char *rbuf, int count )
 		addmodifyop( &pmods, modop, val.bv_val, NULL );
 		goto end_line;
 	    } else {	/* no modify op:  use default */
-		modop = replace ? LDAP_MOD_REPLACE : LDAP_MOD_ADD;
+		modop = ldapadd ? LDAP_MOD_ADD : LDAP_MOD_REPLACE;
 	    }
 	}
 
@@ -804,6 +1060,16 @@ domodify(
 	fprintf( stderr, "%s: no attributes to change or add (entry=\"%s\")\n",
 		prog, dn );
 	return( LDAP_PARAM_ERROR );
+    } 
+
+    for ( i = 0; pmods[ i ] != NULL; ++i ) {
+	op = pmods[ i ]->mod_op & ~LDAP_MOD_BVALUES;
+	if( op == LDAP_MOD_ADD && ( pmods[i]->mod_bvalues == NULL )) {
+		fprintf( stderr,
+			"%s: attribute \"%s\" has no values (entry=\"%s\")\n",
+			prog, pmods[i]->mod_type, dn );
+		return LDAP_PARAM_ERROR;
+	}
     }
 
     if ( verbose ) {
@@ -845,7 +1111,9 @@ domodify(
 	    i = ldap_modify_s( ld, dn, pmods );
 	}
 	if ( i != LDAP_SUCCESS ) {
-	    ldap_perror( ld, newentry ? "ldap_add" : "ldap_modify" );
+		/* print error message about failed update including DN */
+		fprintf( stderr, "%s: update failed: %s\n", prog, dn );
+		ldap_perror( ld, newentry ? "ldap_add" : "ldap_modify" );
 	} else if ( verbose ) {
 	    printf( "modify complete\n" );
 	}
@@ -868,7 +1136,8 @@ dodelete(
     printf( "%sdeleting entry \"%s\"\n", not ? "!" : "", dn );
     if ( !not ) {
 	if (( rc = ldap_delete_s( ld, dn )) != LDAP_SUCCESS ) {
-	    ldap_perror( ld, "ldap_delete" );
+		fprintf( stderr, "%s: delete failed: %s\n", prog, dn );
+		ldap_perror( ld, "ldap_delete" );
 	} else if ( verbose ) {
 	    printf( "delete complete" );
 	}
@@ -899,8 +1168,10 @@ dorename(
     }
     if ( !not ) {
 	if (( rc = ldap_rename2_s( ld, dn, newrdn, newsup, deleteoldrdn ))
-		!= LDAP_SUCCESS ) {
-	    ldap_perror( ld, "ldap_modrdn" );
+		!= LDAP_SUCCESS )
+	{
+		fprintf( stderr, "%s: rename failed: %s\n", prog, dn );
+		ldap_perror( ld, "ldap_modrdn" );
 	} else {
 	    printf( "modrdn completed\n" );
 	}
@@ -926,7 +1197,7 @@ read_one_record( FILE *fp )
     while ( fgets( line, sizeof(line), fp ) != NULL ) {
     	int len = strlen( line );
 
-		if( len < 2 ) {
+		if( len < 2 || ( len == 3 && *line == '\r' )) {
 			if( buf == NULL ) {
 				continue;
 			} else {
