@@ -1,12 +1,13 @@
-/* $OpenLDAP: pkg/ldap/include/ldap.h,v 1.65.2.7 2000/07/06 17:05:57 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/include/ldap.h,v 1.168.2.3 2002/04/18 15:39:20 kurt Exp $ */
 /*
- * Copyright 1998-2000 The OpenLDAP Foundation, Redwood City, California, USA
+ * Copyright 1998-2002 The OpenLDAP Foundation, Redwood City, California, USA
  * All rights reserved.
  *
- * Redistribution and use in source and binary forms are permitted only
- * as authorized by the OpenLDAP Public License.  A copy of this
- * license is available at http://www.OpenLDAP.org/license.html or
- * in file LICENSE in the top-level directory of the distribution.
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted only as authorized by the OpenLDAP
+ * Public License.  A copy of this license is available at
+ * http://www.OpenLDAP.org/license.html or in file LICENSE in the
+ * top-level directory of the distribution.
  */
 /* Portions
  * Copyright (c) 1990 Regents of the University of Michigan.
@@ -26,6 +27,9 @@
 /* pull in lber */
 #include <lber.h>
 
+/* include version and API feature defines */
+#include <ldap_features.h>
+
 LDAP_BEGIN_DECL
 
 #define LDAP_VERSION1	1
@@ -38,20 +42,15 @@ LDAP_BEGIN_DECL
 
 /*
  * We'll use 2000+draft revision for our API version number
- * As such, the number will be above the old RFC but below 
+ * As such, the number will be above the old RFC but below
  * whatever number does finally get assigned
  */
 #define LDAP_API_VERSION	2004
 #define LDAP_VENDOR_NAME	"OpenLDAP"
-/* We'll eventually release as 20000 */
-#define LDAP_VENDOR_VERSION	19910
 
 /* OpenLDAP API Features */
 #define LDAP_API_FEATURE_X_OPENLDAP LDAP_VENDOR_VERSION
 
-/* include LDAP_API_FEATURE defines */
-#include <ldap_features.h>
-
 #if defined( LDAP_API_FEATURE_X_OPENLDAP_REENTRANT ) || \
 	( defined( LDAP_THREAD_SAFE ) && \
 		defined( LDAP_API_FEATURE_X_OPENLDAP_THREAD_SAFE ) )
@@ -113,15 +112,17 @@ LDAP_BEGIN_DECL
 #define LDAP_OPT_URI				0x5006
 
 /* OpenLDAP TLS options */
-#define LDAP_OPT_X_TLS_CACERTFILE	0x6001
-#define LDAP_OPT_X_TLS_CACERTDIR	0x6002
-#define LDAP_OPT_X_TLS_CERT		0x6003
+#define LDAP_OPT_X_TLS				0x6000
+#define LDAP_OPT_X_TLS_CTX			0x6001	/* SSL CTX */
+#define LDAP_OPT_X_TLS_CACERTFILE	0x6002
+#define LDAP_OPT_X_TLS_CACERTDIR	0x6003
 #define LDAP_OPT_X_TLS_CERTFILE		0x6004
 #define LDAP_OPT_X_TLS_KEYFILE		0x6005
 #define LDAP_OPT_X_TLS_REQUIRE_CERT	0x6006
-#define LDAP_OPT_X_TLS			0x6007
-#define LDAP_OPT_X_TLS_PROTOCOL		0x6008
-#define LDAP_OPT_X_TLS_CIPHER_SUITE	0x6009
+/* #define LDAP_OPT_X_TLS_PROTOCOL		0x6007 */
+#define LDAP_OPT_X_TLS_CIPHER_SUITE	0x6008
+#define LDAP_OPT_X_TLS_RANDOM_FILE	0x6009
+#define LDAP_OPT_X_TLS_SSL_CTX		0x600a
 
 #define LDAP_OPT_X_TLS_NEVER		0
 #define LDAP_OPT_X_TLS_HARD		1
@@ -130,9 +131,17 @@ LDAP_BEGIN_DECL
 #define LDAP_OPT_X_TLS_TRY		4
 
 /* OpenLDAP SASL options */
-#define LDAP_OPT_X_SASL_MINSSF		0x6100
-#define LDAP_OPT_X_SASL_MAXSSF		0x6101
-#define LDAP_OPT_X_SASL_ACTSSF		0x6102
+#define LDAP_OPT_X_SASL_MECH			0x6100
+#define LDAP_OPT_X_SASL_REALM			0x6101
+#define LDAP_OPT_X_SASL_AUTHCID			0x6102
+#define LDAP_OPT_X_SASL_AUTHZID			0x6103
+#define LDAP_OPT_X_SASL_SSF				0x6104 /* read-only */
+#define LDAP_OPT_X_SASL_SSF_EXTERNAL	0x6105 /* write-only */
+#define LDAP_OPT_X_SASL_SECPROPS		0x6106 /* write-only */
+#define LDAP_OPT_X_SASL_SSF_MIN			0x6107
+#define LDAP_OPT_X_SASL_SSF_MAX			0x6108
+#define	LDAP_OPT_X_SASL_MAXBUFSIZE		0x6109
+
 
 /* on/off values */
 #define LDAP_OPT_ON		((void *) 1)
@@ -171,14 +180,8 @@ typedef struct ldapcontrol {
 } LDAPControl;
 
 /* LDAP Controls */
-	/* chase referrals controls */
-#define LDAP_CONTROL_REFERRALS	"1.2.840.113666.1.4.616"
-#define LDAP_CHASE_SUBORDINATE_REFERRALS	0x0020U
-#define LDAP_CHASE_EXTERNAL_REFERRALS	0x0040U
 
-#define LDAP_CONTROL_MANAGEDSAIT "2.16.840.1.113730.3.4.2"
-
-/* Experimental Controls */
+#define LDAP_CONTROL_MANAGEDSAIT	"2.16.840.1.113730.3.4.2"
 
 #define LDAP_CONTROL_SORTREQUEST    "1.2.840.113556.1.4.473"
 #define LDAP_CONTROL_SORTRESPONSE	"1.2.840.113556.1.4.474"
@@ -190,15 +193,17 @@ typedef struct ldapcontrol {
 #define LDAP_NOTICE_DISCONNECT LDAP_NOTICE_OF_DISCONNECTION
 
 /* LDAP Extended Operations */
-#define LDAP_EXOP_START_TLS "1.3.6.1.4.1.1466.20037"
+#define LDAP_EXOP_START_TLS	"1.3.6.1.4.1.1466.20037"
 
-#define LDAP_EXOP_X_MODIFY_PASSWD "1.3.6.1.4.1.4203.666.6.1"
-#define LDAP_TAG_EXOP_X_MODIFY_PASSWD_ID	((ber_tag_t) 0x80U)
-#define LDAP_TAG_EXOP_X_MODIFY_PASSWD_OLD	((ber_tag_t) 0x81U)
-#define LDAP_TAG_EXOP_X_MODIFY_PASSWD_NEW	((ber_tag_t) 0x82U)
-#define LDAP_TAG_EXOP_X_MODIFY_PASSWD_GEN	((ber_tag_t) 0x80U)
+#define LDAP_EXOP_MODIFY_PASSWD	"1.3.6.1.4.1.4203.1.11.1"
+#define LDAP_TAG_EXOP_MODIFY_PASSWD_ID	((ber_tag_t) 0x80U)
+#define LDAP_TAG_EXOP_MODIFY_PASSWD_OLD	((ber_tag_t) 0x81U)
+#define LDAP_TAG_EXOP_MODIFY_PASSWD_NEW	((ber_tag_t) 0x82U)
+#define LDAP_TAG_EXOP_MODIFY_PASSWD_GEN	((ber_tag_t) 0x80U)
 
-/* 
+#define LDAP_EXOP_X_WHO_AM_I	"1.3.6.1.4.1.4203.1.11.3"
+
+/*
  * specific LDAP instantiations of BER types we know about
  */
 
@@ -222,8 +227,8 @@ typedef struct ldapcontrol {
 /* general stuff */
 #define LDAP_TAG_MESSAGE	((ber_tag_t) 0x30U)	/* constructed + 16 */
 #define LDAP_TAG_MSGID		((ber_tag_t) 0x02U)	/* integer */
-#define LDAP_TAG_LDAPDN		((ber_tag_t) 0x04U)	/* octect string */
-#define LDAP_TAG_LDAPCRED	((ber_tag_t) 0x04U)	/* octect string */
+#define LDAP_TAG_LDAPDN		((ber_tag_t) 0x04U)	/* octet string */
+#define LDAP_TAG_LDAPCRED	((ber_tag_t) 0x04U)	/* octet string */
 #define LDAP_TAG_CONTROLS	((ber_tag_t) 0xa0U)	/* context specific + constructed + 0 */
 #define LDAP_TAG_REFERRAL	((ber_tag_t) 0xa3U)	/* context specific + constructed + 3 */
 
@@ -244,9 +249,9 @@ typedef struct ldapcontrol {
 #define LDAP_REQ_MODIFY			((ber_tag_t) 0x66U)	/* application + constructed */
 #define LDAP_REQ_ADD			((ber_tag_t) 0x68U)	/* application + constructed */
 #define LDAP_REQ_DELETE			((ber_tag_t) 0x4aU)	/* application + primitive   */
-#define LDAP_REQ_MODRDN			((ber_tag_t) 0x6cU)	/* application + constructed */
-#define LDAP_REQ_MODDN			LDAP_REQ_MODRDN	
-#define LDAP_REQ_RENAME			LDAP_REQ_MODRDN	
+#define LDAP_REQ_MODDN			((ber_tag_t) 0x6cU)	/* application + constructed */
+#define LDAP_REQ_MODRDN			LDAP_REQ_MODDN
+#define LDAP_REQ_RENAME			LDAP_REQ_MODDN
 #define LDAP_REQ_COMPARE		((ber_tag_t) 0x6eU)	/* application + constructed */
 #define LDAP_REQ_ABANDON		((ber_tag_t) 0x50U)	/* application + primitive   */
 #define LDAP_REQ_EXTENDED		((ber_tag_t) 0x77U)	/* application + constructed */
@@ -259,9 +264,9 @@ typedef struct ldapcontrol {
 #define LDAP_RES_MODIFY			((ber_tag_t) 0x67U)	/* application + constructed */
 #define LDAP_RES_ADD			((ber_tag_t) 0x69U)	/* application + constructed */
 #define LDAP_RES_DELETE			((ber_tag_t) 0x6bU)	/* application + constructed */
-#define LDAP_RES_MODRDN			((ber_tag_t) 0x6dU)	/* application + constructed */
-#define LDAP_RES_MODDN			LDAP_RES_MODRDN	/* application + constructed */
-#define LDAP_RES_RENAME			LDAP_RES_MODRDN	/* application + constructed */
+#define LDAP_RES_MODDN			((ber_tag_t) 0x6dU)	/* application + constructed */
+#define LDAP_RES_MODRDN			LDAP_RES_MODDN	/* application + constructed */
+#define LDAP_RES_RENAME			LDAP_RES_MODDN	/* application + constructed */
 #define LDAP_RES_COMPARE		((ber_tag_t) 0x6fU)	/* application + constructed */
 #define LDAP_RES_EXTENDED		((ber_tag_t) 0x78U)	/* V3: application + constructed */
 #define LDAP_RES_EXTENDED_PARTIAL	((ber_tag_t) 0x79U)	/* V3+: application + constructed */
@@ -271,12 +276,14 @@ typedef struct ldapcontrol {
 
 
 /* sasl methods */
-#define LDAP_SASL_SIMPLE			NULL
+#define LDAP_SASL_SIMPLE		((char*)0)
+#define LDAP_SASL_NULL			("")
+
 
 /* authentication methods available */
 #define LDAP_AUTH_NONE		((ber_tag_t) 0x00U)	/* no authentication		  */
 #define LDAP_AUTH_SIMPLE	((ber_tag_t) 0x80U)	/* context specific + primitive   */
-#define LDAP_AUTH_SASL		((ber_tag_t) 0xa3U)	/* context specific + primitive   */
+#define LDAP_AUTH_SASL		((ber_tag_t) 0xa3U)	/* context specific + constructed */
 #define LDAP_AUTH_KRBV4		((ber_tag_t) 0xffU)	/* means do both of the following */
 #define LDAP_AUTH_KRBV41	((ber_tag_t) 0x81U)	/* context specific + primitive   */
 #define LDAP_AUTH_KRBV42	((ber_tag_t) 0x82U)	/* context specific + primitive   */
@@ -306,6 +313,7 @@ typedef struct ldapcontrol {
 #define LDAP_SUBSTRING_FINAL	((ber_tag_t) 0x82U)	/* context specific */
 
 /* search scopes */
+#define LDAP_SCOPE_DEFAULT	((ber_int_t) -1)
 #define LDAP_SCOPE_BASE		((ber_int_t) 0x0000)
 #define LDAP_SCOPE_ONELEVEL	((ber_int_t) 0x0001)
 #define LDAP_SCOPE_SUBTREE	((ber_int_t) 0x0002)
@@ -315,7 +323,7 @@ typedef struct ldapcontrol {
 #define LDAP_SUBSTRING_ANY	((ber_tag_t) 0x81U)	/* context specific */
 #define LDAP_SUBSTRING_FINAL	((ber_tag_t) 0x82U)	/* context specific */
 
-/* 
+/*
  * possible error codes we can return
  */
 
@@ -331,13 +339,13 @@ typedef struct ldapcontrol {
 #define LDAP_AUTH_METHOD_NOT_SUPPORTED	0x07
 #define LDAP_STRONG_AUTH_NOT_SUPPORTED	LDAP_AUTH_METHOD_NOT_SUPPORTED
 #define LDAP_STRONG_AUTH_REQUIRED	0x08
-#define LDAP_PARTIAL_RESULTS		0x09	/* not listed in v3 */
+#define LDAP_PARTIAL_RESULTS		0x09	/* LDAPv2+ (not LDAPv3) */
 
 #define	LDAP_REFERRAL				0x0a /* LDAPv3 */
 #define LDAP_ADMINLIMIT_EXCEEDED	0x0b /* LDAPv3 */
 #define	LDAP_UNAVAILABLE_CRITICAL_EXTENSION	0x0c /* LDAPv3 */
 #define LDAP_CONFIDENTIALITY_REQUIRED	0x0d /* LDAPv3 */
-#define	LDAP_SASL_BIND_IN_PROGRESS	0x0e /* LDAPv3 */	
+#define	LDAP_SASL_BIND_IN_PROGRESS	0x0e /* LDAPv3 */
 
 #define LDAP_ATTR_ERROR(n)	LDAP_RANGE((n),0x10,0x15) /* 16-21 */
 
@@ -383,7 +391,10 @@ typedef struct ldapcontrol {
 #define LDAP_OTHER			0x50
 
 #define LDAP_API_ERROR(n)		LDAP_RANGE((n),0x51,0x61) /* 81-97 */
+#define LDAP_API_RESULT(n)		(((n) == LDAP_SUCCESS) || \
+									LDAP_RANGE((n),0x51,0x61)) /* 0,81-97 */
 
+/* reserved for APIs */
 #define LDAP_SERVER_DOWN		0x51
 #define LDAP_LOCAL_ERROR		0x52
 #define LDAP_ENCODING_ERROR		0x53
@@ -395,7 +406,7 @@ typedef struct ldapcontrol {
 #define LDAP_PARAM_ERROR		0x59
 #define LDAP_NO_MEMORY			0x5a
 
-/* not technically reserved for APIs */
+/* used but not reserved for APIs */
 #define LDAP_CONNECT_ERROR				0x5b	/* draft-ietf-ldap-c-api-xx */
 #define LDAP_NOT_SUPPORTED				0x5c	/* draft-ietf-ldap-c-api-xx */
 #define LDAP_CONTROL_NOT_FOUND			0x5d	/* draft-ietf-ldap-c-api-xx */
@@ -489,28 +500,20 @@ typedef struct ldap LDAP;
 #define LDAP_MSG_ALL		0x01
 #define LDAP_MSG_RECEIVED	0x02
 
-/*
- * structure for ldap friendly mapping routines
- */
-
-typedef struct ldap_friendly {
-	char	*lf_unfriendly;
-	char	*lf_friendly;
-} LDAPFriendlyMap;
-
 /*
  * types for ldap URL handling
  */
 typedef struct ldap_url_desc {
-    struct ldap_url_desc *lud_next;
-    char	*lud_scheme;
-    char	*lud_host;
-    int		lud_port;
-    char	*lud_dn;
-    char	**lud_attrs;
-    int		lud_scope;
-    char	*lud_filter;
-    char	**lud_exts;
+	struct ldap_url_desc *lud_next;
+	char	*lud_scheme;
+	char	*lud_host;
+	int		lud_port;
+	char	*lud_dn;
+	char	**lud_attrs;
+	int		lud_scope;
+	char	*lud_filter;
+	char	**lud_exts;
+	int		lud_crit_exts;
 } LDAPURLDesc;
 
 #define LDAP_URL_SUCCESS		0x00	/* Success */
@@ -549,16 +552,26 @@ ldap_set_option LDAP_P((
 
 /* V3 REBIND Function Callback Prototype */
 typedef int (LDAP_REBIND_PROC) LDAP_P((
-	LDAP *ld, LDAP_CONST char *url, int request, ber_int_t msgid ));
+	LDAP *ld, LDAP_CONST char *url,
+	ber_tag_t request, ber_int_t msgid,
+	void *params ));
 
 LDAP_F( int )
 ldap_set_rebind_proc LDAP_P((
 	LDAP *ld,
-	LDAP_REBIND_PROC *ldap_proc));
+	LDAP_REBIND_PROC *rebind_proc,
+	void *params ));
 
 /*
  * in controls.c:
  */
+LDAP_F( int )
+ldap_create_control LDAP_P((
+	LDAP_CONST char *requestOID,
+	BerElement *ber,
+	int iscritical,
+	LDAPControl **ctrlp ));
+
 LDAP_F( void )
 ldap_control_free LDAP_P((
 	LDAPControl *ctrl ));
@@ -686,16 +699,35 @@ ldap_sasl_bind LDAP_P((
 	LDAPControl		**clientctrls,
 	int				*msgidp ));
 
+/* Interaction flags (should be passed about in a control)
+ *  Automatic (default): use defaults, prompt otherwise
+ *  Interactive: prompt always
+ *  Quiet: never prompt
+ */
+#define LDAP_SASL_AUTOMATIC		0U
+#define LDAP_SASL_INTERACTIVE	1U
+#define LDAP_SASL_QUIET			2U
+
+/*
+ * V3 SASL Interaction Function Callback Prototype
+ *	when using Cyrus SASL, interact is pointer to sasl_interact_t
+ *  should likely passed in a control (and provided controls)
+ */
+typedef int (LDAP_SASL_INTERACT_PROC) LDAP_P((
+	LDAP *ld, unsigned flags, void* defaults, void *interact ));
+
 LDAP_F( int )
-ldap_negotiated_sasl_bind_s LDAP_P((
+ldap_sasl_interactive_bind_s LDAP_P((
 	LDAP *ld,
 	LDAP_CONST char *dn, /* usually NULL */
-	LDAP_CONST char *authenticationId, 
-	LDAP_CONST char *authorizationId, /* usually NULL */
 	LDAP_CONST char *saslMechanism,
-	struct berval *passPhrase,
 	LDAPControl **serverControls,
-	LDAPControl **clientControls ));
+	LDAPControl **clientControls,
+
+	/* should be client controls */
+	unsigned flags,
+	LDAP_SASL_INTERACT_PROC *proc,
+	void *defaults ));
 
 LDAP_F( int )
 ldap_sasl_bind_s LDAP_P((
@@ -883,7 +915,7 @@ ldap_parse_result LDAP_P((
 	LDAPControl		***serverctrls,
 	int				freeit ));
 
-LDAP_F( char *)
+LDAP_F( char * )
 ldap_err2string LDAP_P((
 	int err ));
 
@@ -1002,18 +1034,18 @@ ldap_modrdn2_s LDAP_P((	/* deprecated */
 /*
  * in open.c:
  */
-LDAP_F( LDAP *)
+LDAP_F( LDAP * )
 ldap_init LDAP_P((
 	LDAP_CONST char *host,
 	int port ));
 
-LDAP_F( LDAP *)
+LDAP_F( LDAP * )
 ldap_open LDAP_P((	/* deprecated */
 	LDAP_CONST char *host,
 	int port ));
 
 LDAP_F( int )
-ldap_create LDAP_P(( 
+ldap_create LDAP_P((
 	LDAP **ldp ));
 
 LDAP_F( int )
@@ -1030,12 +1062,12 @@ ldap_start_tls_s LDAP_P((
 /*
  * in messages.c:
  */
-LDAP_F( LDAPMessage *)
+LDAP_F( LDAPMessage * )
 ldap_first_message LDAP_P((
 	LDAP *ld,
 	LDAPMessage *chain ));
 
-LDAP_F( LDAPMessage *)
+LDAP_F( LDAPMessage * )
 ldap_next_message LDAP_P((
 	LDAP *ld,
 	LDAPMessage *msg ));
@@ -1049,12 +1081,12 @@ ldap_count_messages LDAP_P((
 /*
  * in references.c:
  */
-LDAP_F( LDAPMessage *)
+LDAP_F( LDAPMessage * )
 ldap_first_reference LDAP_P((
 	LDAP *ld,
 	LDAPMessage *chain ));
 
-LDAP_F( LDAPMessage *)
+LDAP_F( LDAPMessage * )
 ldap_next_reference LDAP_P((
 	LDAP *ld,
 	LDAPMessage *ref ));
@@ -1076,12 +1108,12 @@ ldap_parse_reference LDAP_P((
 /*
  * in getentry.c:
  */
-LDAP_F( LDAPMessage *)
+LDAP_F( LDAPMessage * )
 ldap_first_entry LDAP_P((
 	LDAP *ld,
 	LDAPMessage *chain ));
 
-LDAP_F( LDAPMessage *)
+LDAP_F( LDAPMessage * )
 ldap_next_entry LDAP_P((
 	LDAP *ld,
 	LDAPMessage *entry ));
@@ -1101,7 +1133,7 @@ ldap_get_entry_controls LDAP_P((
 /*
  * in addentry.c
  */
-LDAP_F( LDAPMessage *)
+LDAP_F( LDAPMessage * )
 ldap_delete_result_entry LDAP_P((
 	LDAPMessage **list,
 	LDAPMessage *e ));
@@ -1120,36 +1152,138 @@ ldap_get_dn LDAP_P((
 	LDAP *ld,
 	LDAPMessage *entry ));
 
+typedef struct ldap_ava {
+	struct berval la_attr;
+	struct berval la_value;
+	unsigned la_flags;
+#define LDAP_AVA_STRING		0x0000U
+#define LDAP_AVA_BINARY		0x0001U
+#define LDAP_AVA_NONPRINTABLE	0x0002U
+
+	void *la_private;
+} LDAPAVA;
+
+typedef LDAPAVA** LDAPRDN;
+typedef LDAPRDN** LDAPDN;
+
+/* DN formats */
+#define LDAP_DN_FORMAT_LDAP		0x0000U
+#define LDAP_DN_FORMAT_LDAPV3		0x0010U
+#define LDAP_DN_FORMAT_LDAPV2		0x0020U
+#define LDAP_DN_FORMAT_DCE		0x0030U
+#define LDAP_DN_FORMAT_UFN		0x0040U	/* dn2str only */
+#define LDAP_DN_FORMAT_AD_CANONICAL	0x0050U	/* dn2str only */
+#define LDAP_DN_FORMAT_LBER		0x00F0U /* for testing only */
+#define LDAP_DN_FORMAT_MASK		0x00F0U
+
+/* DN flags */
+#define LDAP_DN_PRETTY			0x0100U
+#define LDAP_DN_SKIP			0x0200U
+#define LDAP_DN_P_NOLEADTRAILSPACES	0x1000U
+#define LDAP_DN_P_NOSPACEAFTERRDN	0x2000U
+#define LDAP_DN_PEDANTIC		0xF000U
+
+LDAP_F( void )
+ldap_avafree LDAP_P(( LDAPAVA *ava ));
+LDAP_F( void )
+ldap_rdnfree LDAP_P(( LDAPRDN *rdn ));
+LDAP_F( void )
+ldap_dnfree LDAP_P(( LDAPDN *dn ));
+
+LDAP_F( int )
+ldap_bv2dn LDAP_P(( 
+	struct berval *bv, 
+	LDAPDN **dn, 
+	unsigned flags ));
+
+LDAP_F( int )
+ldap_str2dn LDAP_P((
+	LDAP_CONST char *str,
+	LDAPDN **dn,
+	unsigned flags ));
+
+LDAP_F( int )
+ldap_dn2bv LDAP_P((
+	LDAPDN *dn,
+	struct berval *bv,
+	unsigned flags ));
+
+LDAP_F( int )
+ldap_dn2str LDAP_P((
+	LDAPDN *dn,
+	char **str,
+	unsigned flags ));
+
+LDAP_F( int )
+ldap_bv2rdn LDAP_P((
+	struct berval *bv,
+	LDAPRDN **rdn,
+	char **next,
+	unsigned flags ));
+
+LDAP_F( int )
+ldap_str2rdn LDAP_P((
+	LDAP_CONST char *str,
+	LDAPRDN **rdn,
+	char **next,
+	unsigned flags ));
+
+LDAP_F( int )
+ldap_rdn2bv LDAP_P((
+	LDAPRDN *rdn,
+	struct berval *bv,
+	unsigned flags ));
+
+LDAP_F( int )
+ldap_rdn2str LDAP_P((
+	LDAPRDN *rdn,
+	char **str,
+	unsigned flags ));
+
+LDAP_F( int )
+ldap_dn_normalize LDAP_P((
+	LDAP_CONST char *in, unsigned iflags,
+	char **out, unsigned oflags ));
+
 LDAP_F( char * )
-ldap_dn2ufn LDAP_P((
+ldap_dn2ufn LDAP_P(( /* deprecated */
 	LDAP_CONST char *dn ));
 
 LDAP_F( char ** )
-ldap_explode_dn LDAP_P((
+ldap_explode_dn LDAP_P(( /* deprecated */
 	LDAP_CONST char *dn,
 	int notypes ));
 
 LDAP_F( char ** )
-ldap_explode_rdn LDAP_P((
+ldap_explode_rdn LDAP_P(( /* deprecated */
 	LDAP_CONST char *rdn,
 	int notypes ));
 
+typedef int LDAPDN_rewrite_func LDAP_P(( LDAPDN *dn, unsigned flags ));
+
+LDAP_F( int )
+ldap_X509dn2bv LDAP_P(( void *x509_name, struct berval *dn,
+	LDAPDN_rewrite_func *func, unsigned flags ));
+
 LDAP_F( char * )
 ldap_dn2dcedn LDAP_P(( LDAP_CONST char *dn ));	/* deprecated */
 
 LDAP_F( char * )
 ldap_dcedn2dn LDAP_P(( LDAP_CONST char *dce ));	/* deprecated */
 
+LDAP_F( char * )
+ldap_dn2ad_canonical LDAP_P(( LDAP_CONST char *dn ));	/* deprecated */
+
 /*
  * in getattr.c
  */
-LDAP_F( char *)
-ldap_first_attribute LDAP_P((									 
+LDAP_F( char * )
+ldap_first_attribute LDAP_P((
 	LDAP *ld,
 	LDAPMessage *entry,
 	BerElement **ber ));
 
-LDAP_F( char *)
+LDAP_F( char * )
 ldap_next_attribute LDAP_P((
 	LDAP *ld,
 	LDAPMessage *entry,
@@ -1159,33 +1293,33 @@ ldap_next_attribute LDAP_P((
 /*
  * in getvalues.c
  */
-LDAP_F( char **)
-ldap_get_values LDAP_P((
-	LDAP *ld,
-	LDAPMessage *entry,
-	LDAP_CONST char *target ));
-
-LDAP_F( struct berval **)
+LDAP_F( struct berval ** )
 ldap_get_values_len LDAP_P((
 	LDAP *ld,
 	LDAPMessage *entry,
 	LDAP_CONST char *target ));
 
-LDAP_F( int )
-ldap_count_values LDAP_P((
-	char **vals ));
-
 LDAP_F( int )
 ldap_count_values_len LDAP_P((
 	struct berval **vals ));
 
 LDAP_F( void )
-ldap_value_free LDAP_P((
+ldap_value_free_len LDAP_P((
+	struct berval **vals ));
+
+LDAP_F( char ** )
+ldap_get_values LDAP_P((	/* deprecated */
+	LDAP *ld,
+	LDAPMessage *entry,
+	LDAP_CONST char *target ));
+
+LDAP_F( int )
+ldap_count_values LDAP_P((	/* deprecated */
 	char **vals ));
 
 LDAP_F( void )
-ldap_value_free_len LDAP_P((
-	struct berval **vals ));
+ldap_value_free LDAP_P((	/* deprecated */
+	char **vals ));
 
 /*
  * in result.c:
@@ -1277,57 +1411,6 @@ ldap_search_st LDAP_P((	/* deprecated */
 	struct timeval *timeout,
 	LDAPMessage **res ));
 
-
-/*
- * in ufn.c						 	
- *	(deprecated)
- */
-LDAP_F( int )
-ldap_ufn_search_c LDAP_P(( /* deprecated */
-	LDAP *ld,
-	LDAP_CONST char *ufn,
-	char **attrs,
-	int attrsonly,
-	LDAPMessage **res,
-	int (*cancelproc)( void *cl ),
-	void *cancelparm ));
-
-LDAP_F( int )
-ldap_ufn_search_ct LDAP_P(( /* deprecated */
-	LDAP *ld,
-	LDAP_CONST char *ufn,
-	char **attrs,
-	int attrsonly,
-	LDAPMessage **res,
-	int (*cancelproc)( void *cl ),
-	void *cancelparm,
-	char *tag1,
-	char *tag2,
-	char *tag3 ));
-
-LDAP_F( int )
-ldap_ufn_search_s LDAP_P(( /* deprecated */
-	LDAP *ld,
-	LDAP_CONST char *ufn,
-	char **attrs,
-	int attrsonly,
-	LDAPMessage **res ));
-
-LDAP_F( LDAPFiltDesc *)
-ldap_ufn_setfilter LDAP_P(( /* deprecated */
-	LDAP *ld,
-	LDAP_CONST char *fname ));
-
-LDAP_F( void )
-ldap_ufn_setprefix LDAP_P(( /* deprecated */
-	LDAP *ld,
-	LDAP_CONST char *prefix ));
-
-LDAP_F( int )
-ldap_ufn_timeout LDAP_P(( /* deprecated */
-	void *tvparam ));
-
-
 /*
  * in unbind.c
  */
@@ -1355,42 +1438,20 @@ ldap_unbind_ext_s LDAP_P((
  * in getfilter.c
  *	(deprecated)
  */
-LDAP_F( LDAPFiltDesc *)
+LDAP_F( LDAPFiltDesc * )
 ldap_init_getfilter LDAP_P(( /* deprecated */
 	LDAP_CONST char *fname ));
 
-LDAP_F( LDAPFiltDesc *)
-ldap_init_getfilter_buf LDAP_P(( /* deprecated */
-	/* LDAP_CONST */ char *buf,
-	ber_len_t buflen ));
-
-LDAP_F( LDAPFiltInfo *)
+LDAP_F( LDAPFiltInfo * )
 ldap_getfirstfilter LDAP_P(( /* deprecated */
 	LDAPFiltDesc *lfdp,
 	/* LDAP_CONST */ char *tagpat,
 	/* LDAP_CONST */ char *value ));
 
-LDAP_F( LDAPFiltInfo *)
+LDAP_F( LDAPFiltInfo * )
 ldap_getnextfilter LDAP_P(( /* deprecated */
 	LDAPFiltDesc *lfdp ));
 
-LDAP_F( void )
-ldap_setfilteraffixes LDAP_P(( /* deprecated */
-	LDAPFiltDesc *lfdp,
-	LDAP_CONST char *prefix,
-	LDAP_CONST char *suffix ));
-
-LDAP_F( void )
-ldap_build_filter LDAP_P(( /* deprecated */
-	char *buf,
-	ber_len_t buflen,
-	LDAP_CONST char *pattern,
-	LDAP_CONST char *prefix,
-	LDAP_CONST char *suffix,
-	LDAP_CONST char *attr,
-	LDAP_CONST char *value,
-	char **valwords ));
-
 
 /*
  * in free.c
@@ -1432,64 +1493,19 @@ ldap_mods_free LDAP_P((
 	int freemods ));
 
 
-/*
- * in friendly.c
- *	(deprecated)
- */
-LDAP_F( char * )
-ldap_friendly_name LDAP_P(( /* deprecated */
-	LDAP_CONST char *filename,
-	/* LDAP_CONST */ char *uname,
-	LDAPFriendlyMap **map ));
-
-LDAP_F( void )
-ldap_free_friendlymap LDAP_P(( /* deprecated */
-	LDAPFriendlyMap **map ));
-
-
-/*
- * in cldap.c
- *	(deprecated)
- */
-LDAP_F( LDAP * )
-cldap_open LDAP_P((	/* deprecated */
-	LDAP_CONST char *host,
-	int port ));
-
-LDAP_F( void )
-cldap_close LDAP_P((	/* deprecated */
-	LDAP *ld ));
-
-LDAP_F( int )
-cldap_search_s LDAP_P(( /* deprecated */
-	LDAP *ld,
-	LDAP_CONST char *base,
-	int scope,
-	LDAP_CONST char *filter,
-	char **attrs,
-	int attrsonly,
-	LDAPMessage **res,
-	char *logdn ));
-
-LDAP_F( void )
-cldap_setretryinfo LDAP_P(( /* deprecated */
-	LDAP *ld,
-	int tries,
-	int timeout ));
-
-
 /*
  * in sort.c
+ *	(deprecated)
  */
-typedef int (LDAP_SORT_AD_CMP_PROC) LDAP_P((
+typedef int (LDAP_SORT_AD_CMP_PROC) LDAP_P(( /* deprecated */
 	LDAP_CONST char *left,
 	LDAP_CONST char *right ));
 
-typedef int (LDAP_SORT_AV_CMP_PROC) LDAP_P((
+typedef int (LDAP_SORT_AV_CMP_PROC) LDAP_P(( /* deprecated */
 	LDAP_CONST void *left,
 	LDAP_CONST void *right ));
 
-LDAP_F( int )
+LDAP_F( int )	/* deprecated */
 ldap_sort_entries LDAP_P(( LDAP *ld,
 	LDAPMessage **chain,
 	LDAP_CONST char *attr,
@@ -1509,8 +1525,6 @@ ldap_sort_strcasecmp LDAP_P((
 
 /*
  * in url.c
- *
- * need _ext varients
  */
 LDAP_F( int )
 ldap_is_ldap_url LDAP_P((
@@ -1529,36 +1543,19 @@ ldap_url_parse LDAP_P((
 	LDAP_CONST char *url,
 	LDAPURLDesc **ludpp ));
 
+LDAP_F( char * )
+ldap_url_desc2str LDAP_P((
+	LDAPURLDesc *ludp ));
+
 LDAP_F( void )
 ldap_free_urldesc LDAP_P((
 	LDAPURLDesc *ludp ));
 
-LDAP_F( int )
-ldap_url_search LDAP_P((
-	LDAP *ld,
-	LDAP_CONST char *url,
-	int attrsonly ));
-
-LDAP_F( int )
-ldap_url_search_s LDAP_P((
-	LDAP *ld,
-	LDAP_CONST char *url,
-	int attrsonly,
-	LDAPMessage **res ));
-
-LDAP_F( int )
-ldap_url_search_st LDAP_P((
-	LDAP *ld,
-	LDAP_CONST char *url,
-	int attrsonly,
-	struct timeval *timeout,
-	LDAPMessage **res ));
-
-/* 
- * in sortctrl.c  
+/*
+ * in sortctrl.c
  */
 /*
- * structure for a sort-key 
+ * structure for a sort-key
  */
 typedef struct ldapsortkey {
 	char *  attributeType;
@@ -1576,22 +1573,22 @@ ldap_free_sort_keylist LDAP_P((
 	LDAPSortKey **sortkeylist ));
 
 LDAP_F( int )
-ldap_create_sort_control LDAP_P((	
-	LDAP *ld, 
+ldap_create_sort_control LDAP_P((
+	LDAP *ld,
 	LDAPSortKey **keyList,
 	int ctl_iscritical,
 	LDAPControl **ctrlp ));
 
 LDAP_F( int )
 ldap_parse_sort_control LDAP_P((
-	LDAP           *ld, 
-	LDAPControl    **ctrlp,  
+	LDAP           *ld,
+	LDAPControl    **ctrlp,
 	unsigned long  *result,
 	char           **attribute ));
 
 
-/* 
- * in vlvctrl.c  
+/*
+ * in vlvctrl.c
  */
 
 /*
@@ -1599,27 +1596,27 @@ ldap_parse_sort_control LDAP_P((
  */
 typedef struct ldapvlvinfo {
 	int             ldvlv_version;
-    unsigned long   ldvlv_before_count;      
-    unsigned long   ldvlv_after_count;                     
-    unsigned long   ldvlv_offset;              
+    unsigned long   ldvlv_before_count;
+    unsigned long   ldvlv_after_count;
+    unsigned long   ldvlv_offset;
     unsigned long   ldvlv_count;
     struct berval  *ldvlv_attrvalue;
     struct berval  *ldvlv_context;
     void           *ldvlv_extradata;
 } LDAPVLVInfo;
 
-LDAP_F( int ) 
+LDAP_F( int )
 ldap_create_vlv_control LDAP_P((
-	LDAP *ld, 
+	LDAP *ld,
 	LDAPVLVInfo *ldvlistp,
 	LDAPControl **ctrlp ));
 
 LDAP_F( int )
-ldap_parse_vlv_control LDAP_P(( 
-	LDAP          *ld, 
+ldap_parse_vlv_control LDAP_P((
+	LDAP          *ld,
 	LDAPControl   **ctrls,
-	unsigned long *target_posp, 
-	unsigned long *list_countp, 
+	unsigned long *target_posp,
+	unsigned long *list_countp,
 	struct berval **contextp,
 	int           *errcodep ));
 
