@@ -1,7 +1,7 @@
 /* io.c - ber general i/o routines */
-/* $OpenLDAP: pkg/ldap/libraries/liblber/io.c,v 1.38.4.3 2000/06/17 16:16:56 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/libraries/liblber/io.c,v 1.70.2.1 2002/04/15 15:21:45 kurt Exp $ */
 /*
- * Copyright 1998-2000 The OpenLDAP Foundation, All Rights Reserved.
+ * Copyright 1998-2002 The OpenLDAP Foundation, All Rights Reserved.
  * COPYING RESTRICTIONS APPLY, see COPYRIGHT file
  */
 /* Portions
@@ -34,49 +34,6 @@
 
 #include "lber-int.h"
 
-static ber_slen_t BerRead LDAP_P((
-	Sockbuf *sb,
-	char *buf,
-	ber_len_t len ));
-
-#define EXBUFSIZ	1024
-
-/* probably far too large... */
-#define MAX_BERBUFSIZE	(128*1024)
-
-#if defined( DOS ) && !defined( _WIN32 ) && (MAX_BERBUFSIZE > 65535)
-# undef MAX_BERBUFSIZE
-# define MAX_BERBUFSIZE 65535
-#endif
-
-static ber_slen_t
-BerRead(
-	Sockbuf *sb,
-	char *buf,
-	ber_len_t len )
-{
-	ber_slen_t	c;
-	ber_slen_t	nread = 0;
-
-	assert( sb != NULL );
-	assert( buf != NULL );
-
-	assert( SOCKBUF_VALID( sb ) );
-
-	while ( len > 0 ) {
-		if ( (c = ber_int_sb_read( sb, buf, len )) <= 0 ) {
-			if ( nread > 0 )
-				break;
-			return( c );
-		}
-		buf+= c;
-		nread+=c;
-		len-=c;
-	}
-
-	return( nread );
-}
-
 ber_slen_t
 ber_read(
 	BerElement *ber,
@@ -88,12 +45,12 @@ ber_read(
 	assert( ber != NULL );
 	assert( buf != NULL );
 
-	assert( BER_VALID( ber ) );
+	assert( LBER_VALID( ber ) );
 
-	nleft = ber->ber_end - ber->ber_ptr;
+	nleft = ber_pvt_ber_remaining( ber );
 	actuallen = nleft < len ? nleft : len;
 
-	SAFEMEMCPY( buf, ber->ber_ptr, actuallen );
+	AC_MEMCPY( buf, ber->ber_ptr, actuallen );
 
 	ber->ber_ptr += actuallen;
 
@@ -110,14 +67,14 @@ ber_write(
 	assert( ber != NULL );
 	assert( buf != NULL );
 
-	assert( BER_VALID( ber ) );
+	assert( LBER_VALID( ber ) );
 
 	if ( nosos || ber->ber_sos == NULL ) {
 		if ( ber->ber_ptr + len > ber->ber_end ) {
 			if ( ber_realloc( ber, len ) != 0 )
 				return( -1 );
 		}
-		SAFEMEMCPY( ber->ber_ptr, buf, (size_t)len );
+		AC_MEMCPY( ber->ber_ptr, buf, (size_t)len );
 		ber->ber_ptr += len;
 		return( (ber_slen_t) len );
 
@@ -126,7 +83,7 @@ ber_write(
 			if ( ber_realloc( ber, len ) != 0 )
 				return( -1 );
 		}
-		SAFEMEMCPY( ber->ber_sos->sos_ptr, buf, (size_t)len );
+		AC_MEMCPY( ber->ber_sos->sos_ptr, buf, (size_t)len );
 		ber->ber_sos->sos_ptr += len;
 		ber->ber_sos->sos_clen += len;
 		return( (ber_slen_t) len );
@@ -136,23 +93,36 @@ ber_write(
 int
 ber_realloc( BerElement *ber, ber_len_t len )
 {
-	ber_len_t	need, have, total;
+	ber_len_t	total;
 	Seqorset	*s;
 	long		off;
 	char		*oldbuf;
 
 	assert( ber != NULL );
 	assert( len > 0 );
+	assert( LBER_VALID( ber ) );
 
-	assert( BER_VALID( ber ) );
+	total = ber_pvt_ber_total( ber );
 
-	have = (ber->ber_end - ber->ber_buf) / EXBUFSIZ;
-	need = (len < EXBUFSIZ ? 1 : (len + (EXBUFSIZ - 1)) / EXBUFSIZ);
-	total = have * EXBUFSIZ + need * EXBUFSIZ;
+#define LBER_EXBUFSIZ	1000 /* a few words less than 2^N for binary buddy */
+#if defined( LBER_EXBUFSIZ ) && LBER_EXBUFSIZ > 0
+# ifndef notdef
+	/* don't realloc by small amounts */
+	total += len < LBER_EXBUFSIZ ? LBER_EXBUFSIZ : len;
+# else
+	{	/* not sure what value this adds */
+		ber_len_t have = (total + (LBER_EXBUFSIZE - 1)) / LBER_EXBUFSIZ;
+		ber_len_t need = (len + (LBER_EXBUFSIZ - 1)) / LBER_EXBUFSIZ;
+		total = ( have + need ) * LBER_EXBUFSIZ;
+	}
+# endif
+#else
+	total += len;	/* realloc just what's needed */
+#endif
 
 	oldbuf = ber->ber_buf;
 
-	ber->ber_buf = (char *) LBER_REALLOC( ber->ber_buf, total );
+	ber->ber_buf = (char *) LBER_REALLOC( oldbuf, total );
 	
 	if ( ber->ber_buf == NULL ) {
 		ber->ber_buf = oldbuf;
@@ -163,7 +133,7 @@ ber_realloc( BerElement *ber, ber_len_t len )
 
 	/*
 	 * If the stinking thing was moved, we need to go through and
-	 * reset all the sos and ber pointers.  Offsets would've been
+	 * reset all the sos and ber pointers.	Offsets would've been
 	 * a better idea... oh well.
 	 */
 
@@ -182,6 +152,25 @@ ber_realloc( BerElement *ber, ber_len_t len )
 	return( 0 );
 }
 
+void
+ber_free_buf( BerElement *ber )
+{
+	Seqorset *s, *next;
+
+	assert( LBER_VALID( ber ) );
+
+	if ( ber->ber_buf) LBER_FREE( ber->ber_buf );
+
+	for( s = ber->ber_sos ; s != NULL ; s = next ) {
+		next = s->sos_next;
+		LBER_FREE( s );
+	}
+
+	ber->ber_buf = NULL;
+	ber->ber_sos = NULL;
+	ber->ber_valid = LBER_UNINITIALIZED;
+}
+
 void
 ber_free( BerElement *ber, int freebuf )
 {
@@ -193,21 +182,8 @@ ber_free( BerElement *ber, int freebuf )
 		return;
 	}
 
-	assert( BER_VALID( ber ) );
-
-	if ( freebuf ) {
-		Seqorset *s, *next;
-		LBER_FREE( ber->ber_buf );
-
-		for( s = ber->ber_sos ; s != NULL ; s = next ) {
-			next = s->sos_next;
-			LBER_FREE( s );
-		}
-	}
-
-	ber->ber_buf = NULL;
-	ber->ber_sos = NULL;
-	ber->ber_valid = LBER_UNINITIALIZED;
+	if( freebuf )
+		ber_free_buf( ber );
 
 	LBER_FREE( (char *) ber );
 }
@@ -222,7 +198,7 @@ ber_flush( Sockbuf *sb, BerElement *ber, int freeit )
 	assert( ber != NULL );
 
 	assert( SOCKBUF_VALID( sb ) );
-	assert( BER_VALID( ber ) );
+	assert( LBER_VALID( ber ) );
 
 	if ( ber->ber_rwptr == NULL ) {
 		ber->ber_rwptr = ber->ber_buf;
@@ -230,12 +206,20 @@ ber_flush( Sockbuf *sb, BerElement *ber, int freeit )
 	towrite = ber->ber_ptr - ber->ber_rwptr;
 
 	if ( sb->sb_debug ) {
+#ifdef NEW_LOGGING
+		LDAP_LOG(( "liblber", LDAP_LEVEL_DETAIL1,
+			   "ber_flush: %ld bytes to sd %ld%s\n",
+			   towrite, (long)sb->sb_fd,
+			   ber->ber_rwptr != ber->ber_buf ? " (re-flush)" : "" ));
+		BER_DUMP(( "liblber", LDAP_LEVEL_DETAIL2, ber, 1 ));
+#else
 		ber_log_printf( LDAP_DEBUG_ANY, sb->sb_debug,
 			"ber_flush: %ld bytes to sd %ld%s\n",
 			towrite, (long) sb->sb_fd,
 			ber->ber_rwptr != ber->ber_buf ?  " (re-flush)" : "" );
 		ber_log_bprint( LDAP_DEBUG_PACKETS, sb->sb_debug,
 			ber->ber_rwptr, towrite );
+#endif
 	}
 
 	nwritten = 0;
@@ -264,28 +248,29 @@ ber_alloc_t( int options )
 
 	ber = (BerElement *) LBER_CALLOC( 1, sizeof(BerElement) );
 
-	if ( ber == NULL )
-		return( NULL );
+	if ( ber == NULL ) {
+		return NULL;
+	}
 
 	ber->ber_valid = LBER_VALID_BERELEMENT;
 	ber->ber_tag = LBER_DEFAULT;
 	ber->ber_options = options;
 	ber->ber_debug = ber_int_debug;
 
-	assert( BER_VALID( ber ) );
-	return( ber );
+	assert( LBER_VALID( ber ) );
+	return ber;
 }
 
 BerElement *
 ber_alloc( void )	/* deprecated */
 {
-	return( ber_alloc_t( 0 ) );
+	return ber_alloc_t( 0 );
 }
 
 BerElement *
 der_alloc( void )	/* deprecated */
 {
-	return( ber_alloc_t( LBER_USE_DER ) );
+	return ber_alloc_t( LBER_USE_DER );
 }
 
 BerElement *
@@ -294,26 +279,25 @@ ber_dup( BerElement *ber )
 	BerElement	*new;
 
 	assert( ber != NULL );
-	assert( BER_VALID( ber ) );
+	assert( LBER_VALID( ber ) );
 
 	if ( (new = ber_alloc_t( ber->ber_options )) == NULL ) {
-		return( NULL );
+		return NULL;
 	}
 
 	*new = *ber;
 
-	assert( BER_VALID( new ) );
+	assert( LBER_VALID( new ) );
 	return( new );
 }
 
 
-/* OLD U-Mich ber_init() */
 void
-ber_init_w_nullc( BerElement *ber, int options )
+ber_init2( BerElement *ber, struct berval *bv, int options )
 {
 	assert( ber != NULL );
 
-    ber_int_options.lbo_valid = LBER_INITIALIZED;
+	ber_int_options.lbo_valid = LBER_INITIALIZED;
 
 	(void) memset( (char *)ber, '\0', sizeof( BerElement ));
 	ber->ber_valid = LBER_VALID_BERELEMENT;
@@ -321,7 +305,20 @@ ber_init_w_nullc( BerElement *ber, int options )
 	ber->ber_options = (char) options;
 	ber->ber_debug = ber_int_debug;
 
-	assert( BER_VALID( ber ) );
+	if ( bv != NULL ) {
+		ber->ber_buf = bv->bv_val;
+		ber->ber_ptr = ber->ber_buf;
+		ber->ber_end = ber->ber_buf + bv->bv_len;
+	}
+
+	assert( LBER_VALID( ber ) );
+}
+
+/* OLD U-Mich ber_init() */
+void
+ber_init_w_nullc( BerElement *ber, int options )
+{
+	ber_init2( ber, NULL, options );
 }
 
 /* New C-API ber_init() */
@@ -345,19 +342,20 @@ ber_init( struct berval *bv )
 
 	if( ber == NULL ) {
 		/* allocation failed */
-		return ( NULL );
+		return NULL;
 	}
 
 	/* copy the data */
-	if ( ( (ber_len_t) ber_write ( ber, bv->bv_val, bv->bv_len, 0 )) != bv->bv_len ) {
+	if ( ((ber_len_t) ber_write ( ber, bv->bv_val, bv->bv_len, 0 ))
+		!= bv->bv_len )
+	{
 		/* write failed, so free and return NULL */
 		ber_free( ber, 1 );
-		return( NULL );
+		return NULL;
 	}
 
 	ber_reset( ber, 1 );	/* reset the pointer to the start of the buffer */
-
-	return ( ber );
+	return ber;
 }
 
 /* New C-API ber_flatten routine */
@@ -376,11 +374,12 @@ int ber_flatten(
     ber_int_options.lbo_valid = LBER_INITIALIZED;
 
 	if(bvPtr == NULL) {
-		return( -1 );
+		return -1;
 	}
 
-	if ( (bv = LBER_MALLOC( sizeof(struct berval))) == NULL ) {
-		return( -1 );
+	bv = LBER_MALLOC( sizeof(struct berval) );
+	if ( bv == NULL ) {
+		return -1;
 	}
 
 	if ( ber == NULL ) {
@@ -390,31 +389,33 @@ int ber_flatten(
 
 	} else {
 		/* copy the berval */
-		ber_len_t len = ber->ber_ptr - ber->ber_buf;
+		ber_len_t len = ber_pvt_ber_write( ber );
 
-		if ( (bv->bv_val = (char *) LBER_MALLOC( len + 1 )) == NULL ) {
-			ber_bvfree( bv );
-			return( -1 );
+		bv->bv_val = (char *) LBER_MALLOC( len + 1 );
+		if ( bv->bv_val == NULL ) {
+			LBER_FREE( bv );
+			return -1;
 		}
 
-		SAFEMEMCPY( bv->bv_val, ber->ber_buf, len );
+		AC_MEMCPY( bv->bv_val, ber->ber_buf, len );
 		bv->bv_val[len] = '\0';
 		bv->bv_len = len;
 	}
     
 	*bvPtr = bv;
-	return( 0 );
+	return 0;
 }
 
 void
 ber_reset( BerElement *ber, int was_writing )
 {
 	assert( ber != NULL );
-	assert( BER_VALID( ber ) );
+	assert( LBER_VALID( ber ) );
 
 	if ( was_writing ) {
 		ber->ber_end = ber->ber_ptr;
 		ber->ber_ptr = ber->ber_buf;
+
 	} else {
 		ber->ber_ptr = ber->ber_end;
 	}
@@ -422,46 +423,6 @@ ber_reset( BerElement *ber, int was_writing )
 	ber->ber_rwptr = NULL;
 }
 
-#if 0
-/* return the tag - LBER_DEFAULT returned means trouble */
-static ber_tag_t
-get_tag( Sockbuf *sb )
-{
-	unsigned char	xbyte;
-	ber_tag_t	tag;
-	char		*tagp;
-	unsigned int	i;
-
-	assert( sb != NULL );
-	assert( SOCKBUF_VALID( sb ) );
-
-	if ( ber_int_sb_read( sb, (char *) &xbyte, 1 ) != 1 )
-		return( LBER_DEFAULT );
-
-	if ( (xbyte & LBER_BIG_TAG_MASK) != LBER_BIG_TAG_MASK )
-		return( (ber_tag_t) xbyte );
-
-	tagp = (char *) &tag;
-	tagp[0] = xbyte;
-	for ( i = 1; i < sizeof(ber_tag_t); i++ ) {
-		if ( ber_int_sb_read( sb, (char *) &xbyte, 1 ) != 1 )
-			return( LBER_DEFAULT );
-
-		tagp[i] = xbyte;
-
-		if ( ! (xbyte & LBER_MORE_TAG_MASK) )
-			break;
-	}
-
-	/* tag too big! */
-	if ( i == sizeof(ber_tag_t) )
-		return( LBER_DEFAULT );
-
-	/* want leading, not trailing 0's */
-	return( tag >> (sizeof(ber_tag_t) - i - 1) );
-}
-#endif
-
 /*
  * A rewrite of ber_get_next that can safely be called multiple times 
  * for the same packet. It will simply continue where it stopped until
@@ -479,10 +440,14 @@ ber_get_next(
 	assert( ber != NULL );
 
 	assert( SOCKBUF_VALID( sb ) );
-	assert( BER_VALID( ber ) );
+	assert( LBER_VALID( ber ) );
 
+#ifdef NEW_LOGGING
+	LDAP_LOG(( "liblber", LDAP_LEVEL_ENTRY, "ber_get_next: enter\n" ));
+#else
 	ber_log_printf( LDAP_DEBUG_TRACE, ber->ber_debug,
 		"ber_get_next\n" );
+#endif
 
 	/*
 	 * Any ber element looks like this: tag length contents.
@@ -495,105 +460,137 @@ ber_get_next(
 	 *	2) definite lengths
 	 *	3) primitive encodings used whenever possible
 	 */
-	
+
 	if (ber->ber_rwptr == NULL) {
 		/* XXYYZ
 		 * dtest does like this assert.
 		 */
 		/* assert( ber->ber_buf == NULL ); */
-		ber->ber_rwptr = (char *) &ber->ber_tag;
+		ber->ber_rwptr = (char *) &ber->ber_len-1;
+		ber->ber_ptr = ber->ber_rwptr;
 		ber->ber_tag = 0;
 	}
 
-#undef PTR_IN_VAR
-#define PTR_IN_VAR( ptr, var ) \
-	(((ptr)>=(char *) &(var)) && ((ptr)< (char *) &(var)+sizeof(var)))
-	
-	if (PTR_IN_VAR(ber->ber_rwptr, ber->ber_tag)) {
-		if (ber->ber_rwptr == (char *) &ber->ber_tag) {
-			if (ber_int_sb_read( sb, ber->ber_rwptr, 1)<=0)
-				return LBER_DEFAULT;
-			if ((ber->ber_rwptr[0] & LBER_BIG_TAG_MASK)
-				!= LBER_BIG_TAG_MASK) {
-				ber->ber_tag = ber->ber_rwptr[0];
-				ber->ber_rwptr = (char *) &ber->ber_usertag;
-				goto get_lenbyte;
-			}
-			ber->ber_rwptr++;
-		}
-		do {
-			/* reading the tag... */
-			if (ber_int_sb_read( sb, ber->ber_rwptr, 1)<=0) {
-				return LBER_DEFAULT;
-			}
+	while (ber->ber_rwptr > (char *)&ber->ber_tag && ber->ber_rwptr <
+		(char *)(&ber->ber_usertag + 1)) {
+		ber_slen_t i;
+		char buf[sizeof(ber->ber_len)-1];
+		ber_len_t tlen = 0;
 
-			if (! (ber->ber_rwptr[0] & LBER_MORE_TAG_MASK) ) {
-				ber->ber_tag>>=sizeof(ber->ber_tag) -
-				  ((char *) &ber->ber_tag - ber->ber_rwptr);
-				ber->ber_rwptr = (char *) &ber->ber_usertag;
-				goto get_lenbyte;
-			}
-		} while( PTR_IN_VAR(ber->ber_rwptr, ber->ber_tag ));
-
-		errno = ERANGE; /* this is a serious error. */
-		return LBER_DEFAULT;
-	}
-
-get_lenbyte:
-	if (ber->ber_rwptr==(char *) &ber->ber_usertag) {
-		unsigned char c;
-		if (ber_int_sb_read( sb, (char *) &c, 1)<=0)
+		if ((i=ber_int_sb_read( sb, ber->ber_rwptr,
+			(char *)(&ber->ber_usertag+1)-ber->ber_rwptr))<=0) {
 			return LBER_DEFAULT;
-		if (c & 0x80U) {
-			int len = c & 0x7fU;
-			if ( (len==0) || ( len>sizeof( ber->ber_len ) ) ) {
+		}
+
+		ber->ber_rwptr += i;
+
+		/* We got at least one byte, try to parse the tag. */
+		if (ber->ber_ptr == (char *)&ber->ber_len-1) {
+			ber_tag_t tag;
+			unsigned char *p = (unsigned char *)ber->ber_ptr;
+			tag = *p++;
+			if ((tag & LBER_BIG_TAG_MASK) == LBER_BIG_TAG_MASK) {
+				for (i=1; (char *)p<ber->ber_rwptr; i++,p++) {
+					tag <<= 8;
+					tag |= *p;
+					if (!(*p & LBER_MORE_TAG_MASK))
+						break;
+					/* Is the tag too big? */
+					if (i == sizeof(ber_tag_t)-1) {
+						errno = ERANGE;
+						return LBER_DEFAULT;
+					}
+				}
+				/* Did we run out of bytes? */
+				if ((char *)p == ber->ber_rwptr) {
+					return LBER_DEFAULT;
+				}
+				p++;
+			}
+			ber->ber_tag = tag;
+			ber->ber_ptr = (char *)p;
+		}
+
+		/* Now look for the length */
+		if (*ber->ber_ptr & 0x80) {	/* multi-byte */
+			int llen = *(unsigned char *)ber->ber_ptr++ & 0x7f;
+			if (llen > (int)sizeof(ber_len_t)) {
 				errno = ERANGE;
 				return LBER_DEFAULT;
 			}
-			ber->ber_rwptr = (char *) &ber->ber_len +
-				sizeof(ber->ber_len) - len;
-			ber->ber_len = 0;
+			/* Not enough bytes? */
+			if (ber->ber_rwptr - ber->ber_ptr < llen) {
+				return LBER_DEFAULT;
+			}
+			for (i=0; i<llen && ber->ber_ptr<ber->ber_rwptr; i++,ber->ber_ptr++) {
+				tlen <<=8;
+				tlen |= *(unsigned char *)ber->ber_ptr;
+			}
 		} else {
-			ber->ber_len = c;
-			goto fill_buffer;
+			tlen = *(unsigned char *)ber->ber_ptr++;
 		}
-	}
+		/* Are there leftover data bytes inside ber->ber_len? */
+		if (ber->ber_ptr < (char *)&ber->ber_usertag) {
+			if (ber->ber_rwptr < (char *)&ber->ber_usertag)
+				i = ber->ber_rwptr - ber->ber_ptr;
+			else
+				i = (char *)&ber->ber_usertag - ber->ber_ptr;
+			AC_MEMCPY(buf, ber->ber_ptr, i);
+			ber->ber_ptr += i;
+		} else {
+			i = 0;
+		}
+		ber->ber_len = tlen;
 
-	if (PTR_IN_VAR(ber->ber_rwptr, ber->ber_len)) {
-		unsigned char netlen[sizeof(ber_len_t)];
+		/* now fill the buffer. */
 
-		ber_slen_t res;
-		ber_slen_t to_go;
-		to_go = (char *) &ber->ber_len + sizeof( ber->ber_len ) -
-			ber->ber_rwptr;
-		assert( to_go > 0 );
-		res = BerRead( sb, netlen, to_go );
-		if (res <= 0) {
+		/* make sure length is reasonable */
+		if ( ber->ber_len == 0 ) {
+			errno = ERANGE;
 			return LBER_DEFAULT;
-		}
-		ber->ber_rwptr += res;
-
-		/* convert length. */
-		ber->ber_len = 0;
-		for( to_go = 0; to_go < res ; to_go++ ) {
-			ber->ber_len <<= 8;
-			ber->ber_len |= netlen[to_go];
-		}
-	}
-
-fill_buffer:	
-	/* now fill the buffer. */
-	if (ber->ber_buf==NULL) {
-		if (ber->ber_len > MAX_BERBUFSIZE) {
+		} else if ( sb->sb_max_incoming && ber->ber_len > sb->sb_max_incoming ) {
+#ifdef NEW_LOGGING
+			LDAP_LOG(( "liblber", LDAP_LEVEL_ERR, 
+				"ber_get_next: sockbuf_max_incoming limit hit "
+				"(%d > %d)\n", ber->ber_len, sb->sb_max_incoming ));
+#else
+			ber_log_printf( LDAP_DEBUG_CONNS, ber->ber_debug,
+				"ber_get_next: sockbuf_max_incoming limit hit "
+				"(%ld > %ld)\n", ber->ber_len, sb->sb_max_incoming );
+#endif
 			errno = ERANGE;
 			return LBER_DEFAULT;
 		}
-		ber->ber_buf = (char *) LBER_MALLOC( ber->ber_len );
-		if (ber->ber_buf==NULL)
-			return LBER_DEFAULT;
-		ber->ber_rwptr = ber->ber_buf;
-		ber->ber_ptr = ber->ber_buf;
-		ber->ber_end = ber->ber_buf + ber->ber_len;
+
+		if (ber->ber_buf==NULL) {
+			ber_len_t l = ber->ber_rwptr - ber->ber_ptr;
+			/* ber->ber_ptr is always <= ber->ber->ber_rwptr.
+			 * make sure ber->ber_len agrees with what we've
+			 * already read.
+			 */
+			if ( ber->ber_len < i + l ) {
+				errno = ERANGE;
+				return LBER_DEFAULT;
+			}
+			ber->ber_buf = (char *) LBER_MALLOC( ber->ber_len + 1 );
+			if (ber->ber_buf==NULL) {
+				return LBER_DEFAULT;
+			}
+			ber->ber_end = ber->ber_buf + ber->ber_len;
+			if (i) {
+				AC_MEMCPY(ber->ber_buf, buf, i);
+			}
+			if (l > 0) {
+				AC_MEMCPY(ber->ber_buf + i, ber->ber_ptr, l);
+				i += l;
+			}
+			ber->ber_ptr = ber->ber_buf;
+			ber->ber_usertag = 0;
+			if ((ber_len_t)i == ber->ber_len) {
+				goto done;
+			}
+			ber->ber_rwptr = ber->ber_buf + i;
+		}
 	}
 
 	if ((ber->ber_rwptr>=ber->ber_buf) && (ber->ber_rwptr<ber->ber_end)) {
@@ -616,17 +613,25 @@ fill_buffer:
 #endif			
 			return LBER_DEFAULT;
 		}
-		
+done:
 		ber->ber_rwptr = NULL;
 		*len = ber->ber_len;
 		if ( ber->ber_debug ) {
+#ifdef NEW_LOGGING
+			LDAP_LOG(( "liblber", LDAP_LEVEL_DETAIL1,
+				"ber_get_next: tag 0x%lx len %ld\n",
+				ber->ber_tag, ber->ber_len ));
+			BER_DUMP(( "liblber", LDAP_LEVEL_DETAIL2, ber, 1 ));
+#else
 			ber_log_printf( LDAP_DEBUG_TRACE, ber->ber_debug,
 				"ber_get_next: tag 0x%lx len %ld contents:\n",
 				ber->ber_tag, ber->ber_len );
 			ber_log_dump( LDAP_DEBUG_BER, ber->ber_debug, ber, 1 );
+#endif
 		}
 		return (ber->ber_tag);
 	}
+
 	assert( 0 ); /* ber structure is messed up ?*/
 	return LBER_DEFAULT;
 }
