@@ -1,6 +1,6 @@
-/* $OpenLDAP: pkg/ldap/libraries/libldap/os-ip.c,v 1.33.2.4 2000/06/17 16:16:57 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/libraries/libldap/os-ip.c,v 1.72.2.2 2002/04/15 15:21:47 kurt Exp $ */
 /*
- * Copyright 1998-2000 The OpenLDAP Foundation, All Rights Reserved.
+ * Copyright 1998-2002 The OpenLDAP Foundation, All Rights Reserved.
  * COPYING RESTRICTIONS APPLY, see COPYRIGHT file
  */
 /*  Portions
@@ -74,7 +74,7 @@ ldap_int_timeval_dup( struct timeval **dest, const struct timeval *src )
 		return 1;
 	}
 
-	SAFEMEMCPY( (char *) new, (const char *) src, sizeof(struct timeval));
+	AC_MEMCPY( (char *) new, (const char *) src, sizeof(struct timeval));
 
 	*dest = new;
 	return 0;
@@ -95,9 +95,9 @@ ldap_pvt_ndelay_off(LDAP *ld, int fd)
 }
 
 static ber_socket_t
-ldap_pvt_socket(LDAP *ld, int family)
+ldap_int_socket(LDAP *ld, int family, int type )
 {
-	ber_socket_t s = socket(family, SOCK_STREAM, 0);
+	ber_socket_t s = socket(family, type, 0);
 	osip_debug(ld, "ldap_new_socket: %d\n",s,0,0);
 	return ( s );
 }
@@ -110,25 +110,28 @@ ldap_pvt_close_socket(LDAP *ld, int s)
 }
 
 static int
-ldap_pvt_prepare_socket(LDAP *ld, int fd)
+ldap_int_prepare_socket(LDAP *ld, int s, int proto )
 {
-	osip_debug(ld, "ldap_prepare_socket: %d\n",fd,0,0);
+	osip_debug(ld, "ldap_prepare_socket: %d\n", s,0,0);
 
 #ifdef TCP_NODELAY
-{
-	int dummy = 1;
-	if ( setsockopt( fd, IPPROTO_TCP, TCP_NODELAY,
-		(char*) &dummy, sizeof(dummy) ) == AC_SOCKET_ERROR )
-	{
-		osip_debug(ld, "ldap_prepare_socket: "
-			"setsockopt(%d, TCP_NODELAY) failed (ignored).\n",
-			fd, 0, 0);
+	if( proto == LDAP_PROTO_TCP ) {
+		int dummy = 1;
+		if ( setsockopt( s, IPPROTO_TCP, TCP_NODELAY,
+			(char*) &dummy, sizeof(dummy) ) == AC_SOCKET_ERROR )
+		{
+			osip_debug(ld, "ldap_prepare_socket: "
+				"setsockopt(%d, TCP_NODELAY) failed (ignored).\n",
+				s, 0, 0);
+		}
 	}
-}
 #endif
+
 	return 0;
 }
 
+#ifndef HAVE_WINSOCK
+
 #undef TRACE
 #define TRACE do { \
 	osip_debug(ld, \
@@ -149,8 +152,10 @@ ldap_pvt_is_socket_ready(LDAP *ld, int s)
 #if defined( notyet ) /* && defined( SO_ERROR ) */
 {
 	int so_errno;
-	int dummy = sizeof(so_errno);
-	if ( getsockopt( s, SOL_SOCKET, SO_ERROR, &so_errno, &dummy ) == -1 ) {
+	socklen_t dummy = sizeof(so_errno);
+	if ( getsockopt( s, SOL_SOCKET, SO_ERROR, &so_errno, &dummy )
+		== AC_SOCKET_ERROR )
+	{
 		return -1;
 	}
 	if ( so_errno ) {
@@ -165,13 +170,12 @@ ldap_pvt_is_socket_ready(LDAP *ld, int s)
 	/* error slippery */
 	struct sockaddr_in sin;
 	char ch;
-	int dummy = sizeof(sin);
-	if ( getpeername( s, (struct sockaddr *) &sin, &dummy ) == -1 ) {
+	socklen_t dummy = sizeof(sin);
+	if ( getpeername( s, (struct sockaddr *) &sin, &dummy )
+		== AC_SOCKET_ERROR )
+	{
 		/* XXX: needs to be replace with ber_stream_read() */
 		read(s, &ch, 1);
-#ifdef HAVE_WINSOCK
-		ldap_pvt_set_errno( WSAGetLastError() );
-#endif
 		TRACE;
 		return -1;
 	}
@@ -182,8 +186,12 @@ ldap_pvt_is_socket_ready(LDAP *ld, int s)
 }
 #undef TRACE
 
+#endif /* HAVE_WINSOCK */
+
 static int
-ldap_pvt_connect(LDAP *ld, ber_socket_t s, struct sockaddr *sin, socklen_t addrlen, int async)
+ldap_pvt_connect(LDAP *ld, ber_socket_t s,
+	struct sockaddr *sin, socklen_t addrlen,
+	int async)
 {
 	struct timeval	tv, *opt_tv=NULL;
 	fd_set		wfds, *z=NULL;
@@ -191,6 +199,18 @@ ldap_pvt_connect(LDAP *ld, ber_socket_t s, struct sockaddr *sin, socklen_t addrl
 	fd_set		efds;
 #endif
 
+#ifdef LDAP_CONNECTIONLESS
+	/* We could do a connect() but that would interfere with
+	 * attempts to poll a broadcast address
+	 */
+	if (LDAP_IS_UDP(ld)) {
+		if (ld->ld_options.ldo_peer)
+			ldap_memfree(ld->ld_options.ldo_peer);
+		ld->ld_options.ldo_peer=ldap_memalloc(sizeof(struct sockaddr));
+		AC_MEMCPY(ld->ld_options.ldo_peer,sin,sizeof(struct sockaddr));
+		return ( 0 );
+	}
+#endif
 	if ( (opt_tv = ld->ld_options.ldo_tm_net) != NULL ) {
 		tv.tv_usec = opt_tv->tv_usec;
 		tv.tv_sec = opt_tv->tv_sec;
@@ -202,7 +222,7 @@ ldap_pvt_connect(LDAP *ld, ber_socket_t s, struct sockaddr *sin, socklen_t addrl
 	if ( ldap_pvt_ndelay_on(ld, s) == -1 )
 		return ( -1 );
 
-	if ( connect(s, sin, addrlen) == 0 )
+	if ( connect(s, sin, addrlen) != AC_SOCKET_ERROR )
 	{
 		if ( ldap_pvt_ndelay_off(ld, s) == -1 )
 			return ( -1 );
@@ -235,22 +255,33 @@ ldap_pvt_connect(LDAP *ld, ber_socket_t s, struct sockaddr *sin, socklen_t addrl
 #else
 		    z,
 #endif
-		    opt_tv ? &tv : NULL) == -1)
+		    opt_tv ? &tv : NULL) == AC_SOCKET_ERROR )
+	{
 		return ( -1 );
+	}
 
 #ifdef HAVE_WINSOCK
 	/* This means the connection failed */
-	if (FD_ISSET(s, &efds))
-	{
-	    ldap_pvt_set_errno(WSAECONNREFUSED);
+	if ( FD_ISSET(s, &efds) ) {
+	    int so_errno;
+	    int dummy = sizeof(so_errno);
+	    if ( getsockopt( s, SOL_SOCKET, SO_ERROR,
+			(char *) &so_errno, &dummy ) == AC_SOCKET_ERROR || !so_errno )
+	    {
+	    	/* impossible */
+	    	so_errno = WSAGetLastError();
+	    }
+	    ldap_pvt_set_errno(so_errno);
 	    osip_debug(ld, "ldap_pvt_connect: error on socket %d: "
 		       "errno: %d (%s)\n", s, errno, sock_errstr(errno));
 	    return -1;
 	}
 #endif
 	if ( FD_ISSET(s, &wfds) ) {
+#ifndef HAVE_WINSOCK
 		if ( ldap_pvt_is_socket_ready(ld, s) == -1 )
 			return ( -1 );
+#endif
 		if ( ldap_pvt_ndelay_off(ld, s) == -1 )
 			return ( -1 );
 		return ( 0 );
@@ -275,41 +306,69 @@ ldap_pvt_inet_aton( const char *host, struct in_addr *in)
 
 
 int
-ldap_connect_to_host(LDAP *ld, Sockbuf *sb, const char *host,
-		unsigned long address, int port, int async)
+ldap_connect_to_host(LDAP *ld, Sockbuf *sb,
+	int proto,
+	const char *host,
+	unsigned long address, int port, int async )
 {
 	struct sockaddr_in	sin;
-	struct in_addr		in;
 	ber_socket_t		s = AC_SOCKET_INVALID;
 	int			rc, i, use_hp = 0;
-	struct hostent		*hp, he_buf;
-   	int			local_h_errno;
+	struct hostent		*hp = NULL;
 	char   			*ha_buf=NULL, *p, *q;
+	int			socktype;
 
-	osip_debug(ld, "ldap_connect_to_host\n",0,0,0);
 	
+	switch(proto) {
+	case LDAP_PROTO_TCP: socktype = SOCK_STREAM;
+		osip_debug(ld, "ldap_connect_to_host: TCP %s:%d\n",host,port,0);
+		break;
+	case LDAP_PROTO_UDP: socktype = SOCK_DGRAM;
+		osip_debug(ld, "ldap_connect_to_host: TCP %s:%d\n",host,port,0);
+		break;
+
+	default:
+		osip_debug(ld, "ldap_connect_to_host: unknown proto: %d\n",
+			proto, 0, 0);
+		return -1;
+	}
+
 	if (host != NULL) {
-#ifdef HAVE_GETADDRINFO
-	        char serv[7];
+#if defined( HAVE_GETADDRINFO ) && defined( HAVE_INET_NTOP )
+		char serv[7];
+		int err;
 		struct addrinfo hints, *res, *sai;
 
 		memset( &hints, '\0', sizeof(hints) );
 		hints.ai_family = AF_UNSPEC;
-		hints.ai_socktype = SOCK_STREAM;
+		hints.ai_socktype = socktype;
 
-		snprintf(serv, sizeof serv, "%d", ntohs(port));
-		if ( getaddrinfo(host, serv, &hints, &res) ) {
-			osip_debug(ld, "ldap_connect_to_host:getaddrinfo failed\n",0,0,0);
+		snprintf(serv, sizeof serv, "%d", port );
+		if ( ( err = getaddrinfo(host, serv, &hints, &res) ) ) {
+			osip_debug(ld, "ldap_connect_to_host: getaddrinfo failed: %s\n",
+				AC_GAI_STRERROR(err), 0, 0);
 			return -1;
 		}
-		sai = res;
 		rc = -1;
-		do {
-			s = ldap_pvt_socket( ld, sai->ai_family );
-			if ( s == -1 ) {
+
+		for( sai=res; sai != NULL; sai=sai->ai_next) {
+			if( sai->ai_addr == NULL ) {
+				osip_debug(ld, "ldap_connect_to_host: getaddrinfo "
+					"ai_addr is NULL?\n", 0, 0, 0);
 				continue;
 			}
 
+			/* we assume AF_x and PF_x are equal for all x */
+			s = ldap_int_socket( ld, sai->ai_family, socktype );
+			if ( s == AC_SOCKET_INVALID ) {
+				continue;
+			}
+
+			if ( ldap_int_prepare_socket(ld, s, proto ) == -1 ) {
+				ldap_pvt_close_socket(ld, s);
+				break;
+			}
+
 			switch (sai->ai_family) {
 #ifdef LDAP_PF_INET6
 			case AF_INET6: {
@@ -330,17 +389,22 @@ ldap_connect_to_host(LDAP *ld, Sockbuf *sb, const char *host,
 					addr, serv, 0);
 			} break;
 			}
+
 			rc = ldap_pvt_connect(ld, s, sai->ai_addr, sai->ai_addrlen, async);
 			if ( (rc == 0) || (rc == -2) ) {
 				ber_sockbuf_ctrl( sb, LBER_SB_OPT_SET_FD, &s );
 				break;
 			}
 			ldap_pvt_close_socket(ld, s);
-		} while ((sai = sai->ai_next) != NULL);
+		}
 		freeaddrinfo(res);
 		return rc;
+
 #else
+		struct in_addr in;
 		if (! inet_aton( host, &in) ) {
+			int local_h_errno;
+			struct hostent he_buf;
 			rc = ldap_pvt_gethostbyname_a(host, &he_buf, &ha_buf,
 					&hp, &local_h_errno);
 
@@ -362,30 +426,30 @@ ldap_connect_to_host(LDAP *ld, Sockbuf *sb, const char *host,
 
 	rc = s = -1;
 	for ( i = 0; !use_hp || (hp->h_addr_list[i] != 0); ++i, rc = -1 ) {
-
-		s = ldap_pvt_socket( ld, AF_INET );
-		if ( s == -1 ) {
+		s = ldap_int_socket( ld, PF_INET, socktype );
+		if ( s == AC_SOCKET_INVALID ) {
 			/* use_hp ? continue : break; */
 			break;
 		}
 	   
-		if ( ldap_pvt_prepare_socket(ld, s) == -1 ) {
+		if ( ldap_int_prepare_socket( ld, s, proto ) == -1 ) {
 			ldap_pvt_close_socket(ld, s);
-			/* use_hp ? continue : break; */
 			break;
 		}
 
 		(void)memset((char *)&sin, '\0', sizeof(struct sockaddr_in));
 		sin.sin_family = AF_INET;
-		sin.sin_port = port;
+		sin.sin_port = htons((short) port);
 		p = (char *)&sin.sin_addr;
 		q = use_hp ? (char *)hp->h_addr_list[i] : (char *)&address;
-		SAFEMEMCPY(p, q, sizeof(sin.sin_addr) );
+		AC_MEMCPY(p, q, sizeof(sin.sin_addr) );
 
 		osip_debug(ld, "ldap_connect_to_host: Trying %s:%d\n", 
-				inet_ntoa(sin.sin_addr),ntohs(sin.sin_port),0);
+			inet_ntoa(sin.sin_addr),port,0);
 
-		rc = ldap_pvt_connect(ld, s, (struct sockaddr *)&sin, sizeof(struct sockaddr_in), async);
+		rc = ldap_pvt_connect(ld, s,
+			(struct sockaddr *)&sin, sizeof(struct sockaddr_in),
+			async);
    
 		if ( (rc == 0) || (rc == -2) ) {
 			ber_sockbuf_ctrl( sb, LBER_SB_OPT_SET_FD, &s );
@@ -401,27 +465,28 @@ ldap_connect_to_host(LDAP *ld, Sockbuf *sb, const char *host,
 	return rc;
 }
 
-#if defined( LDAP_API_FEATURE_X_OPENLDAP_V2_KBIND ) || defined( HAVE_TLS ) || defined( HAVE_CYRUS_SASL )
+#if defined( LDAP_API_FEATURE_X_OPENLDAP_V2_KBIND ) || \
+	defined( HAVE_CYRUS_SASL )
 char *
 ldap_host_connected_to( Sockbuf *sb )
 {
-	struct hostent		*hp;
-	socklen_t         	len;
-	struct sockaddr	        sa;
-	char                    *addr;
+	struct hostent	*hp;
+	socklen_t		len;
+	struct sockaddr	sa;
+	char			*addr;
+	char			*host;
 
    	/* buffers for gethostbyaddr_r */
-   	struct hostent		he_buf;
-        int			local_h_errno;
+   	struct hostent	he_buf;
+	int				local_h_errno;
    	char			*ha_buf=NULL;
-	ber_socket_t		sd;
-#define DO_RETURN(x) if (ha_buf) LDAP_FREE(ha_buf); return (x);
-   
+	ber_socket_t	sd;
+
 	(void)memset( (char *)&sa, '\0', sizeof( struct sockaddr ));
 	len = sizeof( sa );
 
 	ber_sockbuf_ctrl( sb, LBER_SB_OPT_GET_FD, &sd );
-	if ( getpeername( sd, (struct sockaddr *)&sa, &len ) == -1 ) {
+	if ( getpeername( sd, &sa, &len ) == -1 ) {
 		return( NULL );
 	}
 
@@ -432,6 +497,10 @@ ldap_host_connected_to( Sockbuf *sb )
 	 */
 
 	switch (sa.sa_family) {
+#ifdef LDAP_PF_LOCAL
+	case AF_LOCAL:
+		return LDAP_STRDUP( ldap_int_hostname );
+#endif
 #ifdef LDAP_PF_INET6
 	case AF_INET6:
 		addr = (char *) &((struct sockaddr_in6 *)&sa)->sin6_addr;
@@ -441,26 +510,49 @@ ldap_host_connected_to( Sockbuf *sb )
 	case AF_INET:
 		addr = (char *) &((struct sockaddr_in *)&sa)->sin_addr;
 		len = sizeof( struct in_addr );
+
+		{
+			struct sockaddr_in localhost;
+			localhost.sin_addr.s_addr = htonl( INADDR_ANY );
+
+			if( memcmp ( &localhost.sin_addr,
+				&((struct sockaddr_in *)&sa)->sin_addr,
+				sizeof(localhost.sin_addr) ) == 0 )
+			{
+				return LDAP_STRDUP( ldap_int_hostname );
+			}
+
+#ifdef INADDR_LOOPBACK
+			localhost.sin_addr.s_addr = htonl( INADDR_LOOPBACK );
+
+			if( memcmp ( &localhost.sin_addr,
+				&((struct sockaddr_in *)&sa)->sin_addr,
+				sizeof(localhost.sin_addr) ) == 0 )
+			{
+				return LDAP_STRDUP( ldap_int_hostname );
+			}
+#endif
+		}
 		break;
+
 	default:
 		return( NULL );
 		break;
 	}
+
+	host = NULL;
 	if ((ldap_pvt_gethostbyaddr_a( addr, len,
 		sa.sa_family, &he_buf, &ha_buf,
-		&hp,&local_h_errno ) ==0 ) && (hp != NULL) )
+		&hp,&local_h_errno ) == 0 ) &&
+		(hp != NULL) && ( hp->h_name != NULL ) )
 	{
-		if ( hp->h_name != NULL ) {
-			char *host = LDAP_STRDUP( hp->h_name );   
-			DO_RETURN( host );
-		}
+		host = LDAP_STRDUP( hp->h_name );   
 	}
 
-	DO_RETURN( NULL );
+	LDAP_FREE( ha_buf );
+	return host;
 }
-#undef DO_RETURN   
-   
-#endif /* LDAP_API_FEATURE_X_OPENLDAP_V2_KBIND || HAVE_TLS */
+#endif
 
 
 /* for UNIX */
@@ -585,11 +677,15 @@ ldap_int_ip_init( void )
 
 
 int
-do_ldap_select( LDAP *ld, struct timeval *timeout )
+ldap_int_select( LDAP *ld, struct timeval *timeout )
 {
 	struct selectinfo	*sip;
 
-	Debug( LDAP_DEBUG_TRACE, "do_ldap_select\n", 0, 0, 0 );
+#ifdef NEW_LOGGING
+	LDAP_LOG (( "os-ip", LDAP_LEVEL_ENTRY, "ldap_int_select\n" ));
+#else
+	Debug( LDAP_DEBUG_TRACE, "ldap_int_select\n", 0, 0, 0 );
+#endif
 
 	if ( ldap_int_tblsize == 0 )
 		ldap_int_ip_init();
