@@ -1,12 +1,11 @@
 /* ldbm.c - ldap dbm compatibility routines */
-/* $OpenLDAP: pkg/ldap/libraries/libldbm/ldbm.c,v 1.30.8.4 2000/07/06 15:13:00 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/libraries/libldbm/ldbm.c,v 1.72 2002/01/04 20:17:41 kurt Exp $ */
 /*
- * Copyright 1998-2000 The OpenLDAP Foundation, All Rights Reserved.
+ * Copyright 1998-2002 The OpenLDAP Foundation, All Rights Reserved.
  * COPYING RESTRICTIONS APPLY, see COPYRIGHT file
  */
 
 /* Patched for Berkeley DB version 2.0; /KSp; 98/02/23
- *
  *   - DB version 2.6.4b   ; 1998/12/28, /KSp
  *   - DB_DBT_MALLOC       ; 1998/03/22, /KSp
  *   - basic implementation; 1998/02/23, /KSp
@@ -35,7 +34,6 @@ ldbm_datum_free( LDBM ldbm, Datum data )
 	}
 }
 
-
 Datum
 ldbm_datum_dup( LDBM ldbm, Datum data )
 {
@@ -50,27 +48,48 @@ ldbm_datum_dup( LDBM ldbm, Datum data )
 		return( dup );
 	}
 	dup.dsize = data.dsize;
-	if ( (dup.dptr = (char *) malloc( data.dsize )) != NULL )
-		memcpy( dup.dptr, data.dptr, data.dsize );
+
+	if ( (dup.dptr = (char *) malloc( data.dsize )) != NULL ) {
+		AC_MEMCPY( dup.dptr, data.dptr, data.dsize );
+	}
 
 	return( dup );
 }
 
 static int ldbm_initialized = 0;
 
-#ifdef HAVE_BERKELEY_DB_THREAD
-#define LDBM_LOCK	((void)0)
-#define LDBM_UNLOCK	((void)0)
+#if defined( USE_BERKELEY_CDB )
+	/* not currently supported */
+#define LDBM_RWLOCK_INIT  ((void) 0)
+#define LDBM_RWLOCK_DESTROY ((void) 0)
+#define LDBM_WLOCK		((void) 0)
+#define LDBM_WUNLOCK	((void) 0)
+#define LDBM_RLOCK		((void) 0)
+#define LDBM_RUNLOCK	((void) 0)
+
+#elif defined( HAVE_BERKELEY_DB_THREAD )
+static ldap_pvt_thread_rdwr_t ldbm_big_rdwr;
+#define LDBM_RWLOCK_INIT (ldap_pvt_thread_rdwr_init( &ldbm_big_rdwr ))
+#define LDBM_RWLOCK_DESTROY (ldap_pvt_thread_rdwr_destroy( &ldbm_big_rdwr ))
+#define LDBM_WLOCK		(ldap_pvt_thread_rdwr_wlock(&ldbm_big_rdwr))
+#define LDBM_WUNLOCK	(ldap_pvt_thread_rdwr_wunlock(&ldbm_big_rdwr))
+#define LDBM_RLOCK		(ldap_pvt_thread_rdwr_rlock(&ldbm_big_rdwr))
+#define LDBM_RUNLOCK	(ldap_pvt_thread_rdwr_runlock(&ldbm_big_rdwr))
+
 #else
-
-/* Only DB2 or DB3 with DB_THREAD is thread-free */
 static ldap_pvt_thread_mutex_t ldbm_big_mutex;
-#define LDBM_LOCK	(ldap_pvt_thread_mutex_lock(&ldbm_big_mutex))
-#define LDBM_UNLOCK	(ldap_pvt_thread_mutex_unlock(&ldbm_big_mutex))
-
+#define LDBM_RWLOCK_INIT (ldap_pvt_thread_mutex_init( &ldbm_big_mutex ))
+#define LDBM_RWLOCK_DESTROY (ldap_pvt_thread_mutex_destroy( &ldbm_big_mutex ))
+#define LDBM_WLOCK		(ldap_pvt_thread_mutex_lock(&ldbm_big_mutex))
+#define LDBM_WUNLOCK	(ldap_pvt_thread_mutex_unlock(&ldbm_big_mutex))
+#define LDBM_RLOCK		LDBM_WLOCK
+#define LDBM_RUNLOCK	LDBM_WUNLOCK
 #endif
 
-
+#if !defined( HAVE_BERKELEY_DB ) || (DB_VERSION_MAJOR < 3)
+	/*  a dbEnv for BERKELEYv2  */
+DB_ENV *ldbm_Env = NULL;	/* real or fake, depending on db and version */
+#endif
 
 /*******************************************************************
  *                                                                 *
@@ -80,11 +99,11 @@ static ldap_pvt_thread_mutex_t ldbm_big_mutex;
  *******************************************************************/
 #if defined( HAVE_BERKELEY_DB ) && (DB_VERSION_MAJOR >= 2)
 
-
 void *
 ldbm_malloc( size_t size )
 {
-	return( calloc( 1, size ));
+	/* likely should use ber_mem* routines */
+	return( calloc( 1, size ) );
 }
 
 #ifdef LDAP_SYSLOG
@@ -95,78 +114,67 @@ static void
 ldbm_db_errcall( const char *prefix, char *message )
 {
 #ifdef LDAP_SYSLOG
-	syslog( LOG_INFO, "ldbm_db_errcall(): %s %s", prefix, message );
+	syslog( LOG_INFO, "ldbm: %s %s", prefix, message );
 #endif
 }
 
-/*  a dbEnv for BERKELEYv2  */
-DB_ENV           *ldbm_Env = NULL;
-
-int ldbm_initialize( void )
+int ldbm_initialize( const char* home )
 {
-	int     err;
+#if DB_VERSION_MAJOR < 3
+	int	err;
 	u_int32_t	envFlags;
+#endif
 
 	if(ldbm_initialized++) return 1;
 
+	{
+		char *version;
+		int major, minor, patch;
+		version = db_version( &major, &minor, &patch );
+
+		if( major != DB_VERSION_MAJOR ||
+			minor < DB_VERSION_MINOR )
+		{
+#ifdef LDAP_SYSLOG
+			syslog( LOG_INFO,
+				"ldbm_initialize(): version mismatch\nexpected: %s\ngot: %s\n",
+				DB_VERSION_STRING, version );
+#endif
+			return 1;
+		}
+	}
+
 #if DB_VERSION_MAJOR < 3
 	ldbm_Env = calloc( 1, sizeof( DB_ENV ));
 
 	if( ldbm_Env == NULL ) return 1;
 
-	ldbm_Env->db_errcall   = ldbm_db_errcall;
-	ldbm_Env->db_errpfx    = "==>";
-#else
-	ldbm_Env = NULL;
+	ldbm_Env->db_errcall	= ldbm_db_errcall;
+	ldbm_Env->db_errpfx		= "==>";
+
+	envFlags = DB_CREATE | DB_USE_ENVIRON;
+
+	/* add optional flags */
+#ifdef DB_PRIVATE
+	envFlags |= DB_PRIVATE;
+#endif
+#ifdef HAVE_BERKELEY_DB_THREAD
+	envFlags |= DB_THREAD; 
 #endif
 
-	envFlags = 
-#if defined( HAVE_BERKELEY_DB_THREAD )
-		DB_THREAD |
-#endif
-		DB_CREATE;
-
-#if DB_VERSION_MAJOR >= 3
-	err = db_env_create( &ldbm_Env, 0 );
-#elif DB_VERSION_MAJOR >= 2
-	err = db_appinit( NULL, NULL, ldbm_Env, envFlags );
-#endif
+	err = db_appinit( home, NULL, ldbm_Env, envFlags );
 
 	if ( err ) {
-		char error[BUFSIZ];
-
-		sprintf( error, "%s (%d)\n", STRERROR( err ), err );
-
 #ifdef LDAP_SYSLOG
-		syslog( LOG_INFO,
-			"ldbm_initialize(): FATAL error in db_appinit() : %s\n",
-			error );
+		syslog( LOG_INFO, "ldbm_initialize(): "
+			"FATAL error (%d) in db_appinit()\n", err );
 #endif
 	 	return( 1 );
 	}
-
-#if DB_VERSION_MAJOR >= 3
-	ldbm_Env->set_errcall( ldbm_Env, ldbm_db_errcall );
-	ldbm_Env->set_errpfx( ldbm_Env, "==>" );
-
-        envFlags |= DB_INIT_MPOOL;
-        err = ldbm_Env->open( ldbm_Env, NULL, NULL, envFlags, 0 );
-        if ( err != 0 )
-        {
-            char error[BUFSIZ];
-
-            sprintf( error, "%s (%d)\n", STRERROR( err ), err );
-
-#ifdef LDAP_SYSLOG
-            syslog( LOG_INFO,
-                    "ldbm_initialize(): FATAL error in db_appinit() : %s\n",
-                    error );
-#endif
-		    ldbm_Env->close( ldbm_Env, 0 );
-            return( 1 );
-	}
 #endif
 
+	LDBM_RWLOCK_INIT;
+
 	return 0;
 }
 
@@ -174,22 +182,21 @@ int ldbm_shutdown( void )
 {
 	if( !ldbm_initialized ) return 1;
 
-#if DB_VERSION_MAJOR >= 3
-        ldbm_Env->close( ldbm_Env, 0 );
-#else
+#if DB_VERSION_MAJOR < 3
 	db_appexit( ldbm_Env );
 #endif
 
+	LDBM_RWLOCK_DESTROY;
 	return 0;
 }
 
 #else  /* some DB other than Berkeley V2 or greater */
 
-int ldbm_initialize( void )
+int ldbm_initialize( const char * home )
 {
 	if(ldbm_initialized++) return 1;
 
-	ldap_pvt_thread_mutex_init( &ldbm_big_mutex );
+	LDBM_RWLOCK_INIT;
 
 	return 0;
 }
@@ -198,13 +205,92 @@ int ldbm_shutdown( void )
 {
 	if( !ldbm_initialized ) return 1;
 
-	ldap_pvt_thread_mutex_destroy( &ldbm_big_mutex );
+	LDBM_RWLOCK_DESTROY;
 
 	return 0;
 }
 
-#endif /* ifdef HAVE_BERKELEY_DB */
+#endif /* HAVE_BERKELEY_DB */
 
+#if defined( HAVE_BERKELEY_DB ) && (DB_VERSION_MAJOR >= 3)
+
+DB_ENV *ldbm_initialize_env(const char *home, int dbcachesize, int *envdirok)
+{
+	DB_ENV *env = NULL;    
+	int     err;
+	u_int32_t	envFlags;
+
+	err = db_env_create( &env, 0 );
+
+	if ( err ) {
+#ifdef LDAP_SYSLOG
+		syslog( LOG_INFO, "ldbm_initialize_env(): "
+			"FATAL error in db_env_create() : %s (%d)\n",
+			db_strerror( err ), err );
+#endif
+		return NULL;
+	}
+
+#if DB_VERSION_MAJOR > 3 || DB_VERSION_MINOR >= 3
+	/* This interface appeared in 3.3 */
+	env->set_alloc( env, ldbm_malloc, NULL, NULL );
+#endif
+
+	env->set_errcall( env, ldbm_db_errcall );
+	env->set_errpfx( env, "==>" );
+	if (dbcachesize) {
+		env->set_cachesize( env, 0, dbcachesize, 0 );
+	}
+
+	envFlags = DB_CREATE | DB_INIT_MPOOL | DB_USE_ENVIRON;
+#ifdef DB_PRIVATE
+	envFlags |= DB_PRIVATE;
+#endif
+#ifdef DB_MPOOL_PRIVATE
+	envFlags |= DB_MPOOL_PRIVATE;
+#endif
+#ifdef HAVE_BERKELEY_DB_THREAD
+	envFlags |= DB_THREAD;
+#endif
+
+#if DB_VERSION_MAJOR > 3 || DB_VERSION_MINOR > 0
+	err = env->open( env, home, envFlags, 0 );
+#else
+	/* 3.0.x requires an extra argument */
+	err = env->open( env, home, NULL, envFlags, 0 );
+#endif
+
+	if ( err != 0 ) {
+#ifdef LDAP_SYSLOG
+		syslog(	LOG_INFO, "ldbm_initialize_env(): "
+			"FATAL error in dbEnv->open() : %s (%d)\n",
+			db_strerror( err ), err );
+#endif
+		env->close( env, 0 );
+		return NULL;
+	}
+
+	*envdirok = 1;
+	return env;
+}
+
+void ldbm_shutdown_env(DB_ENV *env)
+{
+	env->close( env, 0 );
+}
+
+#else
+
+DB_ENV *ldbm_initialize_env(const char *home, int dbcachesize, int *envdirok)
+{
+	return ldbm_Env;
+}
+
+void ldbm_shutdown_env(DB_ENV *env)
+{
+}
+
+#endif
 
 #if defined( LDBM_USE_DBHASH ) || defined( LDBM_USE_DBBTREE )
 
@@ -215,62 +301,67 @@ int ldbm_shutdown( void )
  *****************************************************************/
 
 LDBM
-ldbm_open( char *name, int rw, int mode, int dbcachesize )
+ldbm_open( DB_ENV *env, char *name, int rw, int mode, int dbcachesize )
 {
 	LDBM		ret = NULL;
 
 #if DB_VERSION_MAJOR >= 3
-       int err;
-       LDBM_LOCK;
-       err = db_create( &ret, ldbm_Env, 0 );
-       if ( err != 0 )
-       {
-           char error[BUFSIZ];
+	int err;
 
-           sprintf( error, "%s (%d)\n", STRERROR( err ), err );
+	LDBM_WLOCK;
 
-           (void)ret->close(ret, 0);
-           return NULL;
-       }
+	err = db_create( &ret, env, 0 );
+	if ( err != 0 ) {
+		(void)ret->close(ret, 0);
+		LDBM_WUNLOCK;
 
-       ret->set_pagesize( ret, DEFAULT_DB_PAGE_SIZE );
-       ret->set_malloc( ret, ldbm_malloc );
-       ret->set_cachesize( ret, 0, dbcachesize, 0 );
-       err = ret->open( ret, name, NULL, DB_TYPE, rw, mode);
-       LDBM_UNLOCK;
-       if ( err != 0 )
-       {
-           char error[BUFSIZ];
+		return NULL;
+	}
 
-           sprintf( error, "%s (%d)\n", STRERROR( err ), err );
+#if DB_VERSION_MAJOR == 3 && DB_VERSION_MINOR < 3
+	ret->set_malloc( ret, ldbm_malloc );
+#endif
 
-           (void)ret->close(ret, 0);
-           return NULL;
-       }
+	ret->set_pagesize( ret, DEFAULT_DB_PAGE_SIZE );
+
+	/* likely should use ber_mem* routines */
+
+	err = ret->open( ret, name, NULL, DB_TYPE, rw, mode);
+
+	if ( err != 0 ) {
+		int tmp = errno;
+		(void)ret->close(ret, 0);
+		errno = tmp;
+
+		LDBM_WUNLOCK;
+		return NULL;
+	}
+
+	LDBM_WUNLOCK;
  
 #elif DB_VERSION_MAJOR >= 2
 	DB_INFO dbinfo;
 
 	memset( &dbinfo, '\0', sizeof( dbinfo ));
 
-#if defined( DB_VERSION_MAJOR ) && defined( DB_VERSION_MINOR ) && \
-    DB_VERSION_MAJOR == 2 && DB_VERSION_MINOR == 4
+#if	DB_VERSION_MAJOR == 2 && DB_VERSION_MINOR == 4
 	/*
 	 * BerkeleyDB 2.4 do not allow db_cachesize
 	 * to be specified if an DB_ENV is.
 	 */
 #else
 	/* set db_cachesize of MPOOL is NOT being used. */
-	if (( ldbm_Env == NULL ) || ( ldbm_Env->mp_info == NULL ))
+	if (( ldbm_Env == NULL ) || ( ldbm_Env->mp_info == NULL )) {
 		dbinfo.db_cachesize = dbcachesize;
+	}
 #endif
 
-	dbinfo.db_pagesize  = DEFAULT_DB_PAGE_SIZE;
-	dbinfo.db_malloc    = ldbm_malloc;
+	dbinfo.db_pagesize	= DEFAULT_DB_PAGE_SIZE;
+	dbinfo.db_malloc	= ldbm_malloc;
 
-	LDBM_LOCK;
-    (void) db_open( name, DB_TYPE, rw, mode, ldbm_Env, &dbinfo, &ret );
-	LDBM_UNLOCK;
+	LDBM_WLOCK;
+	(void) db_open( name, DB_TYPE, rw, mode, ldbm_Env, &dbinfo, &ret );
+	LDBM_WUNLOCK;
 
 #else
 	void		*info;
@@ -289,35 +380,32 @@ ldbm_open( char *name, int rw, int mode, int dbcachesize )
 		info = NULL;
 	}
 
-	LDBM_LOCK;
+	LDBM_WLOCK;
 	ret = dbopen( name, rw, mode, DB_TYPE, info );
-	LDBM_UNLOCK;
-
+	LDBM_WUNLOCK;
 #endif
 
-	return( ret );
+	return ret;
 }
 
 void
 ldbm_close( LDBM ldbm )
 {
-	LDBM_LOCK;
-#if DB_VERSION_MAJOR >= 3
-        ldbm->close( ldbm, 0 );
-#elif DB_VERSION_MAJOR >= 2
-	(*ldbm->close)( ldbm, 0 );
+	LDBM_WLOCK;
+#if DB_VERSION_MAJOR >= 2
+	ldbm->close( ldbm, 0 );
 #else
-	(*ldbm->close)( ldbm );
+	ldbm->close( ldbm );
 #endif
-	LDBM_UNLOCK;
+	LDBM_WUNLOCK;
 }
 
 void
 ldbm_sync( LDBM ldbm )
 {
-	LDBM_LOCK;
+	LDBM_WLOCK;
 	(*ldbm->sync)( ldbm, 0 );
-	LDBM_UNLOCK;
+	LDBM_WUNLOCK;
 }
 
 Datum
@@ -326,34 +414,30 @@ ldbm_fetch( LDBM ldbm, Datum key )
 	Datum	data;
 	int	rc;
 
-	LDBM_LOCK;
-#if DB_VERSION_MAJOR >= 3
-        ldbm_datum_init( data );
+	LDBM_RLOCK;
 
-        data.flags = DB_DBT_MALLOC;
-
-        if ( (rc = ldbm->get( ldbm, NULL, &key, &data, 0 )) != 0 ) {
-            ldbm_datum_free( ldbm, data );
-
-#elif DB_VERSION_MAJOR >= 2
+#if DB_VERSION_MAJOR >= 2
 	ldbm_datum_init( data );
 
 	data.flags = DB_DBT_MALLOC;
 
-	if ( (rc = (*ldbm->get)( ldbm, NULL, &key, &data, 0 )) != 0 ) {
+	if ( (rc = ldbm->get( ldbm, NULL, &key, &data, 0 )) != 0 ) {
 		ldbm_datum_free( ldbm, data );
+		data.dptr = NULL;
+		data.dsize = 0;
+	}
 #else
-	if ( (rc = (*ldbm->get)( ldbm, &key, &data, 0 )) == 0 ) {
+	if ( (rc = ldbm->get( ldbm, &key, &data, 0 )) == 0 ) {
 		/* Berkeley DB 1.85 don't malloc the data for us */
 		/* duplicate it for to ensure reentrancy */
 		data = ldbm_datum_dup( ldbm, data );
 	} else {
-#endif
 		data.dptr = NULL;
 		data.dsize = 0;
 	}
+#endif
 
-	LDBM_UNLOCK;
+	LDBM_RUNLOCK;
 
 	return( data );
 }
@@ -363,29 +447,19 @@ ldbm_store( LDBM ldbm, Datum key, Datum data, int flags )
 {
 	int	rc;
 
-	LDBM_LOCK;
+	LDBM_WLOCK;
 
-#if DB_VERSION_MAJOR >= 3
-        rc = ldbm->put( ldbm, NULL, &key, &data, flags & ~LDBM_SYNC );
-       if ( rc != 0 )
-       {
-           char error[BUFSIZ];
-
-           sprintf( error, "%s (%d)\n", STRERROR( rc ), rc );
-       }
-       rc = (-1) * rc;
-
-#elif DB_VERSION_MAJOR >= 2
-	rc = (*ldbm->put)( ldbm, NULL, &key, &data, flags & ~LDBM_SYNC );
-	rc = (-1 ) * rc;
+#if DB_VERSION_MAJOR >= 2
+	rc = ldbm->put( ldbm, NULL, &key, &data, flags & ~LDBM_SYNC );
+	rc = (-1) * rc;
 #else
-	rc = (*ldbm->put)( ldbm, &key, &data, flags & ~LDBM_SYNC );
+	rc = ldbm->put( ldbm, &key, &data, flags & ~LDBM_SYNC );
 #endif
 
 	if ( flags & LDBM_SYNC )
-		(*ldbm->sync)( ldbm, 0 );
+		ldbm->sync( ldbm, 0 );
 
-	LDBM_UNLOCK;
+	LDBM_WUNLOCK;
 
 	return( rc );
 }
@@ -395,20 +469,17 @@ ldbm_delete( LDBM ldbm, Datum key )
 {
 	int	rc;
 
-	LDBM_LOCK;
+	LDBM_WLOCK;
 
-#if DB_VERSION_MAJOR >= 3
+#if DB_VERSION_MAJOR >= 2
 	rc = ldbm->del( ldbm, NULL, &key, 0 );
-	rc = (-1 ) * rc;
-#elif DB_VERSION_MAJOR >= 2
-	rc = (*ldbm->del)( ldbm, NULL, &key, 0 );
-	rc = (-1 ) * rc;
+	rc = (-1) * rc;
 #else
-	rc = (*ldbm->del)( ldbm, &key, 0 );
+	rc = ldbm->del( ldbm, &key, 0 );
 #endif
-	(*ldbm->sync)( ldbm, 0 );
+	ldbm->sync( ldbm, 0 );
 
-	LDBM_UNLOCK;
+	LDBM_WUNLOCK;
 
 	return( rc );
 }
@@ -417,6 +488,7 @@ Datum
 ldbm_firstkey( LDBM ldbm, LDBMCursor **dbch )
 {
 	Datum	key, data;
+	int	rc;
 
 #if DB_VERSION_MAJOR >= 2
 	LDBMCursor  *dbci;
@@ -426,84 +498,78 @@ ldbm_firstkey( LDBM ldbm, LDBMCursor **dbch )
 
 	key.flags = data.flags = DB_DBT_MALLOC;
 
-	LDBM_LOCK;
+	LDBM_RLOCK;
 
 	/* acquire a cursor for the DB */
-# if DB_VERSION_MAJOR >= 3
-        if ( ldbm->cursor( ldbm, NULL, &dbci, 0 ) )
-# elif defined( DB_VERSION_MAJOR ) && defined( DB_VERSION_MINOR ) && \
-    (DB_VERSION_MAJOR == 2 && DB_VERSION_MINOR < 6)
+# if DB_VERSION_MAJOR >= 3 || (DB_VERSION_MAJOR == 2 && DB_VERSION_MINOR > 5)
+	rc = ldbm->cursor( ldbm, NULL, &dbci, 0 );
+# else
+	rc = ldbm->cursor( ldbm, NULL, &dbci );
+# endif
 
-	if ( (*ldbm->cursor)( ldbm, NULL, &dbci ))
-
-#  else
-	if ( (*ldbm->cursor)( ldbm, NULL, &dbci, 0 ))
-#  endif
-	{
+	if( rc ) {
 		key.dptr = NULL;
-		return( key );
 	} else {
 		*dbch = dbci;
-		if ( (*dbci->c_get)( dbci, &key, &data, DB_NEXT ) == 0 ) {
+		if ( dbci->c_get( dbci, &key, &data, DB_NEXT ) == 0 ) {
 			ldbm_datum_free( ldbm, data );
+		} else {
+			key.dptr = NULL;
+			key.dsize = 0;
 		}
-	else {
-#else
-	int	rc;
-
-	LDBM_LOCK;
-
-	if ( (rc = (*ldbm->seq)( ldbm, &key, &data, R_FIRST )) == 0 ) {
-		key = ldbm_datum_dup( ldbm, key );
 	}
-	else {
-#endif
+
+	LDBM_RUNLOCK;
+
+#else
+	LDBM_RLOCK;
+
+	rc = ldbm->seq( ldbm, &key, &data, R_FIRST );
+
+	if ( rc == 0 ) {
+		key = ldbm_datum_dup( ldbm, key );
+	} else {
 		key.dptr = NULL;
 		key.dsize = 0;
 	}
 
-#if DB_VERSION_MAJOR >= 2
-	}
+	LDBM_RUNLOCK;
 #endif
 
-	LDBM_UNLOCK;
-
 	return( key );
 }
 
 Datum
 ldbm_nextkey( LDBM ldbm, Datum key, LDBMCursor *dbcp )
 {
+	int	rc;
 	Datum	data;
 
+	LDBM_RLOCK;
+
 #if DB_VERSION_MAJOR >= 2
 	ldbm_datum_init( data );
 
 	ldbm_datum_free( ldbm, key );
 	key.flags = data.flags = DB_DBT_MALLOC;
 
-	LDBM_LOCK;
-
-	if ( (*dbcp->c_get)( dbcp, &key, &data, DB_NEXT ) == 0 ) {
+	rc = dbcp->c_get( dbcp, &key, &data, DB_NEXT );
+	if ( rc == 0 ) {
 		ldbm_datum_free( ldbm, data );
-	}
-	else {
+	} else
 #else
-	int	rc;
+	rc = ldbm->seq( ldbm, &key, &data, R_NEXT );
 
-	LDBM_LOCK;
-
-	if ( (rc = (*ldbm->seq)( ldbm, &key, &data, R_NEXT )) == 0 ) {
+	if ( rc == 0 ) {
 		key = ldbm_datum_dup( ldbm, key );
-	}
-	else {
+	} else
 #endif
+	{
 		key.dptr = NULL;
 		key.dsize = 0;
 	}
 
-	LDBM_UNLOCK;
-
+	LDBM_RUNLOCK;
 	return( key );
 }
 
@@ -532,17 +598,17 @@ ldbm_errno( LDBM ldbm )
  *****************************************************************/
 
 LDBM
-ldbm_open( char *name, int rw, int mode, int dbcachesize )
+ldbm_open( DB_ENV *env, char *name, int rw, int mode, int dbcachesize )
 {
 	LDBM		db;
 #ifdef HAVE_ST_BLKSIZE
 		struct stat	st;
 #endif
 
-	LDBM_LOCK;
+	LDBM_WLOCK;
 
-	if ( (db =  gdbm_open( name, 0, rw | GDBM_FAST, mode, 0 )) == NULL ) {
-		LDBM_UNLOCK;
+	if ( (db = gdbm_open( name, 0, rw | GDBM_FAST, mode, 0 )) == NULL ) {
+		LDBM_WUNLOCK;
 		return( NULL );
 	}
 
@@ -560,7 +626,7 @@ ldbm_open( char *name, int rw, int mode, int dbcachesize )
 	}
 #endif
 
-	LDBM_UNLOCK;
+	LDBM_WUNLOCK;
 
 	return( db );
 }
@@ -568,17 +634,17 @@ ldbm_open( char *name, int rw, int mode, int dbcachesize )
 void
 ldbm_close( LDBM ldbm )
 {
-	LDBM_LOCK;
+	LDBM_WLOCK;
 	gdbm_close( ldbm );
-	LDBM_UNLOCK;
+	LDBM_WUNLOCK;
 }
 
 void
 ldbm_sync( LDBM ldbm )
 {
-	LDBM_LOCK;
+	LDBM_WLOCK;
 	gdbm_sync( ldbm );
-	LDBM_UNLOCK;
+	LDBM_WUNLOCK;
 }
 
 Datum
@@ -586,9 +652,9 @@ ldbm_fetch( LDBM ldbm, Datum key )
 {
 	Datum d;
 
-	LDBM_LOCK;
+	LDBM_RLOCK;
 	d = gdbm_fetch( ldbm, key );
-	LDBM_UNLOCK;
+	LDBM_RUNLOCK;
 
 	return d;
 }
@@ -598,11 +664,11 @@ ldbm_store( LDBM ldbm, Datum key, Datum data, int flags )
 {
 	int	rc;
 
-	LDBM_LOCK;
+	LDBM_WLOCK;
 	rc = gdbm_store( ldbm, key, data, flags & ~LDBM_SYNC );
 	if ( flags & LDBM_SYNC )
 		gdbm_sync( ldbm );
-	LDBM_UNLOCK;
+	LDBM_WUNLOCK;
 
 	return( rc );
 }
@@ -612,10 +678,10 @@ ldbm_delete( LDBM ldbm, Datum key )
 {
 	int	rc;
 
-	LDBM_LOCK;
+	LDBM_WLOCK;
 	rc = gdbm_delete( ldbm, key );
 	gdbm_sync( ldbm );
-	LDBM_UNLOCK;
+	LDBM_WUNLOCK;
 
 	return( rc );
 }
@@ -625,9 +691,9 @@ ldbm_firstkey( LDBM ldbm, LDBMCursor **dbcp )
 {
 	Datum d;
 
-	LDBM_LOCK;
+	LDBM_RLOCK;
 	d = gdbm_firstkey( ldbm );
-	LDBM_UNLOCK;
+	LDBM_RUNLOCK;
 
 	if ( d.dptr != NULL ) {
 		*dbcp = (Datum *) malloc( sizeof( Datum ) );
@@ -642,9 +708,9 @@ ldbm_nextkey( LDBM ldbm, Datum key, LDBMCursor *dbcp )
 {
 	Datum d;
 
-	LDBM_LOCK;
+	LDBM_RLOCK;
 	d = gdbm_nextkey( ldbm, *dbcp );
-	LDBM_UNLOCK;
+	LDBM_RUNLOCK;
 
 	ldbm_datum_free( ldbm, *dbcp );
 
@@ -662,9 +728,9 @@ ldbm_errno( LDBM ldbm )
 {
 	int err;
 
-	LDBM_LOCK;
+	LDBM_WLOCK;
 	err = gdbm_errno;
-	LDBM_UNLOCK;
+	LDBM_WUNLOCK;
 
 	return( err );
 }
@@ -688,7 +754,6 @@ ldbm_errno( LDBM ldbm )
 
 /* Use chaining */
 
-
 #define mdbm_store	mdbm_chain_store
 #define mdbm_fetch	mdbm_chain_fetch
 #define mdbm_delete	mdbm_chain_delete
@@ -706,7 +771,7 @@ ldbm_errno( LDBM ldbm )
  *****************************************************************/
 
 LDBM
-ldbm_open( char *name, int rw, int mode, int dbcachesize )
+ldbm_open( DB_ENV *env, char *name, int rw, int mode, int dbcachesize )
 {
 	LDBM		db;
 
@@ -717,26 +782,23 @@ ldbm_open( char *name, int rw, int mode, int dbcachesize )
 	fflush( stdout );
 #endif
 
-	LDBM_LOCK;	/* We need locking here, this is the only non-thread
-			 * safe function we have.
-			 */
+	LDBM_WLOCK;	/* We need locking here, this is the only non-thread
+		* safe function we have.  */
 
 	if ( (db =  mdbm_open( name, rw, mode, MDBM_PG_SZ )) == NULL ) {
-
-		LDBM_UNLOCK;
+		LDBM_WUNLOCK;
 #ifdef MDBM_DEBUG
 		fprintf( stdout, "<==(mdbm)ldbm_open(db=NULL)\n" );
 		fflush( stdout );
 #endif
 		return( NULL );
-
 	}
 
 #ifdef MDBM_CHAIN
 	(void)mdbm_set_chain(db);
 #endif
 
-	LDBM_UNLOCK;
+	LDBM_WUNLOCK;
 
 #ifdef MDBM_DEBUG
 	fprintf( stdout, "<==(mdbm)ldbm_open(db=%p)\n", db );
@@ -744,16 +806,11 @@ ldbm_open( char *name, int rw, int mode, int dbcachesize )
 #endif
 
 	return( db );
-
 }
 
-
-
-
 void
 ldbm_close( LDBM ldbm )
 {
-
 	/* Open and close are not reentrant so we need to use locks here */
 
 #ifdef MDBM_DEBUG
@@ -762,35 +819,28 @@ ldbm_close( LDBM ldbm )
 	fflush( stdout );
 #endif
 
-	LDBM_LOCK;
+	LDBM_WLOCK;
 	mdbm_close( ldbm );
-	LDBM_UNLOCK;
+	LDBM_WUNLOCK;
 
 #ifdef MDBM_DEBUG
 	fprintf( stdout, "<==(mdbm)ldbm_close()\n" );
 	fflush( stdout );
 #endif
-
 }
 
-
-
-
 void
 ldbm_sync( LDBM ldbm )
 {
-
 	/* XXX: Not sure if this is re-entrant need to check code, if so
 	 * you can leave LOCKS out.
 	 */
 
-	LDBM_LOCK;
+	LDBM_WLOCK;
 	mdbm_sync( ldbm );
-        LDBM_UNLOCK;
-
+	LDBM_WUNLOCK;
 }
 
-
 #define MAX_MDBM_RETRY	5
 
 Datum
@@ -805,13 +855,11 @@ ldbm_fetch( LDBM ldbm, Datum key )
 	 * mode.
 	 */
 
-	/* LDBM_LOCK; */
-
 #ifdef NO_NULL_KEY
 	k.key.dsize = key.dsize + 1;			
 	k.key.dptr = malloc(k.key.dsize);
 	*(k.key.dptr) = 'l';
-	memcpy( (void *)(k.key.dptr + 1), key.dptr, key.dsize );	
+	AC_MEMCPY( (void *)(k.key.dptr + 1), key.dptr, key.dsize );	
 #else
 	k.key = key;
 #endif	
@@ -819,47 +867,34 @@ ldbm_fetch( LDBM ldbm, Datum key )
 	k.val.dptr = NULL;
 	k.val.dsize = 0;
 
+	/* LDBM_RLOCK; */
 	do {
-
 		d = mdbm_fetch( ldbm, k );
 
 		if ( d.dsize > 0 ) {
-
 			if ( k.val.dptr != NULL ) {
-			    
-			    free( k.val.dptr );
-
+				free( k.val.dptr );
 			}
 
 			if ( (k.val.dptr = malloc( d.dsize )) != NULL ) {
-		
 				k.val.dsize = d.dsize;
 				d = mdbm_fetch( ldbm, k );
 
 			} else { 
-
 				d.dsize = 0;
 				break;
-			
 			}
-
 		}/* if ( d.dsize > 0 ) */
-
 	} while ((d.dsize > k.val.dsize) && (++retry < MAX_MDBM_RETRY));
-
-	/* LDBM_UNLOCK; */
+	/* LDBM_RUNLOCK; */
 
 #ifdef NO_NULL_KEY
 	free(k.key.dptr);
 #endif
 
 	return d;
-
 }
 
-
-
-
 int
 ldbm_store( LDBM ldbm, Datum key, Datum data, int flags )
 {
@@ -873,13 +908,13 @@ ldbm_store( LDBM ldbm, Datum key, Datum data, int flags )
 	fflush( stdout );
 #endif
 
-	/* LDBM_LOCK; */
+	/* LDBM_WLOCK; */
 
 #ifdef NO_NULL_KEY
 	int_key.dsize = key.dsize + 1;
 	int_key.dptr = malloc( int_key.dsize );
 	*(int_key.dptr) = 'l';	/* Must not be NULL !*/
-	memcpy( (void *)(int_key.dptr + 1), key.dptr, key.dsize );
+	AC_MEMCPY( (void *)(int_key.dptr + 1), key.dptr, key.dsize );
 #else
 	int_key = key;
 #endif
@@ -889,7 +924,7 @@ ldbm_store( LDBM ldbm, Datum key, Datum data, int flags )
 		mdbm_sync( ldbm );
 	}
 
-	/* LDBM_UNLOCK; */
+	/* LDBM_WUNLOCK; */
 
 #ifdef MDBM_DEBUG
 	fprintf( stdout, "<==(mdbm)ldbm_store(rc=%d)\n", rc );
@@ -901,46 +936,38 @@ ldbm_store( LDBM ldbm, Datum key, Datum data, int flags )
 #endif
 
 	return( rc );
-
 }
 
-
-
 int
 ldbm_delete( LDBM ldbm, Datum key )
 {
 	int	rc;
 	Datum	int_key;
 
-	/* LDBM_LOCK; */
+	/* LDBM_WLOCK; */
 
 #ifdef NO_NULL_KEY
 	int_key.dsize = key.dsize + 1;
 	int_key.dptr = malloc(int_key.dsize);
 	*(int_key.dptr) = 'l';
-	memcpy( (void *)(int_key.dptr + 1), key.dptr, key.dsize );	
+	AC_MEMCPY( (void *)(int_key.dptr + 1), key.dptr, key.dsize );	
 #else
 	int_key = key;
 #endif
 	
 	rc = mdbm_delete( ldbm, int_key );
 
-	/* LDBM_UNLOCK; */
+	/* LDBM_WUNLOCK; */
 #ifdef NO_NULL_KEY
 	free(int_key.dptr);
 #endif
 
 	return( rc );
-
 }
 
-
-
-
 static Datum
 ldbm_get_next( LDBM ldbm, kvpair (*fptr)(MDBM *, kvpair) ) 
 {
-
 	kvpair	out;
 	kvpair	in;
 	Datum	ret;
@@ -951,7 +978,7 @@ ldbm_get_next( LDBM ldbm, kvpair (*fptr)(MDBM *, kvpair) )
 	int	delta = 0;
 #endif
 
-	/* LDBM_LOCK; */
+	/* LDBM_RLOCK; */
 
 	in.key.dsize = sz;	/* Assume first key in one pg */
 	in.key.dptr = malloc(sz);
@@ -965,55 +992,39 @@ ldbm_get_next( LDBM ldbm, kvpair (*fptr)(MDBM *, kvpair) )
 	out = fptr( ldbm, in );
 
 	if (out.key.dsize > 0) {
+		ret.dsize = out.key.dsize - delta;
 
-	    ret.dsize = out.key.dsize - delta;
-	    if ((ret.dptr = (char *)malloc(ret.dsize)) == NULL) { 
-
-		ret.dsize = 0;
-		ret.dptr = NULL;
-
-	    } else {
-
-		memcpy(ret.dptr, (void *)(out.key.dptr + delta),
-		       ret.dsize );
+		if ((ret.dptr = (char *)malloc(ret.dsize)) == NULL) { 
+			ret.dsize = 0;
+			ret.dptr = NULL;
 
+		} else {
+			AC_MEMCPY(ret.dptr, (void *)(out.key.dptr + delta),
+				ret.dsize );
 	    }
-
 	}
 
-	/* LDBM_UNLOCK; */
+	/* LDBM_RUNLOCK; */
 	
 	free(in.key.dptr);
-
 	return ret;
-
 }
 
-
-
-
 Datum
 ldbm_firstkey( LDBM ldbm, LDBMCursor **dbcp )
 {
-
 	return ldbm_get_next( ldbm, mdbm_first );
-
 }
 
-
-
-
 Datum
 ldbm_nextkey( LDBM ldbm, Datum key, LDBMCursor *dbcp )
 {
-
 	/* XXX:
-	 * don't know if this will affect the LDAP server opertaion 
+	 * don't know if this will affect the LDAP server operation 
 	 * but mdbm cannot take and input key.
 	 */
 
 	return ldbm_get_next( ldbm, mdbm_next );
-
 }
 
 int
@@ -1021,12 +1032,8 @@ ldbm_errno( LDBM ldbm )
 {
 	/* XXX: best we can do with current  mdbm interface */
 	return( errno );
-
 }
 
-
-
-
 #elif defined( HAVE_NDBM )
 
 /*****************************************************************
@@ -1037,13 +1044,13 @@ ldbm_errno( LDBM ldbm )
 
 /* ARGSUSED */
 LDBM
-ldbm_open( char *name, int rw, int mode, int dbcachesize )
+ldbm_open( DB_ENV *env, char *name, int rw, int mode, int dbcachesize )
 {
 	LDBM ldbm;
 
-	LDBM_LOCK;
+	LDBM_WLOCK;
 	ldbm = dbm_open( name, rw, mode );
-	LDBM_UNLOCK;
+	LDBM_WUNLOCK;
 
 	return( ldbm );
 }
@@ -1051,9 +1058,9 @@ ldbm_open( char *name, int rw, int mode, int dbcachesize )
 void
 ldbm_close( LDBM ldbm )
 {
-	LDBM_LOCK;
+	LDBM_WLOCK;
 	dbm_close( ldbm );
-	LDBM_UNLOCK;
+	LDBM_WUNLOCK;
 }
 
 /* ARGSUSED */
@@ -1068,9 +1075,9 @@ ldbm_fetch( LDBM ldbm, Datum key )
 {
 	Datum d;
 
-	LDBM_LOCK;
+	LDBM_RLOCK;
 	d = ldbm_datum_dup( ldbm, dbm_fetch( ldbm, key ) );
-	LDBM_UNLOCK;
+	LDBM_RUNLOCK;
 
 	return d;
 }
@@ -1080,9 +1087,9 @@ ldbm_store( LDBM ldbm, Datum key, Datum data, int flags )
 {
 	int rc;
 
-	LDBM_LOCK;
+	LDBM_WLOCK;
 	rc = dbm_store( ldbm, key, data, flags );
-	LDBM_UNLOCK;
+	LDBM_WUNLOCK;
 
 	return rc;
 }
@@ -1092,9 +1099,9 @@ ldbm_delete( LDBM ldbm, Datum key )
 {
 	int rc;
 
-	LDBM_LOCK;
+	LDBM_WLOCK;
 	rc = dbm_delete( ldbm, key );
-	LDBM_UNLOCK;
+	LDBM_WUNLOCK;
 
 	return rc;
 }
@@ -1104,9 +1111,9 @@ ldbm_firstkey( LDBM ldbm, LDBMCursor **dbcp )
 {
 	Datum d;
 
-	LDBM_LOCK;
+	LDBM_RLOCK;
 	d = dbm_firstkey( ldbm );
-	LDBM_UNLOCK;
+	LDBM_RUNLOCK;
 
 	return d;
 }
@@ -1116,9 +1123,9 @@ ldbm_nextkey( LDBM ldbm, Datum key, LDBMCursor *dbcp )
 {
 	Datum d;
 
-	LDBM_LOCK;
+	LDBM_RLOCK;
 	d = dbm_nextkey( ldbm );
-	LDBM_UNLOCK;
+	LDBM_RUNLOCK;
 
 	return d;
 }
@@ -1128,9 +1135,9 @@ ldbm_errno( LDBM ldbm )
 {
 	int err;
 
-	LDBM_LOCK;
+	LDBM_WLOCK;
 	err = dbm_error( ldbm );
-	LDBM_UNLOCK;
+	LDBM_WUNLOCK;
 
 	return err;
 }
