@@ -1,7 +1,7 @@
 /* shell.h - shell backend header file */
-/* $OpenLDAP: pkg/ldap/servers/slapd/back-shell/shell.h,v 1.7.2.2 2000/06/13 17:57:37 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/back-shell/shell.h,v 1.13 2002/01/04 20:17:55 kurt Exp $ */
 /*
- * Copyright 1998-2000 The OpenLDAP Foundation, All Rights Reserved.
+ * Copyright 1998-2002 The OpenLDAP Foundation, All Rights Reserved.
  * COPYING RESTRICTIONS APPLY, see COPYRIGHT file
  */
 
@@ -42,7 +42,7 @@ extern int read_and_send_results LDAP_P((
 	struct slap_conn *conn,
 	struct slap_op *op,
 	FILE *fp,
-	char **attrs,
+	AttributeName *attrs,
 	int attrsonly));
 
 LDAP_END_DECL
