@@ -1,7 +1,7 @@
 /* result.c - routines to send ldap results, errors, and referrals */
-/* $OpenLDAP: pkg/ldap/servers/slapd/result.c,v 1.46.2.4 2000/07/04 17:58:55 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/result.c,v 1.130.2.2 2002/04/18 15:19:59 kurt Exp $ */
 /*
- * Copyright 1998-2000 The OpenLDAP Foundation, All Rights Reserved.
+ * Copyright 1998-2002 The OpenLDAP Foundation, All Rights Reserved.
  * COPYING RESTRICTIONS APPLY, see COPYRIGHT file
  */
 
@@ -11,7 +11,6 @@
 
 #include <ac/socket.h>
 #include <ac/errno.h>
-#include <ac/signal.h>
 #include <ac/string.h>
 #include <ac/ctype.h>
 #include <ac/time.h>
@@ -19,40 +18,44 @@
 
 #include "slap.h"
 
-static char *v2ref( struct berval **ref, const char *text )
+static char *v2ref( BerVarray ref, const char *text )
 {
 	size_t len = 0, i = 0;
 	char *v2;
 
-	if(ref == NULL)
-	{
-	    if (text)
-		return ch_strdup(text);
-	    else
-		return NULL;
+	if(ref == NULL) {
+		if (text) {
+			return ch_strdup(text);
+		} else {
+			return NULL;
+		}
 	}
 	
-	if (text) {
+	if ( text != NULL ) {
 		len = strlen( text );
-		if (text[len-1] != '\n')
+		if (text[len-1] != '\n') {
 		    i = 1;
+		}
 	}
+
 	v2 = ch_malloc( len+i+sizeof("Referral:") );
-	if (text) {
+	if( text != NULL ) {
 		strcpy(v2, text);
-		if (i)
+		if( i ) {
 			v2[len++] = '\n';
+		}
 	}
 	strcpy( v2+len, "Referral:" );
 	len += sizeof("Referral:");
 
-	for( i=0; ref[i] != NULL; i++ ) {
-		v2 = ch_realloc( v2, len + ref[i]->bv_len + 1 );
+	for( i=0; ref[i].bv_val != NULL; i++ ) {
+		v2 = ch_realloc( v2, len + ref[i].bv_len + 1 );
 		v2[len-1] = '\n';
-		memcpy(&v2[len], ref[i]->bv_val, ref[i]->bv_len );
-		len += ref[i]->bv_len;
-		if (ref[i]->bv_val[ref[i]->bv_len-1] != '/')
+		AC_MEMCPY(&v2[len], ref[i].bv_val, ref[i].bv_len );
+		len += ref[i].bv_len;
+		if (ref[i].bv_val[ref[i].bv_len-1] != '/') {
 			++len;
+		}
 	}
 
 	v2[len-1] = '\0';
@@ -85,93 +88,12 @@ static ber_tag_t req2res( ber_tag_t tag )
 		break;
 
 	default:
-		assert( 0 );
-		tag = LBER_ERROR;
+		tag = LBER_SEQUENCE;
 	}
+
 	return tag;
 }
 
-static void trim_refs_urls(
-	struct berval **refs )
-{
-	unsigned i;
-
-	if( refs == NULL ) return;
-
-	for( i=0; refs[i] != NULL; i++ ) {
-		if(	refs[i]->bv_len > sizeof("ldap://")-1 &&
-			strncasecmp( refs[i]->bv_val, "ldap://",
-				sizeof("ldap://")-1 ) == 0 )
-		{
-			unsigned j;
-			for( j=sizeof("ldap://")-1; j<refs[i]->bv_len ; j++ ) {
-				if( refs[i]->bv_val[j] == '/' ) {
-					refs[i]->bv_val[j] = '\0';
-					refs[i]->bv_len = j;
-					break;
-				}
-			}
-		}
-	}
-}
-
-struct berval **get_entry_referrals(
-	Backend *be,
-	Connection *conn,
-	Operation *op,
-	Entry *e )
-{
-	Attribute *attr;
-	struct berval **refs;
-	unsigned i, j;
-
-	AttributeDescription *ad_ref = slap_schema.si_ad_ref;
-
-	attr = attr_find( e->e_attrs, ad_ref );
-
-	if( attr == NULL ) return NULL;
-
-	for( i=0; attr->a_vals[i] != NULL; i++ ) {
-		/* count references */
-	}
-
-	if( i < 1 ) return NULL;
-
-	refs = ch_malloc( (i + 1) * sizeof(struct berval *));
-
-	for( i=0, j=0; attr->a_vals[i] != NULL; i++ ) {
-		unsigned k;
-		struct berval *ref = ber_bvdup( attr->a_vals[i] );
-
-		/* trim the label */
-		for( k=0; k<ref->bv_len; k++ ) {
-			if( isspace(ref->bv_val[k]) ) {
-				ref->bv_val[k] = '\0';
-				ref->bv_len = k;
-				break;
-			}
-		}
-
-		if(	ref->bv_len > 0 ) {
-			refs[j++] = ref;
-
-		} else {
-			ber_bvfree( ref );
-		}
-	}
-
-	refs[j] = NULL;
-
-	if( j == 0 ) {
-		ber_bvecfree( refs );
-		refs = NULL;
-	}
-
-	/* we should check that a referral value exists... */
-
-	return refs;
-}
-
 static long send_ldap_ber(
 	Connection *conn,
 	BerElement *ber )
@@ -210,8 +132,14 @@ static long send_ldap_ber(
 		 * it's a hard error and return.
 		 */
 
+#ifdef NEW_LOGGING
+		LDAP_LOG(( "operation", LDAP_LEVEL_ERR,
+			   "send_ldap_ber: conn %d  ber_flush failed err=%d (%s)\n",
+			   conn ? conn->c_connid : 0, err, sock_errstr(err) ));
+#else
 		Debug( LDAP_DEBUG_CONNS, "ber_flush failed errno=%d reason=\"%s\"\n",
 		    err, sock_errstr(err), 0 );
+#endif
 
 		if ( err != EWOULDBLOCK && err != EAGAIN ) {
 			connection_closing( conn );
@@ -246,90 +174,143 @@ send_ldap_response(
     ber_int_t	err,
     const char	*matched,
     const char	*text,
-	struct berval	**ref,
+	BerVarray	ref,
 	const char	*resoid,
 	struct berval	*resdata,
 	struct berval	*sasldata,
 	LDAPControl **ctrls
 )
 {
-	BerElement	*ber;
+	char berbuf[256];
+	BerElement	*ber = (BerElement *)berbuf;
 	int		rc;
 	long	bytes;
 
+	if (op->o_callback && op->o_callback->sc_response) {
+		op->o_callback->sc_response( conn, op, tag, msgid, err, matched,
+			text, ref, resoid, resdata, sasldata, ctrls );
+		return;
+	}
+		
 	assert( ctrls == NULL ); /* ctrls not implemented */
 
-	ber = ber_alloc_t( LBER_USE_DER );
+	ber_init_w_nullc( ber, LBER_USE_DER );
 
-	Debug( LDAP_DEBUG_TRACE, "send_ldap_response: msgid=%ld tag=%ld err=%ld\n",
+#ifdef NEW_LOGGING
+	LDAP_LOG(( "operation", LDAP_LEVEL_ENTRY,
+		   "send_ldap_response: conn %d	 msgid=%ld tag=%ld err=%ld\n",
+		   conn ? conn->c_connid : 0, (long)msgid, (long)tag, (long)err ));
+#else
+	Debug( LDAP_DEBUG_TRACE,
+		"send_ldap_response: msgid=%ld tag=%ld err=%ld\n",
 		(long) msgid, (long) tag, (long) err );
+#endif
+
 	if( ref ) {
-		Debug( LDAP_DEBUG_ARGS, "send_ldap_response: ref=%s\n",
-			ref[0] && ref[0]->bv_val ? ref[0]->bv_val : "NULL",
+#ifdef NEW_LOGGING
+		LDAP_LOG(( "operation", LDAP_LEVEL_ARGS,
+			   "send_ldap_response: conn %d  ref=\"%s\"\n",
+			   conn ? conn->c_connid : 0,
+			   ref[0].bv_val ? ref[0].bv_val : "NULL" ));
+#else
+		Debug( LDAP_DEBUG_ARGS, "send_ldap_response: ref=\"%s\"\n",
+			ref[0].bv_val ? ref[0].bv_val : "NULL",
 			NULL, NULL );
-	}
+#endif
 
-	if ( ber == NULL ) {
-		Debug( LDAP_DEBUG_ANY, "ber_alloc failed\n", 0, 0, 0 );
-		return;
 	}
 
 #ifdef LDAP_CONNECTIONLESS
-	if ( op->o_cldap ) {
-		rc = ber_printf( ber, "{is{t{essN}N}N}", msgid, "", tag,
-		    err, matched ? matched : "", text ? text : "" );
+	if (conn->c_is_udp) {
+	    rc = ber_write(ber, (char *)&op->o_peeraddr, sizeof(struct sockaddr), 0);
+	    if (rc != sizeof(struct sockaddr)) {
+#ifdef NEW_LOGGING
+		LDAP_LOG(( "operation", LDAP_LEVEL_ERR,
+			   "send_ldap_response: conn %d  ber_write failed\n",
+			   conn ? conn->c_connid : 0 ));
+#else
+		Debug( LDAP_DEBUG_ANY, "ber_write failed\n", 0, 0, 0 );
+#endif
+		ber_free_buf( ber );
+		return;
+	    }
+	}
+	if (conn->c_is_udp && op->o_protocol == LDAP_VERSION2) {
+	    rc = ber_printf( ber, "{is{t{ess",
+		msgid, "", tag, err,
+		matched == NULL ? "" : matched,
+		text == NULL ? "" : text );
 	} else
 #endif
 	{
-		rc = ber_printf( ber, "{it{ess",
-			msgid, tag, err,
-			matched == NULL ? "" : matched,
-			text == NULL ? "" : text );
+	    rc = ber_printf( ber, "{it{ess",
+		msgid, tag, err,
+		matched == NULL ? "" : matched,
+		text == NULL ? "" : text );
+	}
 
-		if( rc != -1 ) {
-			if ( ref != NULL ) {
-				assert( err == LDAP_REFERRAL );
-				rc = ber_printf( ber, "t{V}",
-					LDAP_TAG_REFERRAL, ref );
-			} else {
-				assert( err != LDAP_REFERRAL );
-			}
-		}
-
-		if( rc != -1 && sasldata != NULL ) {
-			rc = ber_printf( ber, "tO",
-				LDAP_TAG_SASL_RES_CREDS, sasldata );
-		}
-
-		if( rc != -1 && resoid != NULL ) {
-			rc = ber_printf( ber, "ts",
-				LDAP_TAG_EXOP_RES_OID, resoid );
-		}
-
-		if( rc != -1 && resdata != NULL ) {
-			rc = ber_printf( ber, "tO",
-				LDAP_TAG_EXOP_RES_VALUE, resdata );
-		}
-
-		if( rc != -1 ) {
-			rc = ber_printf( ber, "N}N}" );
+	if( rc != -1 ) {
+		if ( ref != NULL ) {
+			assert( err == LDAP_REFERRAL );
+			rc = ber_printf( ber, "t{W}",
+				LDAP_TAG_REFERRAL, ref );
+		} else {
+			assert( err != LDAP_REFERRAL );
 		}
 	}
 
+	if( rc != -1 && sasldata != NULL ) {
+		rc = ber_printf( ber, "tO",
+			LDAP_TAG_SASL_RES_CREDS, sasldata );
+	}
+
+	if( rc != -1 && resoid != NULL ) {
+		rc = ber_printf( ber, "ts",
+			LDAP_TAG_EXOP_RES_OID, resoid );
+	}
+
+	if( rc != -1 && resdata != NULL ) {
+		rc = ber_printf( ber, "tO",
+			LDAP_TAG_EXOP_RES_VALUE, resdata );
+	}
+
+	if( rc != -1 ) {
+		rc = ber_printf( ber, "N}N}" );
+	}
+#ifdef LDAP_CONNECTIONLESS
+	if( conn->c_is_udp && op->o_protocol == LDAP_VERSION2 && rc != -1 ) {
+		rc = ber_printf( ber, "N}" );
+	}
+#endif
+
 	if ( rc == -1 ) {
+#ifdef NEW_LOGGING
+		LDAP_LOG(( "operation", LDAP_LEVEL_ERR,
+			   "send_ldap_response: conn %d  ber_printf failed\n",
+			   conn ? conn->c_connid : 0 ));
+#else
 		Debug( LDAP_DEBUG_ANY, "ber_printf failed\n", 0, 0, 0 );
-		ber_free( ber, 1 );
+#endif
+
+		ber_free_buf( ber );
 		return;
 	}
 
 	/* send BER */
 	bytes = send_ldap_ber( conn, ber );
-	ber_free( ber, 1 );
+	ber_free_buf( ber );
 
 	if ( bytes < 0 ) {
+#ifdef NEW_LOGGING
+		LDAP_LOG(( "operation", LDAP_LEVEL_ERR,
+			   "send_ldap_response: conn %d ber write failed\n",
+			   conn ? conn->c_connid : 0 ));
+#else
 		Debug( LDAP_DEBUG_ANY,
 			"send_ldap_response: ber write failed\n",
 			0, 0, 0 );
+#endif
+
 		return;
 	}
 
@@ -360,9 +341,16 @@ send_ldap_disconnect(
 
 	assert( LDAP_UNSOLICITED_ERROR( err ) );
 
+#ifdef NEW_LOGGING
+	LDAP_LOG(( "operation", LDAP_LEVEL_ENTRY,
+		   "send_ldap_disconnect: conn %d  %d:%s\n",
+		   conn ? conn->c_connid : 0, err, text ? text : "" ));
+#else
 	Debug( LDAP_DEBUG_TRACE,
 		"send_ldap_disconnect %d:%s\n",
 		err, text ? text : "", NULL );
+#endif
+
 
 	if ( op->o_protocol < LDAP_VERSION3 ) {
 		reqoid = NULL;
@@ -375,26 +363,14 @@ send_ldap_disconnect(
 		msgid = 0;
 	}
 
-#ifdef LDAP_CONNECTIONLESS
-	if ( op->o_cldap ) {
-		ber_sockbuf_ctrl( conn->c_sb, LBER_SB_OPT_UDP_SET_DST,
-		    (void *)&op->o_clientaddr );
-		Debug( LDAP_DEBUG_TRACE, "UDP response to %s port %d\n", 
-		    inet_ntoa(((struct sockaddr_in *)
-		    &op->o_clientaddr)->sin_addr ),
-		    ((struct sockaddr_in *) &op->o_clientaddr)->sin_port,
-		    0 );
-	}
-#endif
-
 	send_ldap_response( conn, op, tag, msgid,
 		err, NULL, text, NULL,
 		reqoid, NULL, NULL, NULL );
 
 	Statslog( LDAP_DEBUG_STATS,
-	    "conn=%ld op=%ld DISCONNECT err=%ld tag=%lu text=%s\n",
+	    "conn=%ld op=%ld DISCONNECT tag=%lu err=%ld text=%s\n",
 		(long) op->o_connid, (long) op->o_opid,
-		(long) tag, (long) err, text ? text : "" );
+		(unsigned long) tag, (long) err, text ? text : "" );
 }
 
 void
@@ -404,7 +380,7 @@ send_ldap_result(
     ber_int_t	err,
     const char	*matched,
     const char	*text,
-	struct berval **ref,
+	BerVarray ref,
 	LDAPControl **ctrls
 )
 {
@@ -414,22 +390,42 @@ send_ldap_result(
 
 	assert( !LDAP_API_ERROR( err ) );
 
-	Debug( LDAP_DEBUG_TRACE, "send_ldap_result: conn=%ld op=%ld p=%d\n",
+#ifdef NEW_LOGGING
+	LDAP_LOG(( "operation", LDAP_LEVEL_ENTRY,
+		   "send_ldap_result : conn %ld	  op=%ld p=%d\n",
+		   (long)op->o_connid, (long)op->o_opid, op->o_protocol ));
+#else
+	Debug( LDAP_DEBUG_TRACE,
+		"send_ldap_result: conn=%ld op=%ld p=%d\n",
 		(long) op->o_connid, (long) op->o_opid, op->o_protocol );
-	Debug( LDAP_DEBUG_ARGS, "send_ldap_result: %d:%s:%s\n",
-		err, matched ?  matched : "", text ? text : "" );
+#endif
+
+#ifdef NEW_LOGGING
+	LDAP_LOG(( "operation", LDAP_LEVEL_ARGS,
+		   "send_ldap_result: conn=%ld err=%d matched=\"%s\" text=\"%s\"\n",
+		   (long)op->o_connid, err, matched ? matched : "", text ? text : "" ));
+#else
+	Debug( LDAP_DEBUG_ARGS,
+		"send_ldap_result: err=%d matched=\"%s\" text=\"%s\"\n",
+		err, matched ?	matched : "", text ? text : "" );
+#endif
+
+
 	if( ref ) {
-	Debug( LDAP_DEBUG_ARGS, "send_ldap_result: referral: %s\n",
-		ref[0] && ref[0]->bv_val ? ref[0]->bv_val : "NULL",
-		NULL, NULL );
+#ifdef NEW_LOGGING
+		LDAP_LOG(( "operation", LDAP_LEVEL_ARGS,
+			"send_ldap_result: referral=\"%s\"\n",
+			ref[0].bv_val ? ref[0].bv_val : "NULL" ));
+#else
+		Debug( LDAP_DEBUG_ARGS,
+			"send_ldap_result: referral=\"%s\"\n",
+			ref[0].bv_val ? ref[0].bv_val : "NULL",
+			NULL, NULL );
+#endif
 	}
 
 	assert( err != LDAP_PARTIAL_RESULTS );
 
-	if( op->o_tag != LDAP_REQ_SEARCH ) {
-		trim_refs_urls( ref );
-	}
-
 	if ( err == LDAP_REFERRAL ) {
 		if( ref == NULL ) {
 			err = LDAP_NO_SUCH_OBJECT;
@@ -447,18 +443,6 @@ send_ldap_result(
 	tag = req2res( op->o_tag );
 	msgid = (tag != LBER_SEQUENCE) ? op->o_msgid : 0;
 
-#ifdef LDAP_CONNECTIONLESS
-	if ( op->o_cldap ) {
-		ber_sockbuf_ctrl( conn->c_sb, LBER_SB_OPT_UDP_SET_DST,
-		    (void *)&op->o_clientaddr );
-		Debug( LDAP_DEBUG_TRACE, "UDP response to %s port %d\n", 
-		    inet_ntoa(((struct sockaddr_in *)
-		    &op->o_clientaddr)->sin_addr ),
-		    ((struct sockaddr_in *) &op->o_clientaddr)->sin_port,
-		    0 );
-	}
-#endif
-
 	send_ldap_response( conn, op, tag, msgid,
 		err, matched, text, ref,
 		NULL, NULL, NULL, ctrls );
@@ -466,7 +450,7 @@ send_ldap_result(
 	Statslog( LDAP_DEBUG_STATS,
 	    "conn=%ld op=%ld RESULT tag=%lu err=%ld text=%s\n",
 		(long) op->o_connid, (long) op->o_opid,
-		(long) tag, (long) err, text ? text : "" );
+		(unsigned long) tag, (long) err, text ? text : "" );
 
 	if( tmp != NULL ) {
 		ch_free(tmp);
@@ -480,7 +464,7 @@ send_ldap_sasl(
     ber_int_t	err,
     const char	*matched,
     const char	*text,
-	struct berval **ref,
+	BerVarray ref,
 	LDAPControl **ctrls,
 	struct berval *cred
 )
@@ -488,27 +472,22 @@ send_ldap_sasl(
 	ber_tag_t tag;
 	ber_int_t msgid;
 
-	Debug( LDAP_DEBUG_TRACE, "send_ldap_sasl %ld\n",
-		(long) err, NULL, NULL );
+#ifdef NEW_LOGGING
+	LDAP_LOG(( "operation", LDAP_LEVEL_ENTRY,
+		   "send_ldap_sasl: conn %d err=%ld len=%ld\n",
+		   op->o_connid, (long)err, cred ? cred->bv_len : -1 ));
+#else
+	Debug( LDAP_DEBUG_TRACE, "send_ldap_sasl: err=%ld len=%ld\n",
+		(long) err, cred ? cred->bv_len : -1, NULL );
+#endif
+
 
 	tag = req2res( op->o_tag );
 	msgid = (tag != LBER_SEQUENCE) ? op->o_msgid : 0;
 
-#ifdef LDAP_CONNECTIONLESS
-	if ( op->o_cldap ) {
-		ber_sockbuf_ctrl( conn->c_sb, LBER_SB_OPT_UDP_SET_DST,
-		    (void *)&op->o_clientaddr );
-		Debug( LDAP_DEBUG_TRACE, "UDP response to %s port %d\n", 
-		    inet_ntoa(((struct sockaddr_in *)
-		    &op->o_clientaddr)->sin_addr ),
-		    ((struct sockaddr_in *) &op->o_clientaddr)->sin_port,
-		    0 );
-	}
-#endif
-
 	send_ldap_response( conn, op, tag, msgid,
 		err, matched, text, ref,
-		NULL, NULL, cred, ctrls  );
+		NULL, NULL, cred, ctrls	 );
 }
 
 void
@@ -518,7 +497,7 @@ send_ldap_extended(
     ber_int_t	err,
     const char	*matched,
     const char	*text,
-    struct berval **refs,
+    BerVarray	refs,
     const char		*rspoid,
 	struct berval *rspdata,
 	LDAPControl **ctrls
@@ -527,27 +506,23 @@ send_ldap_extended(
 	ber_tag_t tag;
 	ber_int_t msgid;
 
+#ifdef NEW_LOGGING
+	LDAP_LOG(( "operation", LDAP_LEVEL_ENTRY,
+		   "send_ldap_extended: conn %d	 err=%ld oid=%s len=%ld\n",
+		   op->o_connid, (long)err, rspoid ? rspoid : "",
+		   rspdata != NULL ? (long)rspdata->bv_len : (long)0 ));
+#else
 	Debug( LDAP_DEBUG_TRACE,
-		"send_ldap_extended %ld:%s (%ld)\n",
+		"send_ldap_extended err=%ld oid=%s len=%ld\n",
 		(long) err,
 		rspoid ? rspoid : "",
 		rspdata != NULL ? (long) rspdata->bv_len : (long) 0 );
+#endif
+
 
 	tag = req2res( op->o_tag );
 	msgid = (tag != LBER_SEQUENCE) ? op->o_msgid : 0;
 
-#ifdef LDAP_CONNECTIONLESS
-	if ( op->o_cldap ) {
-		ber_sockbuf_ctrl( conn->c_sb, LBER_SB_OPT_UDP_SET_DST,
-		    (void *)&op->o_clientaddr );
-		Debug( LDAP_DEBUG_TRACE, "UDP response to %s port %d\n", 
-		    inet_ntoa(((struct sockaddr_in *)
-		    &op->o_clientaddr)->sin_addr ),
-		    ((struct sockaddr_in *) &op->o_clientaddr)->sin_port,
-		    0 );
-	}
-#endif
-
 	send_ldap_response( conn, op, tag, msgid,
 		err, matched, text, refs,
 		rspoid, rspdata, NULL, ctrls );
@@ -561,7 +536,7 @@ send_search_result(
     ber_int_t	err,
     const char	*matched,
 	const char	*text,
-    struct berval **refs,
+    BerVarray	refs,
 	LDAPControl **ctrls,
     int		nentries
 )
@@ -569,15 +544,29 @@ send_search_result(
 	ber_tag_t tag;
 	ber_int_t msgid;
 	char *tmp = NULL;
+
 	assert( !LDAP_API_ERROR( err ) );
 
-	Debug( LDAP_DEBUG_TRACE, "send_ldap_search_result %d:%s:%s\n",
-		err, matched ?  matched : "", text ? text : "" );
+	if (op->o_callback && op->o_callback->sc_sresult) {
+		op->o_callback->sc_sresult(conn, op, err, matched, text, refs,
+			ctrls, nentries);
+		return;
+	}
+
+#ifdef NEW_LOGGING
+	LDAP_LOG(( "operation", LDAP_LEVEL_ENTRY,
+		   "send_search_result: conn %d err=%d matched=\"%s\"\n",
+		   op->o_connid, err, matched ? matched : "",
+		   text ? text : "" ));
+#else
+	Debug( LDAP_DEBUG_TRACE,
+		"send_search_result: err=%d matched=\"%s\" text=\"%s\"\n",
+		err, matched ?	matched : "", text ? text : "" );
+#endif
+
 
 	assert( err != LDAP_PARTIAL_RESULTS );
 
-	trim_refs_urls( refs );
-
 	if( op->o_protocol < LDAP_VERSION3 ) {
 		/* send references in search results */
 		if( err == LDAP_REFERRAL ) {
@@ -587,6 +576,7 @@ send_search_result(
 		tmp = v2ref( refs, text );
 		text = tmp;
 		refs = NULL;
+
 	} else {
 		/* don't send references in search results */
 		assert( refs == NULL );
@@ -600,31 +590,29 @@ send_search_result(
 	tag = req2res( op->o_tag );
 	msgid = (tag != LBER_SEQUENCE) ? op->o_msgid : 0;
 
-#ifdef LDAP_CONNECTIONLESS
-	if ( op->o_cldap ) {
-		ber_sockbuf_ctrl( conn->c_sb, LBER_SB_OPT_UDP_SET_DST,
-		    (void *)&op->o_clientaddr );
-		Debug( LDAP_DEBUG_TRACE, "UDP response to %s port %d\n", 
-		    inet_ntoa(((struct sockaddr_in *)
-		    &op->o_clientaddr)->sin_addr ),
-		    ((struct sockaddr_in *) &op->o_clientaddr)->sin_port,
-		    0 );
-	}
-#endif
-
 	send_ldap_response( conn, op, tag, msgid,
 		err, matched, text, refs,
 		NULL, NULL, NULL, ctrls );
 
-	Statslog( LDAP_DEBUG_STATS,
-	    "conn=%ld op=%ld SEARCH RESULT tag=%lu err=%ld text=%s\n",
-		(long) op->o_connid, (long) op->o_opid,
-		(long) tag, (long) err, text ? text : "" );
+	{
+	char nbuf[64];
+	sprintf( nbuf, "%ld nentries=%d", (long) err, nentries );
 
-	if (tmp != NULL)
+	Statslog( LDAP_DEBUG_STATS,
+	    "conn=%ld op=%ld SEARCH RESULT tag=%lu err=%s text=%s\n",
+		(long) op->o_connid, (long) op->o_opid,
+		(unsigned long) tag, nbuf, text ? text : "" );
+	}
+
+	if (tmp != NULL) {
 	    ch_free(tmp);
+	}
 }
 
+static struct berval AllUser = { sizeof(LDAP_ALL_USER_ATTRIBUTES)-1,
+	LDAP_ALL_USER_ATTRIBUTES };
+static struct berval AllOper = { sizeof(LDAP_ALL_OPERATIONAL_ATTRIBUTES)-1,
+	LDAP_ALL_OPERATIONAL_ATTRIBUTES };
 
 int
 send_search_entry(
@@ -632,47 +620,94 @@ send_search_entry(
     Connection	*conn,
     Operation	*op,
     Entry	*e,
-    char	**attrs,
+    AttributeName	*attrs,
     int		attrsonly,
 	LDAPControl **ctrls
 )
 {
-	BerElement	*ber;
+	char		berbuf[256];
+	BerElement	*ber = (BerElement *)berbuf;
 	Attribute	*a, *aa;
 	int		i, rc=-1, bytes;
-	char            *edn;
+	char		*edn;
 	int		userattrs;
 	int		opattrs;
+	static AccessControlState acl_state_init = ACL_STATE_INIT;
+	AccessControlState acl_state;
 
 	AttributeDescription *ad_entry = slap_schema.si_ad_entry;
 
-	Debug( LDAP_DEBUG_TRACE, "=> send_search_entry: \"%s\"\n", e->e_dn, 0, 0 );
+	if (op->o_callback && op->o_callback->sc_sendentry) {
+		return op->o_callback->sc_sendentry( be, conn, op, e, attrs,
+			attrsonly, ctrls );
+	}
+
+#ifdef NEW_LOGGING
+	LDAP_LOG(( "operation", LDAP_LEVEL_ENTRY,
+		   "send_search_entry: conn %d	dn=\"%s\"%s\n",
+		   op->o_connid, e->e_dn,
+		   attrsonly ? " (attrsOnly)" : "" ));
+#else
+	Debug( LDAP_DEBUG_TRACE,
+		"=> send_search_entry: dn=\"%s\"%s\n",
+		e->e_dn, attrsonly ? " (attrsOnly)" : "", 0 );
+#endif
 
 	if ( ! access_allowed( be, conn, op, e,
-		ad_entry, NULL, ACL_READ ) )
+		ad_entry, NULL, ACL_READ, NULL ) )
 	{
-		Debug( LDAP_DEBUG_ACL, "acl: access to entry not allowed\n",
+#ifdef NEW_LOGGING
+		LDAP_LOG(( "acl", LDAP_LEVEL_INFO,
+			   "send_search_entry: conn %d access to entry (%s) not allowed\n",
+			   op->o_connid, e->e_dn ));
+#else
+		Debug( LDAP_DEBUG_ACL,
+			"send_search_entry: access to entry not allowed\n",
 		    0, 0, 0 );
+#endif
+
 		return( 1 );
 	}
 
 	edn = e->e_ndn;
 
-	ber = ber_alloc_t( LBER_USE_DER );
+	ber_init_w_nullc( ber, LBER_USE_DER );
 
-	if ( ber == NULL ) {
-		Debug( LDAP_DEBUG_ANY, "ber_alloc failed\n", 0, 0, 0 );
-		send_ldap_result( conn, op, LDAP_OTHER,
-			NULL, "BER allocation error", NULL, NULL );
-		goto error_return;
+#ifdef LDAP_CONNECTIONLESS
+	if (conn->c_is_udp) {
+	    rc = ber_write(ber, (char *)&op->o_peeraddr, sizeof(struct sockaddr), 0);
+	    if (rc != sizeof(struct sockaddr)) {
+#ifdef NEW_LOGGING
+		LDAP_LOG(( "operation", LDAP_LEVEL_ERR,
+			   "send_search_entry: conn %d  ber_printf failed\n",
+			   conn ? conn->c_connid : 0 ));
+#else
+		Debug( LDAP_DEBUG_ANY, "ber_printf failed\n", 0, 0, 0 );
+#endif
+		ber_free_buf( ber );
+		return;
+	    }
+	}
+	if (conn->c_is_udp && op->o_protocol == LDAP_VERSION2) {
+	    rc = ber_printf( ber, "{is{t{O{" /*}}}*/,
+		op->o_msgid, "", LDAP_RES_SEARCH_ENTRY, &e->e_name );
+	} else
+#endif
+	{
+	    rc = ber_printf( ber, "{it{O{" /*}}}*/, op->o_msgid,
+		LDAP_RES_SEARCH_ENTRY, &e->e_name );
 	}
 
-	rc = ber_printf( ber, "{it{s{" /*}}}*/, op->o_msgid,
-		LDAP_RES_SEARCH_ENTRY, e->e_dn );
-
 	if ( rc == -1 ) {
+#ifdef NEW_LOGGING
+		LDAP_LOG(( "operation", LDAP_LEVEL_ERR,
+			   "send_search_entry: conn %d  ber_printf failed\n",
+			   op->o_connid ));
+#else
 		Debug( LDAP_DEBUG_ANY, "ber_printf failed\n", 0, 0, 0 );
-		ber_free( ber, 1 );
+#endif
+
+		ber_free_buf( ber );
 		send_ldap_result( conn, op, LDAP_OTHER,
 		    NULL, "encoding DN error", NULL, NULL );
 		goto error_return;
@@ -680,30 +715,28 @@ send_search_entry(
 
 	/* check for special all user attributes ("*") type */
 	userattrs = ( attrs == NULL ) ? 1
-		: charray_inlist( attrs, LDAP_ALL_USER_ATTRIBUTES );
+		: an_find( attrs, &AllUser );
 
 	/* check for special all operational attributes ("+") type */
 	opattrs = ( attrs == NULL ) ? 0
-		: charray_inlist( attrs, LDAP_ALL_OPERATIONAL_ATTRIBUTES );
+		: an_find( attrs, &AllOper );
 
 	for ( a = e->e_attrs; a != NULL; a = a->a_next ) {
 		AttributeDescription *desc = a->a_desc;
-		char *type = desc->ad_cname->bv_val;
 
 		if ( attrs == NULL ) {
-			/* all addrs request, skip operational attributes */
-			if( is_at_operational( desc->ad_type ) )
-			{
+			/* all attrs request, skip operational attributes */
+			if( is_at_operational( desc->ad_type ) ) {
 				continue;
 			}
 
 		} else {
-			/* specific addrs requested */
-			if ( is_at_operational( desc->ad_type ) )
-			{
+			/* specific attrs requested */
+			if ( is_at_operational( desc->ad_type ) ) {
 				if( !opattrs && !ad_inlist( desc, attrs ) ) {
 					continue;
 				}
+
 			} else {
 				if (!userattrs && !ad_inlist( desc, attrs ) ) {
 					continue;
@@ -711,35 +744,68 @@ send_search_entry(
 			}
 		}
 
-		if ( ! access_allowed( be, conn, op, e, desc, NULL, ACL_READ ) ) {
-			Debug( LDAP_DEBUG_ACL, "acl: access to attribute %s not allowed\n",
-			    desc->ad_cname->bv_val, 0, 0 );
+		acl_state = acl_state_init;
+
+		if ( ! access_allowed( be, conn, op, e, desc, NULL,
+			ACL_READ, &acl_state ) )
+		{
+#ifdef NEW_LOGGING
+			LDAP_LOG(( "acl", LDAP_LEVEL_INFO, "send_search_entry: "
+				"conn %d  access to attribute %s not allowed\n",
+				op->o_connid, desc->ad_cname.bv_val ));
+#else
+			Debug( LDAP_DEBUG_ACL, "acl: "
+				"access to attribute %s not allowed\n",
+			    desc->ad_cname.bv_val, 0, 0 );
+#endif
 			continue;
 		}
 
-		if (( rc = ber_printf( ber, "{s[" /*]}*/ , type )) == -1 ) {
+		if (( rc = ber_printf( ber, "{O[" /*]}*/ , &desc->ad_cname )) == -1 ) {
+#ifdef NEW_LOGGING
+			LDAP_LOG(( "operation", LDAP_LEVEL_ERR,
+				"send_search_entry: conn %d  ber_printf failed\n",
+				op->o_connid ));
+#else
 			Debug( LDAP_DEBUG_ANY, "ber_printf failed\n", 0, 0, 0 );
-			ber_free( ber, 1 );
+#endif
+
+			ber_free_buf( ber );
 			send_ldap_result( conn, op, LDAP_OTHER,
 			    NULL, "encoding description error", NULL, NULL );
 			goto error_return;
 		}
 
 		if ( ! attrsonly ) {
-			for ( i = 0; a->a_vals[i] != NULL; i++ ) {
+			for ( i = 0; a->a_vals[i].bv_val != NULL; i++ ) {
 				if ( ! access_allowed( be, conn, op, e,
-					desc, a->a_vals[i], ACL_READ ) )
+					desc, &a->a_vals[i], ACL_READ, &acl_state ) )
 				{
+#ifdef NEW_LOGGING
+					LDAP_LOG(( "acl", LDAP_LEVEL_INFO,
+						"send_search_entry: conn %d "
+						"access to attribute %s, value %d not allowed\n",
+						op->o_connid, desc->ad_cname.bv_val, i ));
+#else
 					Debug( LDAP_DEBUG_ACL,
 						"acl: access to attribute %s, value %d not allowed\n",
-			    		desc->ad_cname->bv_val, i, 0 );
+					desc->ad_cname.bv_val, i, 0 );
+#endif
+
 					continue;
 				}
 
-				if (( rc = ber_printf( ber, "O", a->a_vals[i] )) == -1 ) {
+				if (( rc = ber_printf( ber, "O", &a->a_vals[i] )) == -1 ) {
+#ifdef NEW_LOGGING
+					LDAP_LOG(( "operation", LDAP_LEVEL_ERR,
+						"send_search_entry: conn %d  ber_printf failed.\n",
+						op->o_connid ));
+#else
 					Debug( LDAP_DEBUG_ANY,
 					    "ber_printf failed\n", 0, 0, 0 );
-					ber_free( ber, 1 );
+#endif
+
+					ber_free_buf( ber );
 					send_ldap_result( conn, op, LDAP_OTHER,
 						NULL, "encoding values error", NULL, NULL );
 					goto error_return;
@@ -748,8 +814,15 @@ send_search_entry(
 		}
 
 		if (( rc = ber_printf( ber, /*{[*/ "]N}" )) == -1 ) {
+#ifdef NEW_LOGGING
+			LDAP_LOG(( "operation", LDAP_LEVEL_ERR,
+				"send_search_entry: conn %d  ber_printf failed\n",
+				op->o_connid ));
+#else
 			Debug( LDAP_DEBUG_ANY, "ber_printf failed\n", 0, 0, 0 );
-			ber_free( ber, 1 );
+#endif
+
+			ber_free_buf( ber );
 			send_ldap_result( conn, op, LDAP_OTHER,
 			    NULL, "encode end error", NULL, NULL );
 			goto error_return;
@@ -758,22 +831,21 @@ send_search_entry(
 
 	/* eventually will loop through generated operational attributes */
 	/* only have subschemaSubentry implemented */
-	aa = backend_operational( be, e );
+	aa = backend_operational( be, conn, op, e, attrs, opattrs );
 	
 	for (a = aa ; a != NULL; a = a->a_next ) {
 		AttributeDescription *desc = a->a_desc;
 
 		if ( attrs == NULL ) {
-			/* all addrs request, skip operational attributes */
+			/* all attrs request, skip operational attributes */
 			if( is_at_operational( desc->ad_type ) ) {
 				continue;
 			}
 
 		} else {
-			/* specific addrs requested */
+			/* specific attrs requested */
 			if( is_at_operational( desc->ad_type ) ) {
-				if( !opattrs && !ad_inlist( desc, attrs ) )
-				{
+				if( !opattrs && !ad_inlist( desc, attrs ) ) {
 					continue;
 				}
 			} else {
@@ -784,49 +856,92 @@ send_search_entry(
 			}
 		}
 
-		if ( ! access_allowed( be, conn, op, e,	desc, NULL, ACL_READ ) ) {
+		acl_state = acl_state_init;
+
+		if ( ! access_allowed( be, conn, op, e,	desc, NULL,
+			ACL_READ, &acl_state ) )
+		{
+#ifdef NEW_LOGGING
+			LDAP_LOG(( "acl", LDAP_LEVEL_INFO,
+				"send_search_entry: conn %s "
+				"access to attribute %s not allowed\n",
+				op->o_connid, desc->ad_cname.bv_val ));
+#else
 			Debug( LDAP_DEBUG_ACL, "acl: access to attribute %s not allowed\n",
-			    desc->ad_cname->bv_val, 0, 0 );
+			    desc->ad_cname.bv_val, 0, 0 );
+#endif
+
 			continue;
 		}
 
-		rc = ber_printf( ber, "{s[" /*]}*/ , desc->ad_cname->bv_val );
+		rc = ber_printf( ber, "{O[" /*]}*/ , &desc->ad_cname );
 		if ( rc == -1 ) {
+#ifdef NEW_LOGGING
+			LDAP_LOG(( "operation", LDAP_LEVEL_ERR,
+				"send_search_entry: conn %d  ber_printf failed\n",
+				op->o_connid ));
+#else
 			Debug( LDAP_DEBUG_ANY, "ber_printf failed\n", 0, 0, 0 );
-			ber_free( ber, 1 );
+#endif
+
+			ber_free_buf( ber );
 			send_ldap_result( conn, op, LDAP_OTHER,
 			    NULL, "encoding description error", NULL, NULL );
+			attrs_free( aa );
 			goto error_return;
 		}
 
 		if ( ! attrsonly ) {
-			for ( i = 0; a->a_vals[i] != NULL; i++ ) {
+			for ( i = 0; a->a_vals[i].bv_val != NULL; i++ ) {
 				if ( ! access_allowed( be, conn, op, e,
-					desc, a->a_vals[i], ACL_READ ) )
+					desc, &a->a_vals[i], ACL_READ, &acl_state ) )
 				{
+#ifdef NEW_LOGGING
+					LDAP_LOG(( "acl", LDAP_LEVEL_INFO,
+						"send_search_entry: conn %d "
+						"access to %s, value %d not allowed\n",
+						op->o_connid, desc->ad_cname.bv_val, i ));
+#else
 					Debug( LDAP_DEBUG_ACL,
 						"acl: access to attribute %s, value %d not allowed\n",
-			    		desc->ad_cname->bv_val, i, 0 );
+					desc->ad_cname.bv_val, i, 0 );
+#endif
+
 					continue;
 				}
 
-
-				if (( rc = ber_printf( ber, "O", a->a_vals[i] )) == -1 ) {
+				if (( rc = ber_printf( ber, "O", &a->a_vals[i] )) == -1 ) {
+#ifdef NEW_LOGGING
+					LDAP_LOG(( "operation", LDAP_LEVEL_ERR,
+						   "send_search_entry: conn %d  ber_printf failed\n",
+						   op->o_connid ));
+#else
 					Debug( LDAP_DEBUG_ANY,
 					    "ber_printf failed\n", 0, 0, 0 );
-					ber_free( ber, 1 );
+#endif
+
+					ber_free_buf( ber );
 					send_ldap_result( conn, op, LDAP_OTHER,
 						NULL, "encoding values error", NULL, NULL );
+					attrs_free( aa );
 					goto error_return;
 				}
 			}
 		}
 
 		if (( rc = ber_printf( ber, /*{[*/ "]N}" )) == -1 ) {
+#ifdef NEW_LOGGING
+			LDAP_LOG(( "operation", LDAP_LEVEL_ERR,
+				   "send_search_entry: conn %d  ber_printf failed\n",
+				   op->o_connid ));
+#else
 			Debug( LDAP_DEBUG_ANY, "ber_printf failed\n", 0, 0, 0 );
-			ber_free( ber, 1 );
+#endif
+
+			ber_free_buf( ber );
 			send_ldap_result( conn, op, LDAP_OTHER,
 			    NULL, "encode end error", NULL, NULL );
+			attrs_free( aa );
 			goto error_return;
 		}
 	}
@@ -835,21 +950,39 @@ send_search_entry(
 
 	rc = ber_printf( ber, /*{{{*/ "}N}N}" );
 
+#ifdef LDAP_CONNECTIONLESS
+	if (conn->c_is_udp && op->o_protocol == LDAP_VERSION2 && rc != -1)
+		rc = ber_printf( ber, "}" );
+#endif
 	if ( rc == -1 ) {
+#ifdef NEW_LOGGING
+		LDAP_LOG(( "operation", LDAP_LEVEL_ERR,
+			   "send_search_entry: conn %d ber_printf failed\n",
+			   op->o_connid ));
+#else
 		Debug( LDAP_DEBUG_ANY, "ber_printf failed\n", 0, 0, 0 );
-		ber_free( ber, 1 );
+#endif
+
+		ber_free_buf( ber );
 		send_ldap_result( conn, op, LDAP_OTHER,
 			NULL, "encode entry end error", NULL, NULL );
 		return( 1 );
 	}
 
 	bytes = send_ldap_ber( conn, ber );
-	ber_free( ber, 1 );
+	ber_free_buf( ber );
 
 	if ( bytes < 0 ) {
+#ifdef NEW_LOGGING
+		LDAP_LOG(( "operation", LDAP_LEVEL_ERR,
+			   "send_ldap_response: conn %d  ber write failed.\n",
+			   op->o_connid ));
+#else
 		Debug( LDAP_DEBUG_ANY,
 			"send_ldap_response: ber write failed\n",
 			0, 0, 0 );
+#endif
+
 		return -1;
 	}
 
@@ -862,7 +995,13 @@ send_search_entry(
 	Statslog( LDAP_DEBUG_STATS2, "conn=%ld op=%ld ENTRY dn=\"%s\"\n",
 	    (long) conn->c_connid, (long) op->o_opid, e->e_dn, 0, 0 );
 
+#ifdef NEW_LOGGING
+	LDAP_LOG(( "operation", LDAP_LEVEL_ENTRY,
+		   "send_search_entry: conn %d exit.\n",
+		   op->o_connid ));
+#else
 	Debug( LDAP_DEBUG_TRACE, "<= send_search_entry\n", 0, 0, 0 );
+#endif
 
 	rc = 0;
 
@@ -876,78 +1015,107 @@ send_search_reference(
     Connection	*conn,
     Operation	*op,
     Entry	*e,
-	struct berval **refs,
-	int scope,
+	BerVarray refs,
 	LDAPControl **ctrls,
-    struct berval ***v2refs
+    BerVarray *v2refs
 )
 {
-	BerElement	*ber;
+	char		berbuf[256];
+	BerElement	*ber = (BerElement *)berbuf;
 	int rc;
 	int bytes;
 
 	AttributeDescription *ad_ref = slap_schema.si_ad_ref;
 	AttributeDescription *ad_entry = slap_schema.si_ad_entry;
 
-	Debug( LDAP_DEBUG_TRACE, "=> send_search_reference (%s)\n", e->e_dn, 0, 0 );
+#ifdef NEW_LOGGING
+	LDAP_LOG(( "operation", LDAP_LEVEL_ENTRY,
+		"send_search_reference: conn %d  dn=\"%s\"\n",
+		op->o_connid, e->e_dn ));
+#else
+	Debug( LDAP_DEBUG_TRACE,
+		"=> send_search_reference: dn=\"%s\"\n",
+		e->e_dn, 0, 0 );
+#endif
+
 
 	if ( ! access_allowed( be, conn, op, e,
-		ad_entry, NULL, ACL_READ ) )
+		ad_entry, NULL, ACL_READ, NULL ) )
 	{
+#ifdef NEW_LOGGING
+		LDAP_LOG(( "acl", LDAP_LEVEL_INFO,
+			"send_search_reference: conn %d	access to entry %s not allowed\n",
+			op->o_connid, e->e_dn ));
+#else
 		Debug( LDAP_DEBUG_ACL,
 			"send_search_reference: access to entry not allowed\n",
 		    0, 0, 0 );
+#endif
+
 		return( 1 );
 	}
 
 	if ( ! access_allowed( be, conn, op, e,
-		ad_ref, NULL, ACL_READ ) )
+		ad_ref, NULL, ACL_READ, NULL ) )
 	{
+#ifdef NEW_LOGGING
+		LDAP_LOG(( "acl", LDAP_LEVEL_INFO,
+			"send_search_reference: conn %d access to reference not allowed.\n",
+			op->o_connid ));
+#else
 		Debug( LDAP_DEBUG_ACL,
 			"send_search_reference: access to reference not allowed\n",
 		    0, 0, 0 );
+#endif
+
 		return( 1 );
 	}
 
 	if( refs == NULL ) {
+#ifdef NEW_LOGGING
+		LDAP_LOG(( "operation", LDAP_LEVEL_ERR,
+			"send_search_reference: null ref in (%s).\n",
+			op->o_connid, e->e_dn ));
+#else
 		Debug( LDAP_DEBUG_ANY,
 			"send_search_reference: null ref in (%s)\n", 
 			e->e_dn, 0, 0 );
+#endif
+
 		return( 1 );
 	}
 
 	if( op->o_protocol < LDAP_VERSION3 ) {
 		/* save the references for the result */
-		if( *refs != NULL ) {
+		if( refs[0].bv_val != NULL ) {
 			value_add( v2refs, refs );
 		}
 		return 0;
 	}
 
-	ber = ber_alloc_t( LBER_USE_DER );
+	ber_init_w_nullc( ber, LBER_USE_DER );
 
-	if ( ber == NULL ) {
-		Debug( LDAP_DEBUG_ANY,
-			"send_search_reference: ber_alloc failed\n", 0, 0, 0 );
-		send_ldap_result( conn, op, LDAP_OTHER,
-			NULL, "alloc BER error", NULL, NULL );
-		return -1;
-	}
-
-	rc = ber_printf( ber, "{it{V}N}", op->o_msgid,
+	rc = ber_printf( ber, "{it{W}N}", op->o_msgid,
 		LDAP_RES_SEARCH_REFERENCE, refs );
 
 	if ( rc == -1 ) {
+#ifdef NEW_LOGGING
+		LDAP_LOG(( "operation", LDAP_LEVEL_ERR,
+			"send_search_reference: conn %d	ber_printf failed.\n",
+			op->o_connid ));
+#else
 		Debug( LDAP_DEBUG_ANY,
 			"send_search_reference: ber_printf failed\n", 0, 0, 0 );
-		ber_free( ber, 1 );
+#endif
+
+		ber_free_buf( ber );
 		send_ldap_result( conn, op, LDAP_OTHER,
 			NULL, "encode DN error", NULL, NULL );
 		return -1;
 	}
 
 	bytes = send_ldap_ber( conn, ber );
-	ber_free( ber, 1 );
+	ber_free_buf( ber );
 
 	ldap_pvt_thread_mutex_lock( &num_sent_mutex );
 	num_bytes_sent += bytes;
@@ -955,10 +1123,15 @@ send_search_reference(
 	num_pdu_sent++;
 	ldap_pvt_thread_mutex_unlock( &num_sent_mutex );
 
-	Statslog( LDAP_DEBUG_STATS2, "conn=%ld op=%ld ENTRY dn=\"%s\"\n",
-	    (long) conn->c_connid, (long) op->o_opid, e->e_dn, 0, 0 );
+	Statslog( LDAP_DEBUG_STATS2, "conn=%ld op=%ld REF dn=\"%s\"\n",
+		(long) conn->c_connid, (long) op->o_opid, e->e_dn, 0, 0 );
 
+#ifdef NEW_LOGGING
+	LDAP_LOG(( "operation", LDAP_LEVEL_ENTRY,
+		"send_search_reference: conn %d exit.\n", op->o_connid ));
+#else
 	Debug( LDAP_DEBUG_TRACE, "<= send_search_reference\n", 0, 0, 0 );
+#endif
 
 	return 0;
 }
@@ -980,8 +1153,14 @@ str2result(
 	*info = NULL;
 
 	if ( strncasecmp( s, "RESULT", 6 ) != 0 ) {
+#ifdef NEW_LOGGING
+		LDAP_LOG(( "operation", LDAP_LEVEL_INFO,
+			   "str2result: (%s), expecting \"RESULT\"\n", s ));
+#else
 		Debug( LDAP_DEBUG_ANY, "str2result (%s) expecting \"RESULT\"\n",
 		    s, 0, 0 );
+#endif
+
 
 		return( -1 );
 	}
@@ -1009,8 +1188,14 @@ str2result(
 				*info = c;
 			}
 		} else {
+#ifdef NEW_LOGGING
+			LDAP_LOG(( "operation", LDAP_LEVEL_INFO,
+				"str2result: (%s) unknown.\n", s ));
+#else
 			Debug( LDAP_DEBUG_ANY, "str2result (%s) unknown\n",
 			    s, 0, 0 );
+#endif
+
 			rc = -1;
 		}
 	}
